import java.awt.*;
import java.awt.image.*;
import java.applet.*;
import java.awt.MediaTracker;
import java.util.*;

public abstract class Animate extends OffScr implements Runnable {
    protected Thread    tx_animator;
    protected Thread    img_animator;

    protected int       tx_index;
    protected int       tx_num_images;
    protected int       tx_xoffset[];           // Text X offset
    protected int       tx_yoffset[];           // Text Y offset
    protected int       tx_width[];             // Text width
    protected int       tx_height[];            // Text height
    protected int       tx_x_coord[];           // Text X Coordinate
    protected int       tx_y_coord[];           // Text Y Coordinate
    protected Image     tx_osi[];
    protected int       tx_delay_images;
    protected int       tx_delay_runs;
    protected String    tx_filter_str[];
    protected int       tx_lower_bound;
    protected int       tx_filtered_pixels[][];
    protected AudioClip tx_pi_clip[];            // The per text Audio Clip
    protected String    tx_pi_audio_file[];      // The per text audio file

    protected Image     img[];                  // The image
    protected String    img_file[];
    protected int       img_index;
    protected ImgGetr   img_gettr;            // The Image Getter class
    protected int       img_pixels[][];           // The text image
    protected int       img_num_images;
    protected int       img_xoffset[];           // Image X offset
    protected int       img_yoffset[];           // Image Y offset
    protected int       img_width[];             // Image width
    protected int       img_height[];            // Image height
    protected int       img_width_orig[];        // Image width
    protected int       img_height_orig[];       // Image height
    protected int       img_x_coord[];           // Image X Coordinate
    protected int       img_y_coord[];           // Image Y Coordinate
    protected Image     img_osi[];
    protected int       img_delay_images;
    protected int       img_delay_runs;
    protected String    img_filter_str[];
    protected int       img_lastimage_index;
    protected int       img_lower_bound;
    protected boolean   img_loaded[];
    protected boolean   img_copyfrom[];
    protected int       img_reuseval[];
    protected boolean   img_nofilespec[];
    protected boolean   img_display_first;
    protected int       img_filtered_pixels[][];
    protected AudioClip img_pi_clip[];            // The per image Audio Clip
    protected String    img_pi_audio_file[];      // The per image audio file

    protected Image     offScreen;
    protected Graphics  offGC;

    protected int       tmp_pixels[];
    int applet_width = -1;
    int applet_height = -1;
    protected Filter f;
    protected boolean debugflag;
    protected boolean all_images_loaded;

    public void init() {
        int i;

        super.init();
        tx_index = -1;
        img_index = -1;
        all_images_loaded = false;
        debugflag        = GetParm.toBoolean(this, "Debug", false);
        tx_num_images    = GetParm.toInt(this, "TxNumImages", 0);
        tx_delay_images  = GetParm.toInt(this, "TxDelayBetweenImages", 100);
        tx_delay_runs    = GetParm.toInt(this, "TxDelayBetweenRuns", 2000);
        tx_lower_bound   = GetParm.toInt(this, "TxInitialImage", 0);

        img_num_images   = GetParm.toInt(this, "ImgNumImages", 0);
        img_delay_images = GetParm.toInt(this, "ImgDelayBetweenImages", 100);
        img_delay_runs   = GetParm.toInt(this, "ImgDelayBetweenRuns", 2000);
        img_lower_bound  = GetParm.toInt(this, "ImgInitialImage", 0);
        img_display_first= GetParm.toBoolean(this, "ImgDisplayFirst", false);

        // Since user number start at 1 not 0
        if(tx_lower_bound > 0)
            tx_lower_bound--;
        if(img_lower_bound > 0)
            img_lower_bound--;


        if(tx_num_images > 0) {
            tx_width      = new int[tx_num_images];
            tx_height     = new int[tx_num_images];
            tx_xoffset    = new int[tx_num_images];
            tx_yoffset    = new int[tx_num_images];
            tx_x_coord    = new int[tx_num_images];
            tx_y_coord    = new int[tx_num_images];
            tx_filter_str = new String[tx_num_images];
            tx_osi        = new Image[tx_num_images];
            tx_filtered_pixels = new int[tx_num_images][];
            tx_pi_clip    = new AudioClip[tx_num_images];
            tx_pi_audio_file = new String[tx_num_images];
        }

        if(img_num_images > 0) {
            img_file       = new String[img_num_images];
            img_width      = new int[img_num_images];
            img_height     = new int[img_num_images];
            img_width_orig = new int[img_num_images];
            img_height_orig= new int[img_num_images];
            img_xoffset    = new int[img_num_images];
            img_yoffset    = new int[img_num_images];
            img_x_coord    = new int[img_num_images];
            img_y_coord    = new int[img_num_images];
            img_filter_str = new String[img_num_images];
            img_osi        = new Image[img_num_images];
            img            = new Image[img_num_images];
            img_pixels     = new int[img_num_images][];
            img_loaded     = new boolean[img_num_images];
            img_copyfrom   = new boolean[img_num_images];
            img_nofilespec = new boolean[img_num_images];
            img_reuseval   = new int[img_num_images];
            img_filtered_pixels = new int[img_num_images][];
            img_pi_clip    = new AudioClip[img_num_images];
            img_pi_audio_file = new String[img_num_images];
        }

        f = new Filter();


        if(tx_num_images > 0 && tx != null) {
            ftext ps = new ftext();
            ps.setBackground(tx_bg_color!=null ? tx_bg_color : app_bg_color);
            ps.setTransparent(tx_bg_color!=null ? false : true);
            ps.setForeground(tx_color);
            ps.setFont(tx_font);
            ps.setApplet(this);
            ps.setText(tx);
            ps.setUnderline(tx_underline);
            tx_pixels = ps.filter(null, 0, 0);
            tx_width[0] = ps.getWidth();
            tx_height[0] = ps.getHeight();
            tx_ascent = tx_fontmetrics.getMaxAscent();
            tx_descent = tx_fontmetrics.getDescent();
        }


        for(i=0;i<tx_num_images;i++) {
            tx_filter_str[i] = GetParm.toString(this,
                                        "TxFilter"+(i+1), "noop");
            tx_xoffset[i]   = GetParm.toInt(this,
                                        "TxXOffset" + (i+1), 0x7FFFFFFF);
            tx_yoffset[i]   = GetParm.toInt(this,
                                        "TxYOffset" + (i+1), 0x7FFFFFFF);

            if(i == 0) {
                if(tx_xoffset[0] == 0x7FFFFFFF)
                    tx_xoffset[0] = 0;
                if(tx_yoffset[0] == 0x7FFFFFFF)
                    tx_yoffset[0] = 0;
            } else {
                if(tx_xoffset[i] == 0x7FFFFFFF)
                    tx_xoffset[i] = tx_xoffset[i-1];
                if(tx_yoffset[i] == 0x7FFFFFFF)
                    tx_yoffset[i] = tx_yoffset[i-1];
            }
            tx_pi_audio_file[i]   = GetParm.toString(this,
                                    "TxAudio"+(i+1), null);
            if(tx_pi_audio_file[i] != null)
                tx_pi_clip[i] = getAudioClip(getDocumentBase(),
                                    tx_pi_audio_file[i]);
        }

        img_lastimage_index = -1;
        for(i=0;i<img_num_images;i++) {

            // Get X and X offsets for Images
            img_xoffset[i] = GetParm.toInt(this,"ImgXOffset"+(i+1),0x7FFFFFFF);
            img_yoffset[i] = GetParm.toInt(this,"ImgYOffset"+(i+1),0x7FFFFFFF);

            if(i == 0) {
                if(img_xoffset[0] == 0x7FFFFFFF)
                    img_xoffset[0] = 0;
                if(img_yoffset[0] == 0x7FFFFFFF)
                    img_yoffset[0] = 0;
            } else {
                if(img_xoffset[i] == 0x7FFFFFFF)
                    img_xoffset[i] = img_xoffset[i-1];
                if(img_yoffset[i] == 0x7FFFFFFF)
                    img_yoffset[i] = img_yoffset[i-1];
            }

            // See if the user specified an image file
            img_file[i] = GetParm.toString(this,  "Image"+(i+1), null);

            // Get the Image filter
            img_filter_str[i] = GetParm.toString(this,
                                    "ImgFilter"+(i+1),
                                    "noop");
            img_pi_audio_file[i]   = GetParm.toString(this,
                                     "ImgAudio"+(i+1), null);
            if(img_pi_audio_file[i] != null) {
                img_pi_clip[i] = getAudioClip(getDocumentBase(),
                                    img_pi_audio_file[i]);
            }
        }
        debug("exiting init");
    }

    public void doTxFilters() {
        int i;
        Filter f;

        f = new Filter();
        for(i=0;i<tx_num_images;i++) {
            // This may get called because the thread was stopped but
            // is now being restarted.  No need to go through all of this
            // again
            if(tx_osi[i] != null)
                continue;
            getAppletContext().showStatus("Getting Text "
                                        + (i+1) + " of " + tx_num_images + ": "
                                        + tx_filter_str[i]);
            tx_filtered_pixels[i] = doStandardTxFilters(tx_pixels,
                            tx_width[i], tx_height[i],
                            f, tx_filter_str[i], i);
            tx_width[i] = f.getWidth();
            tx_height[i] = f.getHeight();
            tx_osi[i] = createImage(new MemoryImageSource(tx_width[i],
                                tx_height[i],
                                ColorModel.getRGBdefault(),
                                tx_filtered_pixels[i], 0, tx_width[i]));
            Thread.yield();
        }
        getAppletContext().showStatus("");
    }

    public synchronized void loadImages() {
        int i;

        // Now we have all the parameters
        for(i=0;i<img_num_images;i++) {

            // If the user didn't specify an image file then
            // get then use the last "gotten" image
            if(img_file[i] == null) {
                if(i == 0) {
                    // Error - you must have an image for first image
                    System.err.println("missing Image1");
                    img_num_images = -1;
                    continue;
                }
                img_nofilespec[i] = true;
            } else {
                // If the first character of the imf_file is $ then
                // get the previous image and reuse it
                if(img_file[i].charAt(0) == '$') {
                    char chars[];
                    int reuseval;
                    int len = img_file[i].length();;
                    // Get rid of leading $
                    chars =  new char [len];


                    img_file[i].getChars(0, len, chars, 0);
                    
                    reuseval = Integer.parseInt(new String(chars, 1, len-1));
                    reuseval -= 1;
                    img_reuseval[i] = reuseval;
                    img_copyfrom[i] = true;
                    debug("setting copyfrom to true for " + i);
                } else {
                    // If the user specified an image file then get it
                    img_gettr     = new ImgGetr(this);
                    img[i] = img_gettr.process(getDocumentBase(),
                                            img_file[i], false);

                    debug("prepareImage["+i+"]\n");

                    // It's essential to check the return value of
                    // prepareImage, because if the image is already
                    // prepared then imageUpdate will never be called
                    // and that is the only other place the img_loaded
                    // gets set
                    img_loaded[i] = prepareImage(img[i], this);
                    debug("RETURN from prepareImage["+i+"], img_loaded["+i+"] = " + img_loaded[i]);
                }
            }
        }
    }

    public synchronized boolean imageUpdate(Image image, int flags,
                                int x, int y, int w, int h) {
        int i;
        boolean retval = super.imageUpdate(image, flags, x, y, w, h);
        if((flags & ALLBITS) != 0) {
            for(i=0;i<img_num_images;i++) {
                if(image == img[i]) {
                    img_loaded[i] = true;
                    getAppletContext().showStatus("Loaded Image "
                            + (i+1) + " of " + img_num_images);
                    return retval;
                }
            }
        }
        return retval;
    }



    public int[] doStandardTxFilters(int[] pixels, int width, int height, 
                                    Filter f, String filter_str, int i) {

        StringTokenizer st;
        String first_token;
        int j;
        int num_elements;

        //========= INVOKE TEXT FILTERS ==============
        if(filter_str != null) {
            setFont(tx_font);
            if(tx_color != null)
                setForeground(tx_color);
            else
                tx_color = Color.black;
            setForeground(tx_color);

            // Before we start the filters, see if the user specified
            // $N as the beginning filter which means to reuse a
            // previous filter output
            //
            st = new StringTokenizer(filter_str, "|");
            num_elements = st.countTokens();
            if(num_elements > 0) {
                String newfilter_str = null;

                first_token = st.nextToken();
                for(j=0;j<num_elements-1;j++) {
                    if(j==0)
                        newfilter_str = st.nextToken();
                    else
                        newfilter_str += st.nextToken();
                    if(j < num_elements-2)
                        newfilter_str += "|";
                }

                if(first_token.charAt(0) == '$') {
                    char chars[];
                    int reuseval;
                    int len = first_token.length();;
                    // Get rid of leading $
                    chars =  new char [len];

                    first_token.getChars(0, len, chars, 0);
                    
                    reuseval = Integer.parseInt(new String(chars, 1, len-1));
                    reuseval -= 1;
                    tx_width[i] = tx_width[reuseval];
                    tx_height[i] = tx_height[reuseval];
                    if(tx_bg_color == null) {
                        pixels = f.invoke(this, null, newfilter_str,
                                    tx_filtered_pixels[reuseval],
                                    tx_width[i], tx_height[i],
                                    true, app_bg_color, tx_underline);
                    } else {
                        pixels = f.invoke(this, null, newfilter_str,
                                    tx_filtered_pixels[reuseval],
                                    tx_width[i], tx_height[i],
                                    false, tx_bg_color, tx_underline);
                    }
                    tx_width[i]  = f.getWidth();
                    tx_height[i] = f.getHeight();
                    return pixels;
                }
            }
            if(tx_bg_color == null) {
                pixels = f.invoke(this, tx, filter_str,
                            pixels, tx_width[i], tx_height[i],
                            true, app_bg_color, tx_underline);
            } else {
                pixels = f.invoke(this, tx, filter_str,
                            pixels, tx_width[i], tx_height[i],
                            false, tx_bg_color, tx_underline);
            }
            tx_width[i]  = f.getWidth();
            tx_height[i] = f.getHeight();
        }
        return pixels;
    }

    public int[] doStandardImgFilters(int[] pixels, int width, int height, 
                                    Filter f, String filter_str, int i) {
        String first_token;
        int j;
        StringTokenizer st;
        int num_elements;

        //========= INVOKE IMAGE FILTERS ==============
        if(filter_str != null) {

            // Before we start the filters, see if the user specified
            // $N as the beginning filter which means to reuse a
            // previous filter output
            //
            st = new StringTokenizer(filter_str, "|");
            num_elements = st.countTokens();
            if(num_elements > 0) {
                String newfilter_str = null;

                first_token = st.nextToken();
                for(j=0;j<num_elements-1;j++) {
                    if(j==0)
                        newfilter_str = st.nextToken();
                    else
                        newfilter_str += st.nextToken();
                    if(j < num_elements-2)
                        newfilter_str += "|";
                }

                if(first_token.charAt(0) == '$') {
                    char chars[];
                    int reuseval;
                    int len = first_token.length();;
                    // Get rid of leading $
                    chars =  new char [len];

                    first_token.getChars(0, len, chars, 0);
                    
                    reuseval = Integer.parseInt(new String(chars, 1, len-1));
                    reuseval -= 1;
                    img_width[i] = img_width[reuseval];
                    img_height[i] = img_height[reuseval];

                    // The reuse image may not have already gotten
                    // the pixels associated with it if it did
                    // not explicitly specify a filter.  However,
                    // the user may have specified $N to reuse
                    // an image (or filter) and now the pixels
                    // don't exist.  So, go get them!
                    if(img_filtered_pixels[reuseval] == null) {
                        img_filtered_pixels[reuseval] = 
                                    img_gettr.getPixels(img[reuseval], this);
                    }
                    pixels = f.invoke(this, null, newfilter_str,
                                    img_filtered_pixels[reuseval],
                                    img_width[i], img_height[i],
                                    false, app_bg_color, false);
                    img_width[i]  = f.getWidth();
                    img_height[i] = f.getHeight();
                    return pixels;
                }
            }
            pixels = f.invoke(this, null, filter_str,
                            pixels, width, height,
                            false, app_bg_color, false);
        }
        return pixels;
    }

    public void start() {
        if(tx_num_images > 0 && tx_animator == null) {
            tx_animator = new Thread(this, "Tx_thread");
            tx_animator.start();
        }
        if(img_num_images > 0 && img_animator == null) {
            img_animator = new Thread(this, "Img_thread");
            img_animator.start();
        }
    }

    public void stop() {
        if(tx_animator != null)
            tx_animator.stop();
        tx_animator = null;
        if(img_animator != null)
            img_animator.stop();
        img_animator = null;
    }

    // checkImgLoadStatus is called each time through the while
    // loop in the run method as long as all_images_loaded == false
    // If the image number is marked as "copyfrom" or "nofilespec"
    // then it is skipped. If the img_loaded[i] flag is true then
    // process any filters that may have been specified
    public boolean checkImgLoadStatus() {
        int i;
        boolean all_loaded = true;

        for(i=0;i<img_num_images;i++) {
            if(img_copyfrom[i] == true)
                continue;
            if(img_nofilespec[i] == true)
                continue;
            debug("img_loaded["+i+"] = " + img_loaded[i]);
            if(img_loaded[i] == true) { // flag says it's loaded
                debug("checkImgLoadStatus: img_osi["+i+"] = " + img_osi[i]);
                if(img_osi[i] == null) {        // No osi image yet
                    // We found one that must have just gotten loaded
                    img_width_orig[i] = img_gettr.getWidth(img[i], this);
                    img_height_orig[i] = img_gettr.getHeight(img[i], this);

                    // We save a lot of time by avoid the getPixels
                    // step if there is no filtering going on.  The
                    // problem with this is that it's possible that
                    // a subsequent filter will make reference to
                    // $N (where N is the current i) and the pixels
                    // aren't there.  So, we'll have to deals with
                    // getting the pixels on the fly if they're
                    // called for (see doStandardImgFilters)
                    if(img_filter_str[i].equals("noop")) {
                        img_width[i] = img_width_orig[i];
                        img_height[i] = img_height_orig[i];
                        img_osi[i] = img[i];
                        continue;
                    }
                    img_pixels[i] = img_gettr.getPixels(img[i], this);
                    img_filtered_pixels[i] = doStandardImgFilters(
                                    img_pixels[i],
                                    img_width_orig[i],
                                    img_height_orig[i],
                                    f, img_filter_str[i], i);
                    img_width[i] = f.getWidth();
                    img_height[i]= f.getHeight();
                    img_osi[i] = createImage(new MemoryImageSource(
                                        img_width[i],
                                        img_height[i],
                                        ColorModel.getRGBdefault(),
                                        img_filtered_pixels[i],
                                        0, img_width[i]));
                    return false;
                }
            } else {
                debug("setting false on " + i);
                all_loaded = false;
            }
        }
        if(all_loaded == true) {
            debug("checkImgLoadStatus: ALL LOADED!!!!!");
            getAppletContext().showStatus("");
        }
        return all_loaded;
    }

    //
    // updateImageInfo gets called after all the images have been loaded
    // with getImage.  Now we go through the list of image and see if
    // there were any Image{N} parameters that did not include a
    // file specification (img_nofilespec) and if there were any
    // that were give as "$N" which means to reuse image N
    // In either case, get the appropriate pixels, widht, height,
    // and then process them with the current filter (if any) and
    // create the new image.
    //
    public void updateImageInfo() {
        int i;
        int last_img_file = 0;

        //
        // ===== First, check for images without a image specification
        //
        for(i=0;i<img_num_images;i++) {
            if(img_nofilespec[i] == true) {
                debug("nofilespec is true for " + i + 
                            " last_img_file = " + last_img_file);
                //
                // If an image specification was not given then use
                // the last one that was given
                getAppletContext().showStatus("Getting Image "
                                + (i+1) + " of " + img_num_images + ": "
                                + img_filter_str[i]);
                img[i] = img[img_reuseval[i]];
                img_width[i]  = img_width[last_img_file];
                img_height[i] = img_height[last_img_file];
                img_width_orig[i] = img_width_orig[last_img_file];
                img_height_orig[i] = img_height_orig[last_img_file];

                // If no filters are specified then don't waste time
                // creating the pixels
                if(img_filter_str[i].equals("noop")) {
                    img_osi[i] = img[i];
                    continue;
                } else {
                    img_pixels[i] = img_gettr.getPixels(img[i], this);



                    img_filtered_pixels[i] = doStandardImgFilters(
                                img_pixels[i],
                                img_width_orig[i],
                                img_height_orig[i],
                                f, img_filter_str[i], i);
                    img_width[i] = f.getWidth();
                    img_height[i] = f.getHeight();
                    img_osi[i] = createImage(new MemoryImageSource(
                                        img_width[i],
                                        img_height[i],
                                        ColorModel.getRGBdefault(),
                                        img_filtered_pixels[i],
                                        0, img_width[i]));
                    Thread.yield();
                }
            } else
                last_img_file = i;

            //
            // Second, check for images with a $N image specification
            //
            // img_nofilespec and img_copyfrom should never both be true
            //
            if(img_copyfrom[i] == true) {
                //
                // The img_reuseval array contains the index of the
                // image that we should copy
                //
                img[i] = img[img_reuseval[i]];
                img_width[i] = img_width[img_reuseval[i]];
                img_height[i] = img_height[img_reuseval[i]];
                img_width_orig[i] = img_width_orig[img_reuseval[i]];
                img_height_orig[i] = img_height_orig[img_reuseval[i]];

                // If no filters are specified then don't waste time
                // creating the pixels
                if(img_filter_str[i].equals("noop")) {
                    img_osi[i] = img[i];
                    continue;
                } else {
                    img_pixels[i] = img_gettr.getPixels(img[i], this);
                    img_filtered_pixels[i] = doStandardImgFilters(
                                    img_pixels[i],
                                    img_width_orig[i],
                                    img_height_orig[i],
                                    f, img_filter_str[i], i);
                    img_width[i] = f.getWidth();
                    img_height[i] = f.getHeight();
                    img_osi[i] = createImage(new MemoryImageSource(
                                            img_width[i],
                                            img_height[i],
                                            ColorModel.getRGBdefault(),
                                            img_filtered_pixels[i],
                                            0, img_width[i]));
                }
            }
        }
    }

    public int getNextTxIndex(int current, int max, int min) {
        current++;
        if(current >= max)
            current = min;
        return current;
    }

    public int getNextImgIndex(int current, int max, int min) {
        current++;
        if(current >= max)
            current = min;
        return current;
    }
    
    public void tx_sleep(int s_runs, int s_images,
                        int current, int max, int min) {
        try {
            if(current+1 == max)
                Thread.sleep(s_runs);
            else
                Thread.sleep(s_images);
        } catch (InterruptedException e) {}
    }

    public void img_sleep(int s_runs, int s_images,
                        int current, int max, int min) {
        tx_sleep(s_runs, s_images, current, max, min);
    }

    public void run() {
        Thread current_thread = Thread.currentThread();;

        //
        // It's important to set the priorities on these
        // threads to some value which is smaller than the
        // priority of the thread that is called to do paint
        Thread.currentThread().setPriority(Thread.MIN_PRIORITY);

        tx_index = -1;
        img_index = -1;

        debug("entering run");
        if(current_thread == img_animator) {
            loadImages();
        }

        if(current_thread == tx_animator) {
            doTxFilters();
        }

        while(true) {
            if(Thread.currentThread() == tx_animator) {
                debug("In tx_animator thread");
                // ========== Text ===============
                if(tx_index == -1)      // First time
                    tx_index = tx_lower_bound;
                else
                    tx_index = getNextTxIndex(tx_index, tx_num_images,
                                            tx_lower_bound);
                if(tx_pi_clip[tx_index] != null)
                    tx_pi_clip[tx_index].play();
                repaint();
                tx_sleep(tx_delay_runs, tx_delay_images,
                                    tx_index, tx_num_images,
                                    tx_lower_bound);
            } else if(Thread.currentThread() == img_animator) {
                debug("In img_animator thread");
                // ========== Image ===============
                if(all_images_loaded == false) {
                    all_images_loaded = checkImgLoadStatus();
                    if(all_images_loaded == true) 
                        updateImageInfo();
                }
                if(all_images_loaded == true) {
                    if(img_index == -1) // First time
                        img_index = img_lower_bound;
                    else
                        img_index = getNextImgIndex(img_index, img_num_images,
                                            img_lower_bound);
                    if(img_pi_clip[img_index] != null)
                        img_pi_clip[img_index].play();
                    repaint();
                    img_sleep(img_delay_runs, img_delay_images,
                                    img_index, img_num_images,
                                    img_lower_bound);
                } else {
                    img_index = 0;
                    if(img_display_first) {
                        repaint();
                    } else
                        img_index = -1;
                    try {
                        Thread.sleep(1000);
                    } catch (InterruptedException e) {
                        break;
                    }
                }
            }
        }
    }

    public void paint(Graphics g) {

        // If this is the first time we've been in paint
        // or the applet has changed size then create
        // an image the size of the applet into which
        // we write the background image
        if(applet_width != size().width ||
           applet_height!= size().height) {
            applet_width  = size().width;
            applet_height = size().height;
            offScreen = createImage(applet_width, applet_height);
            offGC = offScreen.getGraphics();
        }
        offGC.setColor(getBackground());
        offGC.fillRect(0, 0, applet_width, applet_height);
        tilebackground(offGC);

        if(img_index >= 0) {    // No images specified
            if(img_osi[img_index] == null) {
                // do nothing
            } else if(img_index < img_num_images && img_index >= 0) {

                img_x_coord[img_index] = calcULXcoord(img_horizcenter,
                                size().width,
                                img_width[img_index],
                                img_xoffset[img_index]);
                img_y_coord[img_index] = calcULYcoord(img_vertcenter,
                                size().height,
                                img_height[img_index],
                                img_yoffset[img_index]);

                debug("drawing Image: " + img_index);
                offGC.drawImage(img_osi[img_index], img_x_coord[img_index],
                                img_y_coord[img_index], this);
                drawFrame(offGC, img_x_coord[img_index],
                                img_y_coord[img_index], img_width[img_index],
                                img_height[img_index], img_fr_type,
                                img_fr_thick,
                                img_fr_margin);
                drawBorder(offGC, img_x_coord[img_index],
                                img_y_coord[img_index],
                                img_width[img_index], img_height[img_index],
                                img_borderwidth,
                                img_bordermargin,
                                img_bordercolor);
            }
        }

        if(tx_index >= 0) {
            // do nothing
            if(tx_osi[tx_index] == null) {
                // do nothing
            } else if(tx_index < tx_num_images && tx_index >= 0) {

                tx_x_coord[tx_index] = calcULXcoord(tx_horizcenter,
                                size().width,
                                tx_width[tx_index],
                                tx_xoffset[tx_index]);
                tx_y_coord[tx_index] = calcULYcoord(tx_vertcenter,
                                size().height,
                                tx_height[tx_index],
                                tx_yoffset[tx_index]);

                debug("drawing Text: " + tx_index);
                offGC.drawImage(tx_osi[tx_index], tx_x_coord[tx_index],
                                tx_y_coord[tx_index], this);
                // Draw frame for text
                drawFrame(offGC, tx_x_coord[tx_index], tx_y_coord[tx_index],
                                    tx_width[tx_index], tx_height[tx_index],
                                    tx_fr_type, tx_fr_thick,
                                    tx_fr_margin);


                // Draw border for text
                drawBorder(offGC, tx_x_coord[tx_index], tx_y_coord[tx_index],
                                    tx_width[tx_index], tx_height[tx_index],
                                    tx_borderwidth, tx_bordermargin,
                                    tx_bordercolor);
            }
        }
        // Draw frame for applet
        drawFrame(offGC, 0, 0, size().width, size().height,
                            app_fr_type, app_fr_thick, 0);
        // Draw border for applet
        drawBorder(offGC, 0, 0, size().width, size().height,
                            app_borderwidth, 0, app_bordercolor);

        g.drawImage(offScreen, 0, 0, this);
    }

    public void debug(String s) {
        if(debugflag)
            System.err.println(s);
    }

    public boolean keyDown(Event evt, int key) {
        int x = evt.x;
        int y = evt.y;
        // Handle the text next (Text takes precedence over images)
        if(tx_index != -1) {
            if(x > tx_x_coord[tx_index] && 
               x < tx_x_coord[tx_index] + tx_width[tx_index] &&
               y > tx_y_coord[tx_index] && 
               y < tx_y_coord[tx_index] + tx_height[tx_index]) {
                if(key == 'f') {
                    tx_delay_images *= .9;
                    if(tx_delay_images < 2)
                        tx_delay_images = 2;
                    getAppletContext().showStatus("Setting image delay to: " +
                                tx_delay_images);
                } else if(key == 's') {
                    if(tx_delay_images < 10)
                        tx_delay_images++;
                    else
                        tx_delay_images *= 1.1;
                    getAppletContext().showStatus("Setting image delay to: " +
                                tx_delay_images);
                } else if(key == 'F') {
                    tx_delay_runs *= .9;
                    if(tx_delay_runs < 2)
                        tx_delay_runs = 2;
                    getAppletContext().showStatus("Setting run delay to: " +
                                tx_delay_runs);
                } else if(key == 'S') {
                    if(tx_delay_runs < 10)
                        tx_delay_runs++;
                    else
                        tx_delay_runs *= 1.1;
                    getAppletContext().showStatus("Setting run delay to: " +
                                tx_delay_runs);
                }
            }    
        }

        if(img_index != -1) {
            if(x > img_x_coord[img_index] && 
               x < img_x_coord[img_index] + img_width[img_index] &&
               y > img_y_coord[img_index] && 
               y < img_y_coord[img_index] + img_height[img_index]) {
                if(key == 'f') {
                    img_delay_images *= .9;
                    if(img_delay_images < 2)
                        img_delay_images = 2;
                    getAppletContext().showStatus("Setting image delay to: " +
                                img_delay_images);
                } else if(key == 's')
                    if(img_delay_images < 10)
                        img_delay_images++;
                    else
                        img_delay_images *= 1.1;
                    getAppletContext().showStatus("Setting image delay to: " +
                                img_delay_images);
                if(key == 'F') {
                    img_delay_runs *= .9;
                    if(img_delay_runs < 2)
                        img_delay_runs = 2;
                    getAppletContext().showStatus("Setting run delay to: " +
                                img_delay_runs);
                } else if(key == 'S') {
                    if(img_delay_runs < 10)
                        img_delay_runs++;
                    else
                        img_delay_runs *= 1.1;
                    getAppletContext().showStatus("Setting run delay to: " +
                                img_delay_runs);
                }
            }    
        }

        return true;
    }

    //
    // If the user clicks the mouse then load the URL
    // We check the text position first, so if there is
    // overlapping text and image then the text URL will
    // be loaded, not the imgae URL
    //
    public boolean mouseDown(Event evt, int x, int y) {
        int i;

        // Handle the image maps first
        for(i=0;i<num_maps;i++) {
            if(x > map_x1[i] && x < map_x2[i] &&
               y > map_y1[i] && y < map_y2[i]) {
                if(map_url[i] != null) {
                    if(testmode) {
                        getAppletContext().showStatus("Loading URL: " +
                                    map_url[i].toString());
                    } else
                        getAppletContext().showDocument(map_url[i]);
                    return true;
                }
            }
        }    

        // Handle the text next (Text takes precedence over images)
        if(tx_index != -1) {
            if(x > tx_x_coord[tx_index] && 
               x < tx_x_coord[tx_index] + tx_width[tx_index] &&
               y > tx_y_coord[tx_index] && 
               y < tx_y_coord[tx_index] + tx_height[tx_index]) {
                if(tx_url != null) {
                    getAppletContext().showDocument(tx_url);
                    return true;
                }
            }    
        }

        // Handle the image last
        if(img_index != -1) {
            if(x > img_x_coord[img_index] &&
               x < img_x_coord[img_index] + img_width[img_index] &&
               y > img_y_coord[img_index] &&
               y < img_y_coord[img_index] + img_height[img_index]) {
                if(img_url != null) {
                    getAppletContext().showDocument(img_url);
                    return true;
                }
            }    
        }
        return true;
    }

    //
    // Check the x and y coordinates to see if we're actually
    // inside the boundaries of the text or image or both
    //
    public boolean mouseMove(Event evt, int x, int y) {
        boolean tx_status_shown = false;
        boolean img_status_shown = false;
        int i;

        if(testmode) {
            getAppletContext().showStatus("X: " + x + " Y: " + y);
            return true;
        }

        //
        // If a image map url was found (hence tx_status_shown == true)
        // then return and don't bother to call the other mouseMove
        // stuff would simply display the url for the text or image.
        //
        if(tx_status_shown == true) {
            return true;
        }

        // ========== AUDIO =============
        // Process the image first
        if(img_index != -1) {
            if(x > img_x_coord[img_index] &&
               x < img_x_coord[img_index] + img_width[img_index] &&
               y > img_y_coord[img_index] &&
               y < img_y_coord[img_index] + img_height[img_index]) {
                if(!img_playing) {
                    img_playing = true;
                    if(img_clip != null)
                        img_clip.play();
                } // else it's already playing, leave it alone
            } else {
                // Since we don't get any kind of notification when the audio
                // clip has terminated (due to playing to completion) we set
                // playing to false everytime we move out of the boundaries
                // of the string
                img_playing = false;
                if(img_clip != null) {
                    img_clip.stop();
                }
            }
        }

        // Process the text last
        if(tx_index != -1) {
            if(x > tx_x_coord[tx_index] &&
               x < tx_x_coord[tx_index] + tx_width[tx_index] &&
               y > tx_y_coord[tx_index] &&
               y < tx_y_coord[tx_index] + tx_height[tx_index]) {
                if(!tx_playing) {
                    tx_playing = true;
                    if(tx_clip != null)
                        tx_clip.play();
                } // else it's already playing, leave it alone
            } else {
                // Since we don't get any kind of notification when the audio
                // clip has terminated (due to playing to completion) we set
                // playing to false everytime we move out of the boundaries
                // of the string
                tx_playing = false;
                if(tx_clip != null) {
                    tx_clip.stop();
                }
            }
        }

        // ========== IMAGE MAPS =============
        for(i=0;i<num_maps;i++) {
            if(x > map_x1[i] && x < map_x2[i] &&
               y > map_y1[i] && y < map_y2[i]) {
                if(map_url[i] != null) {
                    getAppletContext().showStatus(map_url[i].toString());
                    tx_status_shown = true;
                }
            }
        }

        // ========== URLs =============
        // Handle the text first
        if(tx_index != -1) {
            if(x > tx_x_coord[tx_index] &&
               x < tx_x_coord[tx_index] + tx_width[tx_index] &&
               y > tx_y_coord[tx_index] &&
               y < tx_y_coord[tx_index] + tx_height[tx_index]) {
                if(tx_url != null) {
                    getAppletContext().showStatus(tx_url.toString());
                    tx_status_shown = true;
                }
            }    
        }

        // Handle the image last
        // If a Text URL is being displayed then don't
        // display the image URL
        if(img_index != -1) {
            if(x > img_x_coord[img_index] &&
               x < img_x_coord[img_index] + img_width[img_index] &&
               y > img_y_coord[img_index] &&
               y < img_y_coord[img_index] + img_height[img_index]) {
                if(img_url != null && tx_status_shown == false) {
                    getAppletContext().showStatus(img_url.toString());
                    img_status_shown = true;
                }
            }    
        }
        // If we are not over either text or image that has 
        // a corresponding URL then clear the status line
        if(tx_status_shown == false && img_status_shown == false)
            getAppletContext().showStatus(null);
        return true;
    }

    //
    // Just to be sure, when we exit the applet, stop the clips
    // It should already be stopped
    //
    public boolean mouseExit(Event evt, int x, int y) {
        // Just to be sure
        tx_playing = false;
        if(tx_clip != null)
            tx_clip.stop();
        img_playing = false;
        if(img_clip != null)
            img_clip.stop();
        getAppletContext().showStatus(null);
        return true;
    }
}
