import java.awt.*;
import java.awt.image.*;
import java.applet.*;
import java.net.*;
import java.io.*;
import java.util.*;

//
// BasicTxt is the base class applet upon which most of the
// applets in the Fundamentals chapter are based.
// Basic handles lots of generic things like drawing
// borders, frames, and underline.  Many of the basic
// parameters are gotten here.
//
public class AnimateMultiple extends Applet implements Runnable {
    protected int       tx_type[];
    protected int       img_type[];
    protected String    tx_string_type[];
    protected String    img_string_type[];
    protected boolean   tx_goingup[];
    protected boolean   img_goingup[];
    static final int    Continuous  = 0;
    static final int    TwoWay = 1;
    static final int    OnButton  = 2;
    static final int    OnButtonTwoWay = 3;
    static final int    OnEntry  = 4;
    static final int    OnEntryTwoWay = 5;

    protected Thread    tx_animator[];
    protected Thread    img_animator[];

    // Parameters that apply to the entire applet
    protected int         app_borderwidth;      // Applet border width
    protected Color       app_bordercolor;      // Applet border color
    protected String      app_fr_type;          // Applet frame type
    protected int         app_fr_thick;         // Applet frame thickness
    protected Color       app_bg_color;         // Applet background color
    protected int         tx_count;             // Number of Text strings
    protected int         img_count;            // Number of Images

    // Parameters specific to Text
    protected String      tx[];                   // Text
    protected int         tx_xoffset[][];         // Text X offset
    protected int         tx_yoffset[][];         // Text Y offset
    protected Color       tx_color[];             // Text color
    protected Color       tx_bg_color[];          // Text background color
    protected boolean     tx_horizcenter[];       // Text horizontally centered
    protected boolean     tx_vertcenter[];        // Text vertically centered
    protected int         tx_xcenter[];           // Text centered at
    protected int         tx_ycenter[];           // Text centered at
    protected Font        tx_font[];              // Text font
    protected FontMetrics tx_fontmetrics[];       // Text font metrics
    protected boolean     tx_underline[];         // Text underline
    protected int         tx_width[][];           // Text width
    protected int         tx_height[][];          // Text height
    protected int         tx_ascent[];            // Text ascent
    protected int         tx_descent[];           // Text descent
    protected int         tx_borderwidth[];       // Text border width
    protected Color       tx_bordercolor[];       // Text border color
    protected int         tx_bordermargin[];      // Text border margin
    protected String      tx_fr_type[];           // Text frame type
    protected int         tx_fr_thick[];          // Text frame thickness
    protected int         tx_fr_margin[];         // Text frame margin
    protected int         tx_x_coord[][];         // Text X Coordinate
    protected int         tx_y_coord[][];         // Text Y Coordinate
    protected int         tx_num_images[];        // Text Number of Filters
    protected int         tx_delay_images[];
    protected int         tx_delay_runs[];
    protected int         tx_lower_bound[];
    protected boolean     within_tx[];
    protected boolean     last_within_tx[];
    protected boolean     tx_return_start[];
    protected boolean     tx_firsttime[];
    protected boolean     tx_threadSuspended[];    // Used only for Continuous
    protected AudioClip   tx_pi_clip[][];          // The per text Audio Clip
    protected String      tx_pi_audio_file[][];    // The per text audio file
    protected Filter      txFilter[];

    // Parameters specific to the Image
    protected Image       img[][];                // The image
    protected String      img_file[][];           // Image file name
    protected ImgGetr     img_gettr;              // The Image Getter class
    protected int         img_width[][];          // Image width
    protected int         img_height[][];         // Image height
    protected int         img_xoffset[][];        // Image X offset
    protected int         img_yoffset[][];        // Image Y offset
    protected int         img_y_coord[][];        // Image X coordinate
    protected int         img_x_coord[][];        // Image Y coordinate
    protected boolean     img_horizcenter[];      // Image horizontally centered
    protected boolean     img_vertcenter[];       // Image vertically centered
    protected int         img_xcenter[];          // Image centered at
    protected int         img_ycenter[];          // Image centered at
    protected int         img_borderwidth[];      // Image border width
    protected Color       img_bordercolor[];      // Image border color
    protected int         img_bordermargin[];     // Image border margin
    protected String      img_fr_type[];          // Image frame type
    protected int         img_fr_thick[];         // Image frame thickness
    protected int         img_fr_margin[];        // Image frame margin
    protected int         img_num_images[];       // Text Number of Filters
    protected int         img_reuseval[][];       
    protected boolean     img_nofilespec[][];       
    protected boolean     img_display_first[];       
    protected boolean     img_copyfrom[][];       
    protected boolean     img_loaded[][];
    protected int         img_width_orig[][];        // Image width
    protected int         img_height_orig[][];       // Image height
    protected int         img_delay_images[];
    protected int         img_delay_runs[];
    protected int         img_lower_bound[];
    protected boolean     within_img[];
    protected boolean     last_within_img[];
    protected boolean     img_return_start[];
    protected boolean     img_firsttime[];
    protected boolean     img_threadSuspended[];    // Used only for 
    protected AudioClip   img_pi_clip[][];          // The per image Audio Clip
    protected String      img_pi_audio_file[][];    // The per image audio file
    Font                  loading_font;
    protected Filter      imgFilter[];

    // Parameters specific to Audio
    protected AudioClip    tx_clip[];             // The Text Audio Clip
    protected String       tx_audio_file[];       // The Text audio file
    protected boolean      tx_playing[];          // Text clip currently playing
    protected AudioClip    img_clip[];            // The Image Audio Clip
    protected String       img_audio_file[];      // The Image audio file
    protected boolean      img_playing[];         // Image clip currently playing

    // Parameters specific to URLs
    protected URL          tx_url[];              // The Text URL
    protected URL          img_url[];             // The Image URL

    protected boolean      tile;             // Tiled?
    protected Image        bgimage;          // The background image
    protected String       bgimage_file;     // Name of the background image
    protected ImgGetr      ig;               // The Image Getter class
    protected int          bgimage_width;    // Background image width
    protected int          bgimage_height;   // Background image height
    protected int          bgimageXoffset;   // BG image X ofset
    protected int          bgimageYoffset;   // BG image Y ofset
    protected Image        tiled_bgimage;    // Tile BG?
    protected int          app_width;        // App width
    protected int          app_height;       // App height
    protected Graphics     bg_g;             // Background Graphics

    // Parameters specific to Image Maps
    protected int          num_maps;         // Number of Maps
    protected int          map_x1[];         // X1 coord for map
    protected int          map_y1[];         // Y1 coord for map
    protected int          map_x2[];         // X2 coord for map
    protected int          map_y2[];         // Y2 coord for map
    protected URL          map_url[];        // URLs for maps
    protected boolean      testmode;         // Test mode

    // Parameters specific to OffScreen images
    protected   Image         tx_osi[][];         // Text Off-Screen Image
    protected   Graphics      tx_g[][];           // Graphics of tx_osi
    protected   int           tx_pixels[][];      // The text image
    protected   int           tx_filtered_pixels[][][];      // The text image
    protected   String        tx_filter_str[][];  // Text filters in string format
    protected   Image         img_osi[][];        // Image Off-Screen Image
    protected   Graphics      img_g[][];          // Graphics of img_osi
    protected   PixGrab       img_pg[][];         // Image PixelGrabber
    protected   int           img_pixels[][][];     // The text image
    protected   int           img_filtered_pixels[][][];      // The text image
    protected   String        img_filter_str[][]; // Img filters in string format
    boolean                   retval;           // Retval from grabPixels;

    protected   Color         default_foreground;
    protected boolean debugflag = false;
    protected int tx_index[];
    protected int img_index[];
    protected boolean all_images_loaded[];
    int applet_width = -1;
    int applet_height = -1;
    protected Image     offScreen;
    protected Graphics  offGC;


    public void init() {
        int i, j;

        default_foreground = getForeground();

        // ======== Applet ===========
        app_borderwidth = GetParm.toInt(this,     "AppBorderWidth", 0);
        app_bordercolor = GetParm.toColor(this,   "AppBorderColor", Color.black);
        app_fr_type     = GetParm.toString(this,  "AppFrameType", "ShadowIn");
        app_fr_thick    = GetParm.toInt(this,     "AppFrameThickness", 0);
        app_bg_color    = GetParm.toColor(this,   "AppBGColor", null);
        tx_count        = GetParm.toInt(this,     "TxCount", 0);
        img_count       = GetParm.toInt(this,     "ImgCount", 0);

        // Set the background and foreground colors, if specified
        // We must set these colors here before using them
        // in the next block of code
        if(app_bg_color != null)
            setBackground(app_bg_color);
        else
            app_bg_color = getBackground();

        // ======== Text ===========
        tx              = new String[tx_count];
        tx_xoffset      = new int[tx_count][];
        tx_yoffset      = new int[tx_count][];
        tx_color        = new Color[tx_count];
        tx_bg_color     = new Color[tx_count];
        tx_horizcenter  = new boolean[tx_count];
        tx_vertcenter   = new boolean[tx_count];
        tx_xcenter      = new int[tx_count];
        tx_ycenter      = new int[tx_count];
        tx_font         = new Font[tx_count];
        tx_underline    = new boolean[tx_count];
        tx_borderwidth  = new int[tx_count];
        tx_bordercolor  = new Color[tx_count];
        tx_bordermargin = new int[tx_count];
        tx_fr_thick     = new int[tx_count];
        tx_fr_type      = new String[tx_count];
        tx_fr_margin    = new int[tx_count];
        tx_fontmetrics  = new FontMetrics[tx_count];
        tx_width        = new int[tx_count][];
        tx_height       = new int[tx_count][];
        tx_ascent       = new int[tx_count];
        tx_descent      = new int[tx_count];
        tx_x_coord      = new int[tx_count][];
        tx_y_coord      = new int[tx_count][];
        tx_audio_file   = new String[tx_count];
        tx_clip         = new AudioClip[tx_count];
        tx_playing      = new boolean[tx_count];
        tx_url          = new URL[tx_count];
        tx_osi          = new Image[tx_count][];
        tx_g            = new Graphics[tx_count][];
        tx_filter_str   = new String[tx_count][];
        tx_pixels       = new int[tx_count][];
        tx_filtered_pixels = new int[tx_count][][];
        tx_num_images  = new int[tx_count];
        tx_animator     = new Thread[tx_count];
        tx_index        = new int[tx_count];
        tx_delay_runs   = new int[tx_count];
        tx_delay_images = new int[tx_count];
        tx_lower_bound  = new int[tx_count];
        tx_string_type  = new String[tx_count];
        tx_type         = new int[tx_count];
        tx_goingup      = new boolean[tx_count];
        within_tx       = new boolean[tx_count];
        last_within_tx  = new boolean[tx_count];
        tx_return_start = new boolean[tx_count];
        tx_firsttime    = new boolean[tx_count];
        tx_threadSuspended = new boolean[tx_count];
        tx_pi_clip      = new AudioClip[tx_count][];
        tx_pi_audio_file = new String[tx_count][];
	txFilter         = new Filter[tx_count];


        for(i=0;i<tx_count;i++) {
            tx[i]              = GetParm.toString(this,
                                    "Text"+(i+1), null);
            tx_color[i]        = GetParm.toColor(this,
                                    "Tx"+(i+1)+"Color", null);
            tx_bg_color[i]     = GetParm.toColor(this,
                                    "Tx"+(i+1)+"BGColor", null);
            tx_horizcenter[i]  = GetParm.toBoolean(this,
                                    "Tx"+(i+1)+"HorizCenter", false);
            tx_vertcenter[i]   = GetParm.toBoolean(this,
                                    "Tx"+(i+1)+"VertCenter",  false);
            tx_xcenter[i]      = GetParm.toInt(this,
                                    "Tx"+(i+1)+"XCenter", -1);
            tx_ycenter[i]      = GetParm.toInt(this,
                                    "Tx"+(i+1)+"YCenter",  -1);
            tx_font[i]         = GetParm.toFont(this,
                                    "Tx"+(i+1)+"Font",
                                     "Tx"+(i+1)+"Style",
                                     "Tx"+(i+1)+"PointSize");
            tx_underline[i]    = GetParm.toBoolean(this,
                                    "Tx"+(i+1)+"UnderLine", false);
            tx_borderwidth[i]  = GetParm.toInt(this,
                                    "Tx"+(i+1)+"BorderWidth", 0);
            tx_bordercolor[i]  = GetParm.toColor(this,
                                    "Tx"+(i+1)+"BorderColor", Color.black);
            tx_bordermargin[i] = GetParm.toInt(this,
                                    "Tx"+(i+1)+"BorderMargin", 0);
            tx_fr_thick[i]     = GetParm.toInt(this,
                                    "Tx"+(i+1)+"FrameThickness", 0);
            tx_fr_type[i]      = GetParm.toString(this,
                                    "Tx"+(i+1)+"FrameType", "ShadowIn");
            tx_fr_margin[i]    = GetParm.toInt(this,
                                    "Tx"+(i+1)+"FrameMargin", 0);
            tx_delay_images[i] = GetParm.toInt(this,
                                    "Tx"+(i+1)+"DelayBetweenImages", 100);
            tx_delay_runs[i]   = GetParm.toInt(this,
                                    "Tx"+(i+1)+"DelayBetweenRuns", 2000);
            tx_lower_bound[i]  = GetParm.toInt(this,
                                    "Tx"+(i+1)+"InitialImage", 0);
            if(tx_lower_bound[i] > 0)
                tx_lower_bound[i]--;

            if(tx_font[i] == null)
                System.out.println("tx_font == null");
            else 
                tx_fontmetrics[i] = getFontMetrics(tx_font[i]);

            if(tx_color[i] != null)
                setForeground(tx_color[i]);
            else
                tx_color[i] = getForeground();

            tx_string_type[i]  = GetParm.toString(this,
                                    "Tx"+(i+1)+"AnimationType", "Continuous");
            tx_type[i] = getAnimType(tx_string_type[i]);
            tx_return_start[i] = GetParm.toBoolean(this,
                                     "Tx"+(i+1)+"ReturnStart", false);
	    txFilter[i] = new Filter();




            tx_num_images[i]    = GetParm.toInt(this,
                                    "Tx"+(i+1)+"NumImages", 0);

            if(tx_num_images[i] > 0) {
                tx_width[i]      = new int[tx_num_images[i]];
                tx_height[i]     = new int[tx_num_images[i]];
                tx_xoffset[i]    = new int[tx_num_images[i]];
                tx_yoffset[i]    = new int[tx_num_images[i]];
                tx_x_coord[i]    = new int[tx_num_images[i]];
                tx_y_coord[i]    = new int[tx_num_images[i]];
                tx_filter_str[i] = new String[tx_num_images[i]];
                tx_osi[i]        = new Image[tx_num_images[i]];
                tx_filtered_pixels[i] = new int[tx_num_images[i]][];
                tx_pi_audio_file[i] = new String[tx_num_images[i]];
                tx_pi_clip[i]    = new AudioClip[tx_num_images[i]];
            }
                

            for(j=0;j<tx_num_images[i];j++) {
                tx_filter_str[i][j]  = GetParm.toString(this,
                                    "Tx"+(i+1)+"Filter"+(j+1),  null);
            
                tx_xoffset[i][j]  = GetParm.toInt(this,
                                        "Tx"+(i+1)+"XOffset"+(j+1), 0x7FFFFFFF);
                tx_yoffset[i][j]  = GetParm.toInt(this,
                                        "Tx"+(i+1)+"YOffset"+(j+1), 0x7FFFFFFF);
                if(j == 0) {
                    if(tx_xoffset[i][0] == 0x7FFFFFFF)
                        tx_xoffset[i][0] = 0;
                    if(tx_yoffset[i][0] == 0x7FFFFFFF)
                        tx_yoffset[i][0] = 0;
                } else {
                    if(tx_xoffset[i][j] == 0x7FFFFFFF)
                        tx_xoffset[i][j] = tx_xoffset[i][j-1];
                    if(tx_yoffset[i][j] == 0x7FFFFFFF)
                        tx_yoffset[i][j] = tx_yoffset[i][j-1];
                }
                // Get the audio per text
                tx_pi_audio_file[i][j]   = GetParm.toString(this,
                                        "Tx"+(i+1)+"Audio"+(j+1), null);
                if(tx_pi_audio_file[i][j] != null)
                    tx_pi_clip[i][j] = getAudioClip(getDocumentBase(),
                                    tx_pi_audio_file[i][j]);
            }


            tx_pixels[i] = null;
            tx_goingup[i] = true;
            tx_firsttime[i] = true;
        }

        // ======== Image ===========
        img             = new Image[img_count][];
        img_file        = new String[img_count][];
        img_xoffset     = new int[img_count][];
        img_yoffset     = new int[img_count][];
        img_horizcenter = new boolean[img_count];
        img_vertcenter  = new boolean[img_count];
        img_xcenter     = new int[img_count];
        img_ycenter     = new int[img_count];
        img_borderwidth = new int[img_count];
        img_bordercolor = new Color[img_count];
        img_bordermargin= new int[img_count];
        img_fr_thick    = new int[img_count];
        img_fr_type     = new String[img_count];
        img_fr_margin   = new int[img_count];
        img_pg          = new PixGrab[img_count][];
        img_width       = new int[img_count][];
        img_height      = new int[img_count][];
        img_x_coord     = new int[img_count][];
        img_y_coord     = new int[img_count][];
        img_clip        = new AudioClip[img_count];
        img_audio_file  = new String[img_count];
        img_playing     = new boolean[img_count];
        img_url         = new URL[img_count];
        img_osi         = new Image[img_count][];
        img_g           = new Graphics[img_count][];
        img_filter_str  = new String[img_count][];
        img_pixels      = new int[img_count][][];
        img_filtered_pixels = new int[img_count][][];
        img_num_images = new int[img_count];
        img_reuseval    = new int[img_count][];
        img_nofilespec  = new boolean[img_count][];
        img_copyfrom    = new boolean[img_count][];
        img_loaded      = new boolean[img_count][];
        img_display_first    = new boolean[img_count];
        img_loaded       = new boolean[img_count][];
        img_nofilespec   = new boolean[img_count][];
        img_copyfrom     = new boolean[img_count][];
        img_reuseval     = new int[img_count][];
        img_animator     = new Thread[img_count];
        img_width_orig   = new int[img_count][];
        img_height_orig  = new int[img_count][];
        img_index        = new int[img_count];
        img_delay_runs   = new int[img_count];
        img_delay_images = new int[img_count];
        img_lower_bound  = new int[img_count];
        all_images_loaded= new boolean[img_count];
        img_string_type  = new String[img_count];
        img_type         = new int[img_count];
        img_goingup      = new boolean[img_count];
        within_img       = new boolean[img_count];
        last_within_img  = new boolean[img_count];
        img_return_start = new boolean[img_count];
        img_firsttime    = new boolean[img_count];
        img_threadSuspended = new boolean[img_count];
        img_pi_clip      = new AudioClip[img_count][];
        img_pi_audio_file = new String[img_count][];
	imgFilter         = new Filter[img_count];


        img_gettr     = new ImgGetr(this);
        for(i=0;i<img_count;i++) {
            img_horizcenter[i] = GetParm.toBoolean(this,
                                    "Img"+(i+1)+"HorizCenter", false);
            img_vertcenter[i]  = GetParm.toBoolean(this,
                                    "Img"+(i+1)+"VertCenter",  false);
            img_xcenter[i]      = GetParm.toInt(this,
                                    "Img"+(i+1)+"XCenter", -1);
            img_ycenter[i]      = GetParm.toInt(this,
                                    "Img"+(i+1)+"YCenter",  -1);
            img_borderwidth[i] = GetParm.toInt(this, 
                                    "Img"+(i+1)+"BorderWidth", 0);
            img_bordercolor[i] = GetParm.toColor(this,
                                    "Img"+(i+1)+"BorderColor",
                                     Color.black);
            img_bordermargin[i]= GetParm.toInt(this,
                                    "Img"+(i+1)+"BorderMargin", 0);
            img_fr_thick[i]    = GetParm.toInt(this, 
                                    "Img"+(i+1)+"FrameThickness", 0);
            img_fr_type[i]     = GetParm.toString(this,
                                    "Img"+(i+1)+"FrameType", "ShadowIn");
            img_fr_margin[i]   = GetParm.toInt(this,
                                    "Img"+(i+1)+"FrameMargin", 0);
            img_num_images[i]    = GetParm.toInt(this,
                                    "Img"+(i+1)+"NumImages", 0);
            img_delay_images[i] = GetParm.toInt(this,
                                    "Img"+(i+1)+"DelayBetweenImages", 100);
            img_delay_runs[i]   = GetParm.toInt(this,
                                    "Img"+(i+1)+"DelayBetweenRuns", 2000);
            img_lower_bound[i]  = GetParm.toInt(this,
                                    "Img"+(i+1)+"InitialImage", 0);
            if(img_lower_bound[i] > 0)
                img_lower_bound[i]--;
	    img_display_first[i]= GetParm.toBoolean(this,
				    "Img"+(i+1)+"DisplayFirst", false);
            img_string_type[i]  = GetParm.toString(this,
                                    "Img"+(i+1)+"AnimationType", "Continuous");
            img_type[i] = getAnimType(img_string_type[i]);
            img_return_start[i] = GetParm.toBoolean(this,
                                     "Img"+(i+1)+"ReturnStart", false);
	    imgFilter[i] = new Filter();

            if(img_num_images[i] > 0) {
                img[i]            = new Image[img_num_images[i]];
                img_pixels[i]     = new int[img_num_images[i]][];
                img_file[i]       = new String[img_num_images[i]];
                img_width[i]      = new int[img_num_images[i]];
                img_height[i]     = new int[img_num_images[i]];
                img_xoffset[i]    = new int[img_num_images[i]];
                img_yoffset[i]    = new int[img_num_images[i]];
                img_x_coord[i]    = new int[img_num_images[i]];
                img_y_coord[i]    = new int[img_num_images[i]];
                img_filter_str[i] = new String[img_num_images[i]];
                img_osi[i]        = new Image[img_num_images[i]];
                img_filtered_pixels[i] = new int[img_num_images[i]][];
                img_loaded[i]     = new boolean[img_num_images[i]];
                img_nofilespec[i] = new boolean[img_num_images[i]];
                img_copyfrom[i]   = new boolean[img_num_images[i]];
                img_reuseval[i]   = new int[img_num_images[i]];
                img_width_orig[i] = new int[img_num_images[i]];
                img_height_orig[i]= new int[img_num_images[i]];
                img_pi_audio_file[i] = new String[img_num_images[i]];
                img_pi_clip[i]    = new AudioClip[img_num_images[i]];
            }



            for(j=0;j<img_num_images[i];j++) {

                img_filter_str[i][j] = GetParm.toString(this,
                                    "Img"+(i+1)+"Filter"+(j+1), null);

                img_file[i][j]        = GetParm.toString(this,
                                        "Img"+(i+1)+"Image"+(j+1), null);
                img_xoffset[i][j]     = GetParm.toInt(this,
                                        "Img"+(i+1)+"XOffset"+(j+1),0x7FFFFFFF);
                img_yoffset[i][j]     = GetParm.toInt(this,
                                        "Img"+(i+1)+"YOffset"+(j+1),0x7FFFFFFF);
                if(j == 0) {
                    if(img_xoffset[i][0] == 0x7FFFFFFF)
                        img_xoffset[i][0] = 0;
                    if(img_yoffset[i][0] == 0x7FFFFFFF)
                        img_yoffset[i][0] = 0;
                } else {
                    if(img_xoffset[i][j] == 0x7FFFFFFF)
                        img_xoffset[i][j] = img_xoffset[i][j-1];
                    if(img_yoffset[i][j] == 0x7FFFFFFF)
                        img_yoffset[i][j] = img_yoffset[i][j-1];
                }
                // See if the user specified an image file
                img_file[i][j] = GetParm.toString(this,
                                        "Img"+(i+1)+"Image"+(j+1), null);
                // Get the Image filter
                img_filter_str[i][j] = GetParm.toString(this,
                                            "Img"+(i+1)+"Filter"+(j+1),
                                            "noop");
                // Get the audio per text
                img_pi_audio_file[i][j]   = GetParm.toString(this,
                                        "Img"+(i+1)+"Audio"+(j+1), null);
                if(img_pi_audio_file[i][j] != null)
                    img_pi_clip[i][j] = getAudioClip(getDocumentBase(),
                                    img_pi_audio_file[i][j]);
            }
            img_goingup[i] = true;
            img_firsttime[i] = true;
        }

        // ======== Audio ===========
        for(i=0;i<tx_count;i++) {
            tx_audio_file[i] = GetParm.toString(this, "Tx"+(i+1)+"Audio", null);

            // If the user specified an non-existent or invalid
            // file, then print error message and exit
            if(tx_audio_file[i] != null) {
                tx_clip[i] = getAudioClip(getDocumentBase(), tx_audio_file[i]);
                if(tx_clip[i] == null) {
                    System.out.println(
                        "Error getting " + tx_audio_file[i] + ". " +
                        "Make sure the file exists and is an audio file");
                }
            }
        }

        for(i=0;i<img_count;i++) {
            img_audio_file[i] = GetParm.toString(this, "Img"+(i+1)+"Audio", null);

            // If the user specified an non-existent or invalid
            // file, then print error message and exit
            if(img_audio_file[i] != null) {
                img_clip[i] = getAudioClip(getDocumentBase(), img_audio_file[i]);
                if(img_clip[i] == null) {
                    System.out.println(
                        "Error getting " + img_audio_file[i] + ". " +
                        "Make sure the file exists and is an audio file");
                }
            }
        }

        // ======== URL ===========
        for(i=0;i<tx_count;i++)
            tx_url[i]     = GetParm.toURL(this, "Tx"+(i+1)+"URL");
        for(i=0;i<img_count;i++)
            img_url[i]    = GetParm.toURL(this, "Img"+(i+1)+"URL");

        // ======== BG Image ===========
        tile           = GetParm.toBoolean(this, "AppTile",     false);
        bgimage_file   = GetParm.toString(this,  "AppBGImage",  null);
        bgimageXoffset = GetParm.toInt(this,     "AppBGImageXOffset",  0);
        bgimageYoffset = GetParm.toInt(this,     "AppBGImageYOffset",  0);

        if(bgimage_file != null) {
            ig = new ImgGetr(this);
            bgimage = ig.process(getDocumentBase(), bgimage_file, false);
            ig.prepareImage(bgimage, this);
            bgimage_width = ig.getWidth(bgimage, this);
            bgimage_height = ig.getHeight(bgimage, this);
        }

        // ========== Image Maps ===========
        testmode = GetParm.toBoolean(this, "TestMode",   false);
        num_maps = GetParm.toInt(this,     "AppNumMaps", 0);

        if(num_maps > 0) {
            map_x1  = new int[num_maps];
            map_y1  = new int[num_maps];
            map_x2  = new int[num_maps];
            map_y2  = new int[num_maps];
            map_url = new URL[num_maps];
        }

        for(i=0;i<num_maps;i++) {
            map_x1[i]  = GetParm.toInt(this, "Map"+(i+1)+"_X1",  0);
            map_y1[i]  = GetParm.toInt(this, "Map"+(i+1)+"_Y1",  0);
            map_x2[i]  = GetParm.toInt(this, "Map"+(i+1)+"_X2",  0);
            map_y2[i]  = GetParm.toInt(this, "Map"+(i+1)+"_Y2",  0);
            map_url[i] = GetParm.toURL(this, "URL"+(i+1));
        }
    }

    public void doTxFilters(int i) {
        int j;

        for(j=0;j<tx_num_images[i];j++) {
            // This may get called because the thread was stopped but
            // is now being restarted.  No need to go through all of this
            // again
            if(tx_osi[i][j] != null)
                continue;
            getAppletContext().showStatus("Getting Text"+(i+1)+": "
                                + (j+1) + " of " + tx_num_images[i] + ": "
                                + tx_filter_str[i][j]);
            tx_filtered_pixels[i][j] = doStandardTxFilters(tx_pixels[i],
                            tx_width[i][j], tx_height[i][j],
                            txFilter[i], tx_filter_str[i][j], i, j);
            tx_width[i][j] = txFilter[i].getWidth();
            tx_height[i][j] = txFilter[i].getHeight();
            tx_osi[i][j] = createImage(new MemoryImageSource(tx_width[i][j],
                                tx_height[i][j],
                                ColorModel.getRGBdefault(),
                                tx_filtered_pixels[i][j], 0, tx_width[i][j]));
            Thread.yield();
        }
        getAppletContext().showStatus("");
    }

    public synchronized void loadImages(int i) {
        int j;

        // Now we have all the parameters
        for(j=0;j<img_num_images[i];j++) {

            // If the user didn't specify an image file then
            // get then use the last "gotten" image
            if(img_file[i][j] == null) {
                if(j == 0) {
                    // Error - you must have an image for first image
                    img_num_images[i] = -1;
                    continue;
                }
                img_nofilespec[i][j] = true;
            } else {
                // If the first character of the imf_file is $ then
                // get the previous image and reuse it
                if(img_file[i][j].charAt(0) == '$') {
                    char chars[];
                    int reuseval;
                    int len = img_file[i][j].length();;
                    // Get rid of leading $
                    chars =  new char [len];


                    img_file[i][j].getChars(0, len, chars, 0);
                    
                    reuseval = Integer.parseInt(new String(chars, 1, len-1));
                    reuseval -= 1;
                    img_reuseval[i][j] = reuseval;
                    img_copyfrom[i][j] = true;
                    debug("setting copyfrom to true for " + j);
                } else {
                    // If the user specified an image file then get it
                    img_gettr     = new ImgGetr(this);
                    img[i][j] = img_gettr.process(getDocumentBase(),
                                            img_file[i][j], false);
                    debug("prepareImage["+i+"]["+j+"]\n\n\n\n");
                    img_loaded[i][j] = prepareImage(img[i][j], this);
                    debug("RETURN from prepareImage["+i+"]["+j+"]");
                }
            }
        }
    }

    public synchronized boolean imageUpdate(Image image, int flags,
                                int x, int y, int w, int h) {
        int i, j;
        boolean retval = super.imageUpdate(image, flags, x, y, w, h);
        if((flags & ALLBITS) != 0) {
            for(i=0;i<img_count;i++) {
                for(j=0;j<img_num_images[i];j++) {
                    if(image == img[i][j]) {
                        img_loaded[i][j] = true;
                        debug("imageUpdate: image loaded: " + i + ", " + j);
                        return retval;
                    }
                }
            }
        }
        return retval;
    }

    synchronized public int[] doStandardTxFilters(int[] pixels, int width, int height, 
                                    Filter f, String filter_str, int i, int j) {

        StringTokenizer st;
        String first_token;
        int k;
        int num_elements;

        //========= INVOKE TEXT FILTERS ==============
        if(filter_str != null) {
            setFont(tx_font[i]);
            if(tx_color[i] != null)
                setForeground(tx_color[i]);
            else
                tx_color[i] = Color.black;
            setForeground(tx_color[i]);

            // Before we start the filters, see if the user specified
            // $N as the beginning filter which means to reuse a
            // previous filter output
            //
            st = new StringTokenizer(filter_str, "|");
            num_elements = st.countTokens();
            if(num_elements > 0) {
                String newfilter_str = null;

                first_token = st.nextToken();
                for(k=0;k<num_elements-1;k++) {
                    if(k==0)
                        newfilter_str = st.nextToken();
                    else
                        newfilter_str += st.nextToken();
                    if(k < num_elements-2)
                        newfilter_str += "|";
                }

                if(first_token.charAt(0) == '$') {
                    char chars[];
                    int reuseval;
                    int len = first_token.length();;
                    // Get rid of leading $
                    chars =  new char [len];

                    first_token.getChars(0, len, chars, 0);
                    
                    reuseval = Integer.parseInt(new String(chars, 1, len-1));
                    reuseval -= 1;
                    tx_width[i][j] = tx_width[i][reuseval];
                    tx_height[i][j] = tx_height[i][reuseval];
                    if(tx_bg_color[i] == null) {
                        pixels = f.invoke(this, null, newfilter_str,
                                    tx_filtered_pixels[i][reuseval],
                                    tx_width[i][j], tx_height[i][j],
                                    true, app_bg_color, tx_underline[i],
                                    tx_font[i]);
                    } else {
                        pixels = f.invoke(this, null, newfilter_str,
                                    tx_filtered_pixels[i][reuseval],
                                    tx_width[i][j], tx_height[i][j],
                                    false, tx_bg_color[i], tx_underline[i],
                                    tx_font[i]);
                    }
                    tx_width[i][j]  = f.getWidth();
                    tx_height[i][j] = f.getHeight();
                    return pixels;
                }
            }
            if(tx_bg_color[i] == null) {
                pixels = f.invoke(this, tx[i], filter_str,
                            pixels, tx_width[i][j], tx_height[i][j],
                            true, app_bg_color, tx_underline[i],
                            tx_font[i]);
            } else {
                pixels = f.invoke(this, tx[j], filter_str,
                            pixels, tx_width[i][j], tx_height[i][j],
                            false, tx_bg_color[i], tx_underline[i],
                            tx_font[i]);
            }
            tx_width[i][j]  = f.getWidth();
            tx_height[i][j] = f.getHeight();
        }
        return pixels;
    }

    synchronized public int[] doStandardImgFilters(int[] pixels, int width, int height, 
                                    Filter f, String filter_str, int i, int j) {
        String first_token;
        int k;
        StringTokenizer st;
        int num_elements;

        //========= INVOKE IMAGE FILTERS ==============
        if(filter_str != null) {

            // Before we start the filters, see if the user specified
            // $N as the beginning filter which means to reuse a
            // previous filter output
            //
            st = new StringTokenizer(filter_str, "|");
            num_elements = st.countTokens();
            if(num_elements > 0) {
                String newfilter_str = null;

                first_token = st.nextToken();
                for(k=0;k<num_elements-1;k++) {
                    if(k==0)
                        newfilter_str = st.nextToken();
                    else
                        newfilter_str += st.nextToken();
                    if(k < num_elements-2)
                        newfilter_str += "|";
                }

                if(first_token.charAt(0) == '$') {
                    char chars[];
                    int reuseval;
                    int len = first_token.length();;
                    // Get rid of leading $
                    chars =  new char [len];

                    first_token.getChars(0, len, chars, 0);
                    
                    reuseval = Integer.parseInt(new String(chars, 1, len-1));
                    reuseval -= 1;
                    img_width[i][j] = img_width[i][reuseval];
                    img_height[i][j] = img_height[i][reuseval];
                    pixels = f.invoke(this, null, newfilter_str,
                                    img_filtered_pixels[i][reuseval],
                                    img_width[i][j], img_height[i][j],
                                    true, app_bg_color, false);
                    img_width[i][j]  = f.getWidth();
                    img_height[i][j] = f.getHeight();
                    return pixels;
                }
            }
            pixels = f.invoke(this, null, filter_str,
                            pixels, width, height,
                            true, app_bg_color, false);
        }
        return pixels;
    }

    public void start() {
        int i;

        for(i=0;i<tx_count;i++) {
            if(tx_num_images[i] > 0 && tx_animator[i] == null) {
                tx_animator[i] = new Thread(this, "Tx_thread"+i);
                tx_animator[i].start();
            }
        }

        for(i=0;i<img_count;i++) {
            if(img_num_images[i] > 0 && img_animator[i] == null) {
                img_animator[i] = new Thread(this, "Img_thread");
                img_animator[i].start();
            }
        }
    }

    public void stop() {
        int i;

        for(i=0;i<tx_count;i++) {
            if(tx_animator[i] != null)
                tx_animator[i].stop();
            tx_animator[i] = null;
            // set firstime so that if thread is OnEntry then it won't start up
            tx_firsttime[i] = true;
        }

        for(i=0;i<img_count;i++) {
            if(img_animator[i] != null)
                img_animator[i].stop();
            img_animator[i] = null;
            // set firstime so that if thread is OnEntry then it won't start up
            img_firsttime[i] = true;
        }
    }

    public synchronized boolean checkImgLoadStatus(int i) {
        int j;
        boolean all_loaded = true;


        for(j=0;j<img_num_images[i];j++) {
            if(img_copyfrom[i][j] == true)
                continue;
            if(img_nofilespec[i][j] == true)
                continue;
            if(img_loaded[i][j] == true) {      // flag says it's loaded
                if(img_osi[i][j] == null) {     // No osi image yet
                    // We found one that must have just gotten loaded
                    img_width_orig[i][j] = img_gettr.getWidth(img[i][j], this);
                    img_height_orig[i][j] = img_gettr.getHeight(img[i][j], this);
                    img_pixels[i][j] = img_gettr.getPixels(img[i][j], this);
                    getAppletContext().showStatus("Getting Image"+(i+1)+": "
                                + (j+1) + " of " + img_num_images[i] + ": "
                                + img_filter_str[i][j]);
                    img_filtered_pixels[i][j] = doStandardImgFilters(
                                    img_pixels[i][j],
                                    img_width_orig[i][j],
                                    img_height_orig[i][j],
                                    imgFilter[i], img_filter_str[i][j], i, j);
                    img_width[i][j] = imgFilter[i].getWidth();
                    img_height[i][j]= imgFilter[i].getHeight();
                    img_osi[i][j] = createImage(new MemoryImageSource(
                                        img_width[i][j],
                                        img_height[i][j],
                                        ColorModel.getRGBdefault(),
                                        img_filtered_pixels[i][j],
                                        0, img_width[i][j]));
                    return false;
                }
            } else {
                debug("setting false on " + j);
                all_loaded = false;
            }
        }
        if(all_loaded == true)
            debug("checkImgLoadStatus: ALL LOADED!!!!!");
        return all_loaded;
    }

    //
    // updateImageInfo gets called after all the images have been loaded
    // with getImage.  Now we go through the list of image and see if
    // there were any Image{N} parameters that did not include a
    // file specification (img_nofilespec) and if there were any
    // that were give as "$N" which means to reuse image N
    // In either case, get the appropriate pixels, widht, height,
    // and then process them with the current filter (if any) and
    // create the new image.
    //
    public void updateImageInfo(int i) {
        int j;
        int last_img_file = 0;

        //
        // ===== First, check for images without a image specification
        //
        for(j=0;j<img_num_images[i];j++) {
            if(img_nofilespec[i][j] == true) {
                debug("nofilespec is true for " + i + ", "+j+
                                " last_img_file = " + last_img_file);
                //
                // If an image specification was not given then use
                // the last one that was given
                img_filtered_pixels[i][j] = doStandardImgFilters(
                                img_pixels[i][last_img_file],
                                img_width_orig[i][last_img_file],
                                img_height_orig[i][last_img_file],
                                imgFilter[i], img_filter_str[i][j], i, j);
                img_width[i][j] = imgFilter[i].getWidth();
                img_height[i][j] = imgFilter[i].getHeight();
                img_osi[i][j] = createImage(new MemoryImageSource(
                                        img_width[i][j],
                                        img_height[i][j],
                                        ColorModel.getRGBdefault(),
                                        img_filtered_pixels[i][j],
                                        0, img_width[i][j]));
                Thread.yield();
            } else
                last_img_file = j;
        }

        //
        // ===== Second, check for images with a $N image specification
        //
        for(j=0;j<img_num_images[i];j++) {
            if(img_copyfrom[i][j] == true) {
                //
                // The img_reuseval array contains the index of the
                // image that we should copy
                //
                img[i][j] = img[i][img_reuseval[i][j]];
                img_width[i][j] = img_width[i][img_reuseval[i][j]];
                img_height[i][j] = img_height[i][img_reuseval[i][j]];
                img_pixels[i][j] = img_pixels[i][img_reuseval[i][j]];
                img_width_orig[i][j] = img_width_orig[i][img_reuseval[i][j]];
                img_height_orig[i][j] = img_height_orig[i][img_reuseval[i][j]];
                img_filtered_pixels[i][j] = doStandardImgFilters(
                                img_pixels[i][j],
                                img_width_orig[i][j],
                                img_height_orig[i][j],
                                imgFilter[i], img_filter_str[i][j], i, j);
                img_width[i][j] = imgFilter[i].getWidth();
                img_height[i][j] = imgFilter[i].getHeight();
                img_osi[i][j] = createImage(new MemoryImageSource(
                                        img_width[i][j],
                                        img_height[i][j],
                                        ColorModel.getRGBdefault(),
                                        img_filtered_pixels[i][j],
                                        0, img_width[i][j]));
            }
        }
    }

    synchronized public int getNextTxIndex(int current, int max,
                            int min, int type, int index) {
        switch (type) {
        case Continuous:
        case OnButton:
        case OnEntry:
            current++;
            if(current >= max)
                current = min;
            return current;
        case TwoWay:
        case OnButtonTwoWay:
        case OnEntryTwoWay:
            if(tx_goingup[index] == true)
                current++;
            else
                current--;

            if(current >= max) {
                current = max - 2;
                tx_goingup[index] = false;
            }
            if(current < min) {
                //current = min + 1;
                current = min;
                tx_goingup[index] = true;
            }
            return current;
        default:
            return current;
        }
    }

    synchronized public int getNextImgIndex(int current, int max,
                                int min, int type, int index) {
        switch (type) {
        case Continuous:
        case OnButton:
        case OnEntry:
            current++;
            if(current >= max)
                current = min;
            return current;
        case TwoWay:
        case OnButtonTwoWay:
        case OnEntryTwoWay:
            if(img_goingup[index] == true)
                current++;
            else
                current--;

            if(current >= max) {
                current = max - 2;
                img_goingup[index] = false;
            }
            if(current < min) {
                current = min + 1;
                img_goingup[index] = true;
            }
            return current;
        default:
            return current;
        }
    }
    
    public void tx_sleep(int s_runs, int s_images,
                        int current, int max, int min, int type,
                        int index) {
        switch (type) {
        case Continuous:
        case OnEntry:
            try {
                if(current+1 == max)
                    Thread.sleep(s_runs);
                else
                    Thread.sleep(s_images);
            } catch (InterruptedException e) {}
            break;
        case TwoWay:
            try {
                if(current == min || current == max-1)
                    Thread.sleep(s_runs);
                else
                    Thread.sleep(s_images);
            } catch (InterruptedException e) {}
            break;
        case OnButton:
            try {
                if(current == max-1) {
                    tx_animator[index].suspend();
                } else
                    Thread.sleep(s_images);
            } catch (InterruptedException e) {}
            break;
        case OnButtonTwoWay:
        case OnEntryTwoWay:
            try {
                if(((current == max-1 && tx_goingup[index] == true) ||
                    (current == min) && (tx_goingup[index] == false))) {
                    tx_animator[index].suspend();
                } else
                    Thread.sleep(s_images);
            } catch (InterruptedException e) {}
            break;
        }
    }

    public void img_sleep(int s_runs, int s_images,
                        int current, int max, int min, int type,
                        int index) {
        switch (type) {
        case Continuous:
        case OnEntry:
            try {
                if(current+1 == max)
                    Thread.sleep(s_runs);
                else
                    Thread.sleep(s_images);
            } catch (InterruptedException e) {}
            break;
        case TwoWay:
            try {
                if(current == min || current == max-1)
                    Thread.sleep(s_runs);
                else
                    Thread.sleep(s_images);
            } catch (InterruptedException e) {}
            break;
        case OnButton:
            try {
                if(current == max-1) {
                    img_animator[index].suspend();
                } else
                    Thread.sleep(s_images);
            } catch (InterruptedException e) {}
            break;
        case OnButtonTwoWay:
        case OnEntryTwoWay:
            try {
                if(((current == max-1 && img_goingup[index] == true) ||
                    (current == min) && (img_goingup[index] == false))) {
                    img_animator[index].suspend();
                } else
                    Thread.sleep(s_images);
            } catch (InterruptedException e) {}
            break;
        }
    }


    public void run() {
        int i;
        int index = -1;
        Thread current_thread = Thread.currentThread();;

        //
        // It's important to set the priorities on these
        // threads to some value which is smaller than the
        // priority of the thread that is called to do paint
        Thread.currentThread().setPriority(Thread.MIN_PRIORITY);


        debug("entering run");
        for(i=0;i<img_count;i++) {
            if(current_thread == img_animator[i]) {
                loadImages(i);
                index = i;
                img_index[index] = -1;
                break;
            }
        }

        for(i=0;i<tx_count;i++) {
            if(current_thread == tx_animator[i]) {
                doTxFilters(i);
                index = i;
                tx_index[index] = -1;
                break;
            }
        }

        if(index == -1) {
            //System.out.println("in run: index == -1");
            return;
        }

        while(true) {
            if(tx_animator.length > index) {
                if(Thread.currentThread() == tx_animator[index]) {
                    debug("In tx_animator thread");
                    // ========== Text ===============
                    if(tx_index[index] == -1)   // First time
                        tx_index[index] = tx_lower_bound[index];
                    else
                        tx_index[index] = getNextTxIndex(tx_index[index],
                                                tx_num_images[index],
                                                tx_lower_bound[index],
                                                tx_type[index], index);
                    if(tx_pi_clip[index][tx_index[index]] != null)
                        tx_pi_clip[index][tx_index[index]].play();
                    repaint();
                    tx_sleep(tx_delay_runs[index], tx_delay_images[index],
                                        tx_index[index], tx_num_images[index],
                                        tx_lower_bound[index],
                                        tx_type[index], index);
                    if((tx_firsttime[index] == true) && 
                                ((tx_type[index] == OnEntry) || 
                                (tx_type[index] == OnEntryTwoWay))) {
                        tx_firsttime[index] = false;
                        tx_animator[index].suspend();
                    }
                }
            } 
            if(img_animator.length > index) {
                if(Thread.currentThread() == img_animator[index]) {
                    debug("In img_animator thread");
                    // ========== Image ===============
                    if(all_images_loaded[index] == false) {
                        all_images_loaded[index] = checkImgLoadStatus(index);
                        if(all_images_loaded[index] == true) 
                            updateImageInfo(index);
                    }
                    if(all_images_loaded[index] == true) {
                        if(img_index[index] == -1)      // First time
                            img_index[index] = img_lower_bound[index];
                        else
                            img_index[index] = getNextImgIndex(img_index[index],
                                                img_num_images[index],
                                                img_lower_bound[index],
                                                img_type[index], index);
                    if(img_pi_clip[index][img_index[index]] != null)
                        img_pi_clip[index][img_index[index]].play();
                        repaint();
                        img_sleep(img_delay_runs[index], img_delay_images[index],
                                        img_index[index], img_num_images[index],
                                        img_lower_bound[index],
                                        img_type[index], index);
                        if((img_firsttime[index] == true) && 
                                    ((img_type[index] == OnEntry) || 
                                    (img_type[index] == OnEntryTwoWay))) {
                            img_firsttime[index] = false;
                            img_animator[index].suspend();
                        }
                    } else {
                        img_index[index] = 0;
                        if(img_display_first[index]) {
                            repaint();
                        } else
                            img_index[index] = -1;
                        try {
                            Thread.sleep(1000);
                        } catch (InterruptedException e) {
                            break;
                        }
                    }
                }
            }
        }
    }

    //
    // Overwrite the default update method
    // This may not be necessary for this case
    // but it certainly doesn't hurt
    //
    public void update(Graphics g) {
        paint(g);
    }

    //
    // Our paint method handles several tasks:
    //    Calculates the x and y coordinates base on params
    //    Draws the applet frame
    //    Draws the applet border
    //    Draws the text frame
    //    Draws the text border
    //    Draws the underline
    //    Draws the text
    //
    public void paint(Graphics g) {
        int i, j;

        // If this is the first time we've been in paint
        // or the applet has changed size then create
        // an image the size of the applet into which
        // we write the background image
        if(applet_width != size().width ||
           applet_height!= size().height) {
            applet_width  = size().width;
            applet_height = size().height;
            offScreen = createImage(applet_width, applet_height);
            offGC = offScreen.getGraphics();
        }
        offGC.setColor(getBackground());
        offGC.fillRect(0, 0, applet_width, applet_height);
        tilebackground(offGC);

        for(i=0;i<img_count;i++) {
            j = img_index[i];
            if(img_index[i] >= 0) {
                if(img_osi[i][j] != null) {

                    img_x_coord[i][j] = calcULXcoord(img_xcenter[i],
                                                    img_horizcenter[i],
                                                    size().width,
                                                    img_width[i][j],
                                                    img_xoffset[i][j]);
                    img_y_coord[i][j] = calcULYcoord(img_ycenter[i],
                                                    img_vertcenter[i],
                                                    size().height,
                                                    img_height[i][j],
                                                    img_yoffset[i][j]);


                    offGC.drawImage(img_osi[i][j], img_x_coord[i][j],
                                                    img_y_coord[i][j], this);
                    drawFrame(offGC, img_x_coord[i][j], img_y_coord[i][j],
                                    img_width[i][j],
                                    img_height[i][j], img_fr_type[i],
                                    img_fr_thick[i], img_fr_margin[i]);
                    drawBorder(offGC, img_x_coord[i][j], img_y_coord[i][j],
                                    img_width[i][j],
                                    img_height[i][j], img_borderwidth[i],
                                    img_bordermargin[i], img_bordercolor[i]);
                }
            }
        }
        // We never know when the size might change so if horizcenter
        // is true then recalculate the x_coord every time
        for(i=0;i<tx_count;i++) {
            j = tx_index[i];
            if(tx_osi[i][j] != null) {

                tx_x_coord[i][j] = calcULXcoord(tx_xcenter[i],
                                tx_horizcenter[i], size().width,
                                tx_width[i][j], tx_xoffset[i][j]);
                tx_y_coord[i][j] = calcULYcoord(tx_ycenter[i],
                                    tx_vertcenter[i], size().height,
                                    tx_height[i][j], tx_yoffset[i][j]);


                // Draw the text
                offGC.drawImage(tx_osi[i][j], tx_x_coord[i][j],
                                            tx_y_coord[i][j], this);

                // Draw frame for text
                drawFrame(offGC, tx_x_coord[i][j], tx_y_coord[i][j],
                                    tx_width[i][j], tx_height[i][j],
                                    tx_fr_type[i], tx_fr_thick[i],
                                    tx_fr_margin[i]);


                // Draw border for text
                drawBorder(offGC, tx_x_coord[i][j], tx_y_coord[i][j],
                                    tx_width[i][j], tx_height[i][j],
                                    tx_borderwidth[i], tx_bordermargin[i],
                                    tx_bordercolor[i]);
            }
        }
        // Draw frame for applet
        drawFrame(offGC, 0, 0, size().width, size().height,
                            app_fr_type, app_fr_thick, 0);
        // Draw border for applet
        drawBorder(offGC, 0, 0, size().width, size().height,
                            app_borderwidth, 0, app_bordercolor);

        g.drawImage(offScreen, 0, 0, this);
    }

    public boolean keyDown(Event evt, int key) {
        int i, j;
        int x = evt.x;
        int y = evt.y;

        // Handle the text next (Text takes precedence over images)
        for(i=0;i<tx_count;i++) {
            if(tx_index[i] != -1) {
                j = tx_index[i];
                if(x > tx_x_coord[i][j] && 
                   x < tx_x_coord[i][j] + tx_width[i][j] &&
                   y > tx_y_coord[i][j] && 
                   y < tx_y_coord[i][j] + tx_height[i][j]) {
                    if(key == 'f') {
                        tx_delay_images[i] *= .9;
                        getAppletContext().showStatus("Text " + (i+1) + 
                                ": Image delay: " + tx_delay_images[i]);
                        if(tx_delay_images[i] < 2)
                            tx_delay_images[i] = 2;
                    } else if(key == 's') {
                        if(tx_delay_images[i] < 10)
                            tx_delay_images[i]++;
                        else
                            tx_delay_images[i] *= 1.1;
                        getAppletContext().showStatus("Text " + (i+1) + 
                                ": Image delay: " + tx_delay_images[i]);
                    } else if(key == 'F') {
                        tx_delay_runs[i] *= .9;
                        if(tx_delay_runs[i] < 2)
                            tx_delay_runs[i] = 2;
                        getAppletContext().showStatus("Text " + (i+1) + 
                                ": Run delay: " + tx_delay_runs[i]);
                    } else if(key == 'S') {
                        if(tx_delay_runs[i] < 10)
                            tx_delay_runs[i]++;
                        else
                            tx_delay_runs[i] *= 1.1;
                        getAppletContext().showStatus("Text " + (i+1) + 
                                ": Run delay: " + tx_delay_runs[i]);
                    }
                }    
            }
        }

        // Handle the text next (Text takes precedence over images)
        for(i=0;i<img_count;i++) {
            if(img_index[i] != -1) {
                j = img_index[i];
                if(x > img_x_coord[i][j] && 
                   x < img_x_coord[i][j] + img_width[i][j] &&
                   y > img_y_coord[i][j] && 
                   y < img_y_coord[i][j] + img_height[i][j]) {
                    if(key == 'f') {
                        img_delay_images[i] *= .9;
                        getAppletContext().showStatus("Image " + (i+1) + 
                                ": Image delay: " + img_delay_images[i]);
                        if(img_delay_images[i] < 2)
                            img_delay_images[i] = 2;
                    } else if(key == 's') {
                        if(img_delay_images[i] < 10)
                            img_delay_images[i]++;
                        else
                            img_delay_images[i] *= 1.1;
                        getAppletContext().showStatus("Image " + (i+1) + 
                                ": Image delay: " + img_delay_images[i]);
                    } else if(key == 'F') {
                        img_delay_runs[i] *= .9;
                        if(img_delay_runs[i] < 2)
                            img_delay_runs[i] = 2;
                        getAppletContext().showStatus("Image " + (i+1) + 
                                ": Run delay: " + img_delay_runs[i]);
                    } else if(key == 'S') {
                        if(img_delay_runs[i] < 10)
                            img_delay_runs[i]++;
                        else
                            img_delay_runs[i] *= 1.1;
                        getAppletContext().showStatus("Image " + (i+1) + 
                                ": Run delay: " + img_delay_runs[i]);
                    }
                }    
            }
        }
        return true;
    }

    //
    // If the user clicks the mouse then load the URL
    // We check the text position first, so if there is
    // overlapping text and image then the text URL will
    // be loaded, not the imgae URL
    //
    public boolean mouseDown(Event evt, int x, int y) {
        int i, j;

        for(i=0;i<tx_count;i++) {
            if(tx_index[i] != -1) {
                j = tx_index[i];
                switch(tx_type[i]) {
                case OnButton:
                case OnButtonTwoWay:
                    if(x > tx_x_coord[i][j] &&
                       x < tx_x_coord[i][j] + tx_width[i][j] &&
                       y > tx_y_coord[i][j] &&
                       y < tx_y_coord[i][j] + tx_height[i][j]) {
                        if(tx_animator[i].isAlive()) {
                            tx_animator[i].resume();
                        }
                    }
                    break;
                // Toggle thread on mouse press
                case Continuous:
                case TwoWay:
                    if(x > tx_x_coord[i][j] &&
                       x < tx_x_coord[i][j] + tx_width[i][j] &&
                       y > tx_y_coord[i][j] &&
                       y < tx_y_coord[i][j] + tx_height[i][j]) {
                        if(tx_animator[i].isAlive()) {
                            if(tx_threadSuspended[i])
                                tx_animator[i].resume();
                            else
                                tx_animator[i].suspend();
                            tx_threadSuspended[i] = !tx_threadSuspended[i];
                        }
                    }
                    break;
                }
            }
        }

        for(i=0;i<img_count;i++) {
            if(img_index[i] != -1) {
                j = img_index[i];
                switch(img_type[i]) {
                case OnButton:
                case OnButtonTwoWay:
                    if(x > img_x_coord[i][j] &&
                       x < img_x_coord[i][j] + img_width[i][j] &&
                       y > img_y_coord[i][j] &&
                       y < img_y_coord[i][j] + img_height[i][j]) {
                        if(img_animator[i].isAlive()) {
                            img_animator[i].resume();
                        }
                    }    
                    break;
                // Toggle thread on mouse press
                case Continuous:
                case TwoWay:
                    if(x > img_x_coord[i][j] &&
                       x < img_x_coord[i][j] + img_width[i][j] &&
                       y > img_y_coord[i][j] &&
                       y < img_y_coord[i][j] + img_height[i][j]) {
                        if(img_animator[i].isAlive()) {
                            if(img_threadSuspended[i])
                                img_animator[i].resume();
                            else
                                img_animator[i].suspend();
                            img_threadSuspended[i] = !img_threadSuspended[i];
                        }
                    }
                    break;
                }
            }
        }


        // Handle the image maps first
        for(i=0;i<num_maps;i++) {
            if(x > map_x1[i] && x < map_x2[i] &&
               y > map_y1[i] && y < map_y2[i]) {
                if(map_url[i] != null) {
                    if(testmode) {
                        getAppletContext().showStatus("Loading URL: " +
                                    map_url[i].toString());
                        System.out.println("Loading URL: " +
                                    map_url[i].toString());
                    } else
                        getAppletContext().showDocument(map_url[i]);
                    return true;
                }
            }
        }    

        // Handle the text next (Text takes precedence over images)
        // It's possible that the text may be moving
        for(i=0;i<tx_count;i++) {
            // If tx_index[i] is -1 then not ready so skip
            if(tx_index[i] == -1)
                continue;
            j = tx_index[i];
            if(x > tx_x_coord[i][j] && x < tx_x_coord[i][j] + tx_width[i][j] &&
               y > tx_y_coord[i][j] && y < tx_y_coord[i][j] + tx_height[i][j]) {
                if(tx_url[i] != null) {
                    getAppletContext().showDocument(tx_url[i]);
                    System.out.println("loading " + tx_url[i].toString());
                    return true;
                }
            }    
        }

        // Handle the image last
        for(i=0;i<img_count;i++) {
            // If img_index[i] is -1 then not ready so skip
            if(img_index[i] == -1)
                continue;
            j = img_index[i];
            if(x > img_x_coord[i][j] && x < img_x_coord[i][j] + img_width[i][j] &&
               y > img_y_coord[i][j] && y < img_y_coord[i][j] + img_height[i][j]) {
                if(img_url[i] != null) {
                    getAppletContext().showDocument(img_url[i]);
                    System.out.println("loading " + img_url[i].toString());
                    return true;
                }
            }    
        }
        return true;
    }

    //
    // Check the x and y coordinates to see if we're actually
    // inside the boundaries of the text or image or both
    //
    public boolean mouseMove(Event evt, int x, int y) {
        boolean im_status_shown = false;
        boolean tx_status_shown[] = new boolean[tx_count];
        boolean img_status_shown[] = new boolean[img_count];
        int i, j;

        if(testmode) {
            getAppletContext().showStatus("X: " + x + " Y: " + y);
            return true;
        }

        // Handle the text next (Text takes precedence over images)
        for(i=0;i<tx_count;i++) {
            if(tx_index[i] != -1) {
                j = tx_index[i];
                if(x > tx_x_coord[i][j] &&
                   x < tx_x_coord[i][j] + tx_width[i][j] &&
                   y > tx_y_coord[i][j] &&
                   y < tx_y_coord[i][j] + tx_height[i][j]) {
                    within_tx[i] = true;
                } else {
                    within_tx[i] = false;
                }
            }
            switch(tx_type[i]) {
            case OnEntry:
                if(tx_animator[i].isAlive() && within_tx[i] == true &&
                                last_within_tx[i] == false) {
                    tx_animator[i].resume();
                } else if(tx_animator[i].isAlive() && within_tx[i] == false &&
                                last_within_tx[i] == true) {
                    if(tx_return_start[i] == true) {
                        tx_index[i] = tx_lower_bound[i];
                        repaint();
                    }
                    tx_animator[i].suspend();
                }
                break;
            case OnEntryTwoWay:
                if(tx_animator[i].isAlive() && within_tx[i] == true &&
                                last_within_tx[i] == false) {
                    tx_goingup[i] = true;
                    tx_animator[i].resume();
                } else if(tx_animator[i].isAlive() && within_tx[i] == false &&
                                last_within_tx[i] == true) {
                    tx_goingup[i] = false;
                    tx_animator[i].resume();
                }
                break;
            }
            last_within_tx[i] = within_tx[i];
        }

        // Handle the text next (Text takes precedence over images)
        for(i=0;i<img_count;i++) {
            if(img_index[i] != -1) {
                j = img_index[i];
                if(x > img_x_coord[i][j] &&
                   x < img_x_coord[i][j] + img_width[i][j] &&
                   y > img_y_coord[i][j] &&
                   y < img_y_coord[i][j] + img_height[i][j]) {
                    within_img[i] = true;
                } else {
                    within_img[i] = false;
                }
            }
            switch(img_type[i]) {
            case OnEntry:
                if(img_animator[i].isAlive() && within_img[i] == true &&
                                last_within_img[i] == false) {
                    img_animator[i].resume();
                } else if(img_animator[i].isAlive() && within_img[i] == false &&
                                last_within_img[i] == true) {
                    if(img_return_start[i] == true) {
                        img_index[i] = img_lower_bound[i];
                        repaint();
                    }
                    img_animator[i].suspend();
                }
                break;
            case OnEntryTwoWay:
                if(img_animator[i].isAlive() && within_img[i] == true &&
                                last_within_img[i] == false) {
                    img_goingup[i] = true;
                    img_animator[i].resume();
                } else if(img_animator[i].isAlive() && within_img[i] == false &&
                                last_within_img[i] == true) {
                    img_goingup[i] = false;
                    img_animator[i].resume();
                }
                break;
            }
            last_within_img[i] = within_img[i];
        }


        /*
        // Handle the image last
        for(i=0;i<img_count;i++) {
            if(img_index[i] != -1) {
                if(img_type[i] == OnEntry) {
                    j = img_index[i];
                    if(x > img_x_coord[i][j] &&
                       x < img_x_coord[i][j] + img_width[i][j] &&
                       y > img_y_coord[i][j] &&
                       y < img_y_coord[i][j] + img_height[i][j]) {
                        within_img[i] = true;
                        if(img_animator[i].isAlive() && within_img[i] == true && 
                                                    last_within_img[i] == false) {
                            img_animator[i].resume();
                        }
                    } else {
                        if(img_return_start[i] == true) {
                            img_index[i] = img_lower_bound[i];
                            repaint();
                        }
                        img_animator[i].suspend();
                        within_img[i] = false;
                    }
                    last_within_img[i] = within_img[i];
                }
            }
        }
        */





        // =========== AUDIO ==============
        // Process the image first
        for(i=0;i<img_count;i++) {
            j = img_index[i];
            // If the images are not yet loaded then skip
            if(img_index[i] == -1)
                continue;
            if(x > img_x_coord[i][j] && x < img_x_coord[i][j] + img_width[i][j] &&
               y > img_y_coord[i][j] && y < img_y_coord[i][j] + img_height[i][j]) {
                if(!img_playing[i]) {
                    img_playing[i] = true;
                    if(img_clip[i] != null)
                        img_clip[i].play();
                } // else it's already playing, leave it alone
            } else {
                // Since we don't get any kind of notification when the audio
                // clip has terminated (due to playing to completion) we set
                // playing to false everytime we move out of the boundaries
                // of the string
                img_playing[i] = false;
                if(img_clip[i] != null) {
                    img_clip[i].stop();
                }
            }
        }

        // Process the text last
        for(i=0;i<tx_count;i++) {
            // If the text is not yet loaded then skip
            if(tx_index[i] == -1)
                continue;
            j = tx_index[i];
            if(x > tx_x_coord[i][j] && x < tx_x_coord[i][j] + tx_width[i][j] &&
               y > tx_y_coord[i][j] && y < tx_y_coord[i][j] + tx_height[i][j]) {
                if(!tx_playing[i]) {
                    tx_playing[i] = true;
                    if(tx_clip[i] != null)
                        tx_clip[i].play();
                } // else it's already playing, leave it alone
            } else {
                // Since we don't get any kind of notification when the audio
                // clip has terminated (due to playing to completion) we set
                // playing to false everytime we move out of the boundaries
                // of the string
                tx_playing[i] = false;
                if(tx_clip[i] != null) {
                    tx_clip[i].stop();
                }
            }
        }

        // =========== IMAGE MAPS ==============
        // Now check the image maps
        for(i=0;i<num_maps;i++) {
            if(x > map_x1[i] && x < map_x2[i] &&
               y > map_y1[i] && y < map_y2[i]) {
                if(map_url[i] != null) {
                    getAppletContext().showStatus(map_url[i].toString());
                    im_status_shown = true;
                }
            }
        }

        //
        // If a image map url was found (hence tx_status_shown == true)
        // then return and don't bother to call the other mouseMove
        // stuff would simply display the url for the text or image.
        //
        if(im_status_shown == true) {
            return true;
        }

        // =========== URLs ==============
        // Handle the text first
        for(i=0;i<tx_count;i++) {
            // If the text is not yet loaded then skip
            if(tx_index[i] == -1)
                continue;
            j = tx_index[i];
            if(x > tx_x_coord[i][j] && x < tx_x_coord[i][j] + tx_width[i][j] &&
               y > tx_y_coord[i][j] && y < tx_y_coord[i][j] + tx_height[i][j]) {
                if(tx_url[i] != null) {
                    getAppletContext().showStatus(tx_url[i].toString());
                    tx_status_shown[i] = true;
                }
            }    
        }

        // Handle the image last
        // If a Text URL is being displayed then don't
        // display the image URL
        for(i=0;i<img_count;i++) {
            // If the image is not yet loaded then skip
            if(img_index[i] == -1)
                continue;
            j = img_index[i];
            if(x > img_x_coord[i][j] && x < img_x_coord[i][j] + img_width[i][j] &&
               y > img_y_coord[i][j] && y < img_y_coord[i][j] + img_height[i][j]) {
                if(img_url[i] != null && img_status_shown[i] == false) {
                    getAppletContext().showStatus(img_url[i].toString());
                    img_status_shown[i] = true;
                }
            }    
        }

        for(i=0;i<tx_count;i++)
            if(tx_status_shown[i] == true)
                return true;

        for(i=0;i<img_count;i++)
            if(img_status_shown[i] == true)
                return true;

        // We only get this far if nobody was showing status
        // Then we clear it
        getAppletContext().showStatus(null);
        return true;
    }

    //
    // Just to be sure, when we exit the applet, stop the clips
    // It should already be stopped
    //
    public boolean mouseExit(Event evt, int x, int y) {
        int i;
        // Just to be sure

        for(i=0;i<tx_count;i++) {
            tx_playing[i] = false;
            if(tx_clip[i] != null)
                tx_clip[i].stop();
            within_tx[i] = false;
            last_within_tx[i] = false;
        }

        for(i=0;i<img_count;i++) {
            img_playing[i] = false;
            if(img_clip[i] != null)
                img_clip[i].stop();
            within_img[i] = false;
            last_within_img[i] = false;
        }

        getAppletContext().showStatus(null);
        return true;
    }

    //
    // Calculate the upper left corner position when centering
    // an item that's drawn in relation to its upper left corner
    // such as images
    //
    protected int calcULCenter(int panelsize, int objectsize) {
        return ((panelsize / 2) - (objectsize / 2));
    }

    //
    // Calculate the lower left corner position when centering
    // an item that's drawn in relation to its lower left corner
    // such as text
    //
    protected int calcLLCenter(int panelsize, int objectsize) {
        return ((panelsize / 2) + (objectsize / 2));
    }

    //
    // A general purpose text drawing method
    // This method is used by the Img class
    //
    protected void drawText(Graphics g, String txt, int x, int y,
                                            Color c, Font f) {
        if(txt != null) {
            g.setColor(c);
            g.setFont(f);
            g.drawString(txt, x, y);
        }
        
    }

    //
    // The drawBorder method draws a border at the 
    // specified x and y position
    //
    protected void drawBorder(Graphics g,
                           int x,
                           int y,
                           int w,
                           int h,
                           int borderwidth,
                           int margin,
                           Color c) {
        if(borderwidth > 0) {
            g.setColor(c);
            for(int i=0;i<borderwidth;i++) {
                g.drawRect(x-margin+i,
                           y-margin+i,
                           w+(margin*2)-(i*2)-1,
                           h+(margin*2)-(i*2)-1);
            }
        }
    }

    void drawTopLine(Graphics g,
                    int x, int y, int w, int h, int i, int margin) {
        g.drawLine(x-margin+i, y-margin+i, x+w+margin-i-1, y-margin+i);
    }

    void drawBottomLine(Graphics g,
                    int x, int y, int w, int h, int i, int margin) {
        g.drawLine(x-margin+i, y+h+margin-i-1, x+w+margin-i-1, y+h+margin-i-1);
    }

    void drawLeftLine(Graphics g,
                    int x, int y, int w, int h, int i, int margin) {
        g.drawLine(x-margin+i, y-margin+i, x-margin+i, y+h+margin-i-1);
    }

    void drawRightLine(Graphics g,
                    int x, int y, int w, int h, int i, int margin) {
        g.drawLine(x+w+margin-i-1, y-margin+i, x+w+margin-i-1, y+h+margin-i-1);
    }
        

    //
    // The drawFrame draws a frame at the specified x and y position
    // A frame can be one of the following four types: ShadowIn,
    // ShadowOut, ShadowEtchedIn, ShadowEtchedOut
    //
    protected void drawFrame(Graphics g,    // Graphics to draw to
                           int x,           // X upper left position
                           int y,           // Y upper left position
                           int w,           // width
                           int h,           // height
                           String type,     // Type of frame
                           int thickness,   // thickness of frame
                           int margin) {    // Margin around object
        int i;
        if(thickness == 0)
            return;
        Color darker = AdjColor.darkenit(app_bg_color, .50);
        Color slightlydarker = AdjColor.darkenit(app_bg_color, .10);
        Color brighter = AdjColor.brightenit(app_bg_color, .50);

        if(thickness > 0) {
            if(type.equalsIgnoreCase("shadowout")) {
                for(i=0;i<thickness;i++) {
                    g.setColor(brighter);
                    // TOP
                    drawTopLine(g, x, y, w, h, i, margin);
                    // LEFT
                    drawLeftLine(g, x, y, w, h, i, margin);

                    g.setColor(darker);
                    // BOTTOM
                    drawBottomLine(g, x, y, w, h, i, margin);
                    // RIGHT
                    drawRightLine(g, x, y, w, h, i, margin);
                }
            } else if(type.equalsIgnoreCase("shadowetchedin")) {
                for(i=0;i<thickness;i++) {
                    if(i == 0)
                        g.setColor(darker);
                    else if (i == thickness-1)
                        g.setColor(brighter);
                    else
                        g.setColor(slightlydarker);
                    // TOP
                    drawTopLine(g, x, y, w, h, i, margin);
                    // LEFT
                    drawLeftLine(g, x, y, w, h, i, margin);

                    if(i == 0)
                        g.setColor(brighter);
                    else if (i == thickness-1)
                        g.setColor(darker);
                    else
                        g.setColor(slightlydarker);
                    // BOTTOM
                    drawBottomLine(g, x, y, w, h, i, margin);
                    // RIGHT
                    drawRightLine(g, x, y, w, h, i, margin);
                }

            } else if(type.equalsIgnoreCase("shadowetchedout")) {
                for(i=0;i<thickness;i++) {
                    if(i == 0)
                        g.setColor(brighter);
                    else if (i == thickness-1)
                        g.setColor(darker);
                    else
                        g.setColor(app_bg_color);
                    // TOP
                    drawTopLine(g, x, y, w, h, i, margin);
                    // LEFT
                    drawLeftLine(g, x, y, w, h, i, margin);

                    if(i == 0)
                        g.setColor(darker);
                    else if (i == thickness-1)
                        g.setColor(brighter);
                    else
                        g.setColor(app_bg_color);
                    // BOTTOM
                    drawBottomLine(g, x, y, w, h, i, margin);
                    // RIGHT
                    drawRightLine(g, x, y, w, h, i, margin);
                }

            } else {    // Default to "shadowin"
                for(i=0;i<thickness;i++) {
                    g.setColor(darker);
                    // TOP
                    drawTopLine(g, x, y, w, h, i, margin);
                    // LEFT
                    drawLeftLine(g, x, y, w, h, i, margin);

                    g.setColor(brighter);
                    // BOTTOM
                    drawBottomLine(g, x, y, w, h, i, margin);
                    // RIGHT
                    drawRightLine(g, x, y, w, h, i, margin);
                }
            }
        }
    }

    //
    // The drawUnderline method draws an underline
    // The underline is actually a filled rectangle
    // whose size is based on the width of a pipe
    // character from the same font.  It works!
    //
    protected void drawUnderline(Graphics g,
                           int x,
                           int y,
                           int w,
                           Color c) {
        int i;

        for(i=0;i<img_count;i++) {
        if(tx_underline[i] == true) {
                int pipewidth = tx_fontmetrics[i].charWidth('|');;
                g.setColor(c);
                g.fillRect(x, y, w, Math.max(1, pipewidth/4));
            }
        }
    }

    //
    // Calculate the upper left X coordinate based
    // on parameters such as horizcenter and xoffset
    //
    protected int calcULXcoord(int xcenter, boolean hcent, int appletwidth,
                                int objectwidth, int xoffset) {
        int x;

        // Horizontal (x coordinate) Position
        if(xcenter != -1)
            x = xcenter - (objectwidth/2);
        else if(hcent)
            x = calcULCenter(appletwidth, objectwidth);
        else if (xoffset == -1)
            x = 0;
        else
            x = xoffset;
        
        return x;
    }

    //
    // Calculate the upper left Y coordinate based
    // on parameters such as vertcenter and yoffset
    //
    protected int calcULYcoord(int ycenter, boolean hcent, int appletheight,
                            int objectheight, int yoffset) {
        int y;

        // Vertical (y coordinate) Position
        if(ycenter != -1)
            y = ycenter - (objectheight/2);
        else if(hcent)
            y = calcULCenter(appletheight, objectheight);
        else if (yoffset == -1)
            y = 0;
        else
            y = yoffset;

        return y;
    }

    //
    // Calculate the lower left Y coordinate based
    // on parameters such as vertcenter and yoffset
    //
    protected int calcLLYcoord(boolean hcent, int appletheight,
                    int objectheight, int objectdescent, int yoffset) {
        int y;

        // Vertical (y coordinate) Position
        if(hcent)
            y = calcLLCenter(appletheight, objectheight) - objectdescent;
        else if (yoffset == -1)
            y = objectheight - objectdescent;
        else
            y = yoffset;

        return y;
    }

    // 
    // Tile the background with the specified image
    //
    protected void tilebackground(Graphics g) {
        int i, j;

        if(bgimage == null)
            return;
        if(ig.isImagePrepared(bgimage, this) == false)
            ig.loadImageAndWait(bgimage, this);
        //
        // If the applet has changed size or this is the first time
        // tilebackground has been called then create the bg image
        //
        if(app_width != size().width || app_height != size().height) {
            app_width  = size().width;
            app_height = size().height;

            tiled_bgimage = createImage(size().width, size().height); 
            bg_g = tiled_bgimage.getGraphics();
            bg_g.setColor(app_bg_color);
            bg_g.fillRect(0, 0, size().width, size().height);
            if(tile && (bgimageXoffset != 0 || bgimageYoffset != 0)) {
                for(i=-bgimage_height;i<app_height;i+=bgimage_height) {
                    for(j=-bgimage_width;j<app_width;j+=bgimage_width) {
                        bg_g.drawImage(bgimage, j+bgimageYoffset,
                                        i+bgimageXoffset, this);
                    }
                }

            } else if(tile) {
                for(i=-bgimage_height;i<app_height;i+=bgimage_height) {
                    for(j=-bgimage_width;j<app_width;j+=bgimage_width) {
                        bg_g.drawImage(bgimage, j+2, i+2, this);
                    }
                }
            } else {
                bg_g.drawImage(bgimage, bgimageXoffset,
                                        bgimageYoffset, this);
            }
        }

        if(tiled_bgimage != null) {
            g.drawImage(tiled_bgimage, 0, 0, this);
        }
    }

    public void debug(String s) {
        if(debugflag)
            System.err.println(s);
    }

    public int getAnimType(String s) {
        if(s.equalsIgnoreCase("Continuous"))
            return Continuous;
        else if(s.equalsIgnoreCase("TwoWay"))
            return TwoWay;
        else if(s.equalsIgnoreCase("OnButton"))
            return OnButton;
        else if(s.equalsIgnoreCase("OnButtonTwoWay"))
            return OnButtonTwoWay;
        else if(s.equalsIgnoreCase("OnEntry"))
            return OnEntry;
        else if(s.equalsIgnoreCase("OnEntryTwoWay"))
            return OnEntryTwoWay;
        else {
            System.out.println("Unknown animation type: " + s);
            return Continuous;
        }
    }
}
