import java.awt.*;
import java.awt.image.*;
import java.applet.Applet;
import java.awt.MediaTracker;

public class AnimateOnButtonTwoWay extends Animate implements Runnable {
    boolean tx_goingup = true;
    boolean img_goingup = true;
    protected boolean   tx_runonce;
    protected boolean   img_runonce;
    protected boolean   tx_firsttime = true;
    protected boolean   img_firsttime = true;

    public void init() {
        tx_runonce       = GetParm.toBoolean(this, "TxRunOnce", false);
        img_runonce      = GetParm.toBoolean(this, "ImgRunOnce", false);
        super.init();
    }

    // getNextTxIndex() is called right before repaint
    public int getNextTxIndex(int current, int max, int min) {
        if(tx_goingup == true)
            current++;
        else
            current--;
    
        if(current >= max) {
            current = max - 2;
            tx_goingup = false;
        }
        if(current < min) {
            current = min + 1;
            tx_goingup = true;
        }
        return current;
    }

    // getNextImgIndex() is called right before repaint
    public int getNextImgIndex(int current, int max, int min) {
        if(img_goingup == true)
            current++;
        else
            current--;
    
        if(current >= max) {
            current -= 2;
            img_goingup = false;
        }
        if(current < min) {
            current = min+1;
            img_goingup = true;
        }
        return current;
    }

    
    // tx_sleep() is called right after repaint
    public void tx_sleep(int s_runs, int s_images,
                            int current, int max, int min) {
        try {
            if(((current == max-1 && tx_goingup == true) ||
                (current == min) && (tx_goingup == false))) {
                Thread.currentThread().suspend();
            } else
                Thread.sleep(s_images);
        } catch (InterruptedException e) {}
    }

    // img_sleep() is called right after repaint
    public void img_sleep(int s_runs, int s_images,
                            int current, int max, int min) {
        try {
            if(((current == max-1 && img_goingup == true) ||
                (current == min) && (img_goingup == false))) {
                Thread.currentThread().suspend();
            } else
                Thread.sleep(s_images);
        } catch (InterruptedException e) {}
    }

    public boolean mouseDown(Event evt, int x, int y) {
        // Handle the text next (Text takes precedence over images)
        if(tx_index != -1) {
            if(x > tx_x_coord[tx_index] && 
               x < tx_x_coord[tx_index] + tx_width[tx_index] &&
               y > tx_y_coord[tx_index] && 
               y < tx_y_coord[tx_index] + tx_height[tx_index]) {
                if(tx_animator.isAlive()) {
                    tx_animator.resume();
                }
            }    
        }

        // Handle the image last
        if(img_index != -1) {
            if(x > img_x_coord[img_index] &&
               x < img_x_coord[img_index] + img_width[img_index] &&
               y > img_y_coord[img_index] &&
               y < img_y_coord[img_index] + img_height[img_index]) {
                if(img_animator.isAlive()) {
                    img_animator.resume();
                }
            }    
        }
        return super.mouseDown(evt, x, y);
    }

    // In order to get the RunOnce behavior, we have to override
    // the run method.  Most of what's in the run method is unchanged
    // from the parent class with the exception of the run once stuff
    //
    public void run() {
        Thread current_thread = Thread.currentThread();;

        //
        // It's important to set the priorities on these
        // threads to some value which is smaller than the
        // priority of the thread that is called to do paint
        Thread.currentThread().setPriority(Thread.MIN_PRIORITY);

        tx_index = -1;
        img_index = -1;

        debug("entering run");
        if(current_thread == img_animator) {
            loadImages();
        }

        if(current_thread == tx_animator) {
            doTxFilters();
        }

        while(true) {
            if(Thread.currentThread() == tx_animator) {
                debug("In tx_animator thread");
                // ========== Text ===============
                if(tx_index == -1)      // First time
                    tx_index = tx_lower_bound;
                else
                    tx_index = getNextTxIndex(tx_index, tx_num_images,
                                            tx_lower_bound);
                if(tx_pi_clip[tx_index] != null)
                    tx_pi_clip[tx_index].play();
                repaint();
                tx_sleep(tx_delay_runs, tx_delay_images,
                                    tx_index, tx_num_images,
                                    tx_lower_bound);
                if(tx_firsttime == true) {
                    tx_firsttime = false;
                    if(tx_runonce == false) {
                        tx_animator.suspend();
                    }
                }
            } else if(Thread.currentThread() == img_animator) {
                debug("In img_animator thread");
                // ========== Image ===============
                if(all_images_loaded == false) {
                    all_images_loaded = checkImgLoadStatus();
                    if(all_images_loaded == true) 
                        updateImageInfo();
                }
                if(all_images_loaded == true) {
                    if(img_index == -1) // First time
                        img_index = img_lower_bound;
                    else
                        img_index = getNextImgIndex(img_index, img_num_images,
                                            img_lower_bound);
                    if(img_pi_clip[img_index] != null)
                        img_pi_clip[img_index].play();
                    repaint();
                    img_sleep(img_delay_runs, img_delay_images,
                                    img_index, img_num_images,
                                    img_lower_bound);
                    if(img_firsttime == true) {
                        img_firsttime = false;
                        if(img_runonce == false) {
                            img_animator.suspend();
                        }
                    }
                } else {
                    img_index = 0;
                    if(img_display_first) {
                        repaint();
                    } else
                        img_index = -1;
                    try {
                        Thread.sleep(1000);
                    } catch (InterruptedException e) {
                        break;
                    }
                }
            }
        }
    }
}
