import java.awt.*;
import java.awt.image.*;
import java.applet.Applet;
import java.awt.MediaTracker;

public class AnimateOnEntry extends Animate implements Runnable {
    boolean within_tx, last_within_tx;
    boolean within_img, last_within_img;
    protected boolean   tx_firsttime = true;
    protected boolean   img_firsttime = true;
    protected boolean   tx_return_start;
    protected boolean   img_return_start;

    public void init() {
        tx_return_start  = GetParm.toBoolean(this, "TxReturnStart", false);
        img_return_start = GetParm.toBoolean(this, "ImgReturnStart", false);
        super.init();
    }

    // getNextTxIndex() is called right before repaint
    public int getNextTxIndex(int current, int max, int min) {
        current++;
        if(current >= max)
            current = min;
        return current;
    }

    // getNextImgIndex() is called right before repaint
    public int getNextImgIndex(int current, int max, int min) {
        current++;
        if(current >= max)
            current = min;
        return current;
    }

    public void tx_sleep(int s_runs, int s_images,
                        int current, int max, int min) {
        try {
            if(current+1 == max)
                Thread.sleep(s_runs);
            else
                Thread.sleep(s_images);
        } catch (InterruptedException e) {}
    }

    // img_sleep() is called right after repaint
    public void img_sleep(int s_runs, int s_images,
                            int current, int max, int min) {
        tx_sleep(s_runs, s_images, current, max, min);
    }

    public boolean mouseMove(Event evt, int x, int y) {
        // Handle the text next (Text takes precedence over images)
        if(tx_index != -1) {
            if(x > tx_x_coord[tx_index] && 
               x < tx_x_coord[tx_index] + tx_width[tx_index] &&
               y > tx_y_coord[tx_index] && 
               y < tx_y_coord[tx_index] + tx_height[tx_index]) {
                within_tx = true;
                if(tx_animator.isAlive() && within_tx == true &&
                   last_within_tx == false) {
                    tx_animator.resume();
                }
            } else {
                if(tx_return_start == true) {
                    tx_index = tx_lower_bound;
                    repaint();
                }
                tx_animator.suspend();
                within_tx = false;
            }
            last_within_tx = within_tx;
        }

        // Handle the image last
        if(img_index != -1) {
            if(x > img_x_coord[img_index] &&
               x < img_x_coord[img_index] + img_width[img_index] &&
               y > img_y_coord[img_index] &&
               y < img_y_coord[img_index] + img_height[img_index]) {
                within_img = true;
                if(img_animator.isAlive() && within_img == true && last_within_img == false) {
                    img_animator.resume();
                }
            } else {
                if(img_return_start == true) {
                    img_index = img_lower_bound;
                    repaint();
                }
                img_animator.suspend();
                within_img = false;
            }
            last_within_img = within_img;
        }
        return super.mouseMove(evt, x, y);
    }

    public boolean mouseExit(Event evt, int x, int y) {
        within_img = false;
        within_tx  = false;
        last_within_img = false;
        last_within_tx = false;
        if(tx_index != -1) {
            if(tx_return_start == true) {
                tx_index = tx_lower_bound;
                repaint();
            }
            tx_animator.suspend();
        }
        if(img_index != -1) {
            if(img_return_start == true) {
                img_index = img_lower_bound;
                repaint();
            }
            img_animator.suspend();
        }
        return super.mouseExit(evt, x, y);
    }

    public void stop() {
        tx_firsttime = true;
        img_firsttime = true;
        super.stop();
    }

    public void run() {
        Thread current_thread = Thread.currentThread();

        //
        // It's important to set the priorities on these
        // threads to some value which is smaller than the
        // priority of the thread that is called to do paint
        Thread.currentThread().setPriority(Thread.MIN_PRIORITY);

        tx_index = -1;
        img_index = -1;

        debug("entering run");
        if(current_thread == img_animator) {
            loadImages();
        }

        if(current_thread == tx_animator) {
            doTxFilters();
        }

        while(true) {
            if(Thread.currentThread() == tx_animator) {
                debug("In tx_animator thread");
                // ========== Text ===============
                if(tx_index == -1)      // First time
                    tx_index = tx_lower_bound;
                else
                    tx_index = getNextTxIndex(tx_index, tx_num_images,
                                            tx_lower_bound);
                if(tx_pi_clip[tx_index] != null)
                    tx_pi_clip[tx_index].play();
                repaint();
                tx_sleep(tx_delay_runs, tx_delay_images,
                                    tx_index, tx_num_images,
                                    tx_lower_bound);
                if(tx_firsttime == true) {
                    tx_firsttime = false;
                    tx_animator.suspend();
                }
            } else if(Thread.currentThread() == img_animator) {
                debug("In img_animator thread");
                // ========== Image ===============
                if(all_images_loaded == false) {
                    all_images_loaded = checkImgLoadStatus();
                    if(all_images_loaded == true) 
                        updateImageInfo();
                }
                if(all_images_loaded == true) {
                    if(img_index == -1) // First time
                        img_index = img_lower_bound;
                    else
                        img_index = getNextImgIndex(img_index, img_num_images,
                                            img_lower_bound);
                    if(img_pi_clip[img_index] != null)
                        img_pi_clip[img_index].play();
                    repaint();
                    img_sleep(img_delay_runs, img_delay_images,
                                    img_index, img_num_images,
                                    img_lower_bound);
                    if(img_firsttime == true) {
                        img_firsttime = false;
                        img_animator.suspend();
                    }
                } else {
                    img_index = 0;
                    if(img_display_first) {
                        repaint();
                    } else
                        img_index = -1;
                    try {
                        Thread.sleep(1000);
                    } catch (InterruptedException e) {
                        break;
                    }
                }
            }
        }
    }
}
