import java.applet.Applet;
import java.applet.AudioClip;
import java.awt.*;

public class SlideShow extends Applet implements Runnable {
    Image image[];
    int imagedisplayed;
    boolean threadSuspended;
    Thread ssthread = null;
    boolean loaded = false;
    int maxwidth, maxheight;
    boolean ImagePrepared;
    int checkImageFlags;
    int num_images = 0;
    String imagename[];
    boolean imageprepared[];
    int width[];
    int height[];
    String soundname[];
    AudioClip sound[];
    AudioClip currentClip;
    int i;
    int current;
    boolean imagecentered[];
    Color       app_bg_color;         // Applet background color
    Panel       controlPanel;
    ssCanvas    imageCanvas;
    boolean     controls;
    Button      nextButton, prevButton;
    Checkbox    audioCB;
    boolean     autoPlay;
    int         delayBetweenImages;
    boolean     playaudio = true;

    public void init() {
        num_images         = GetParm.toInt(this,   "NumImages", 0);
        app_bg_color       = GetParm.toColor(this, "AppBGColor", null);
        controls           = GetParm.toBoolean(this, "Controls", false);
        autoPlay           = GetParm.toBoolean(this, "AutoPlay", false);
        delayBetweenImages = GetParm.toInt(this, "DelayBetweenImages", 5);
        delayBetweenImages *= 1000;

        // If autoPlay is true then force controls off
        if(autoPlay == true)
            controls = false;

        // Set the background and foreground colors, if specified
        if(app_bg_color != null)
            setBackground(app_bg_color);
        else
            app_bg_color = getBackground();
        if(num_images > 0) {
            image         = new Image[num_images];
            imagename     = new String[num_images];
            imageprepared = new boolean[num_images];
            width         = new int[num_images];
            height        = new int[num_images];
            soundname     = new String[num_images];
            sound         = new AudioClip[num_images];
            imagecentered = new boolean[num_images];
        } else {
            System.out.println("You must specify the number of images (NumImages)");
            return;
        }
        for(i=0;i<num_images;i++) {
            imageprepared[i] = false;
            imagecentered[i] = false;
            imagename[i] = GetParm.toString(this, "Image" + (i+1), null);
            if(imagename == null) {
                System.out.println("Missing image " + (i+1));
                return;
            }
            image[i] = getImage(getDocumentBase(), imagename[i]);
            if(i < 2) {
                imageprepared[i] = prepareImage(image[i], this);
            }

            // Get the sounds
            soundname[i] = GetParm.toString(this, "Sound" + (i+1), null);
            if(soundname[i] != null) {
                sound[i] = getAudioClip(getDocumentBase(), soundname[i]);
                if(sound[i] == null) {
                    soundname[i] = null;
                }
            }
        }

        // Make sure we have the first image loaded
        if(imageprepared[0] == false) {
            while(((checkImageFlags = checkImage(image[0],this))&ALLBITS) == 0) {
                try {Thread.sleep(100);} catch (InterruptedException e){}
            }
            imageprepared[0] = true;
        }
        loaded = true;

        // Start loading the other images
        for(i=1;i<num_images;i++) {
            imageprepared[i] = prepareImage(image[i], this);
        }

        // Get width and height of all images
        for(i=0;i<num_images;i++) {
            while((width[i] = image[i].getWidth(this)) == -1) {
                try {Thread.sleep(100);} catch (InterruptedException e){}
            }
            while((height[i] = image[i].getHeight(this)) == -1) {
                try {Thread.sleep(100);} catch (InterruptedException e){}
            }
        }

        maxwidth = Math.max(width[0], width[1]);
        maxheight = Math.max(height[0], height[1]);
        for(i=2;i<num_images;i++) {
            maxwidth = Math.max(maxwidth, width[i]);
            maxheight = Math.max(maxheight, height[i]);
        }
            
        imagedisplayed = 0;

        current = 0;
        setLayout(new BorderLayout());
        imageCanvas = new ssCanvas(this);
        add("Center", imageCanvas);
        if(controls == true) {
            controlPanel = new Panel();
            nextButton = new Button("Next");
            prevButton = new Button("Previous");
            audioCB    = new Checkbox("Audio");
            audioCB.setState(true);
            controlPanel.add(nextButton);
            controlPanel.add(prevButton);
            controlPanel.add(audioCB);
            add("South", controlPanel);
        }
    }

    public void paint(Graphics g) {
        Graphics tmpGC;
        Image tmpimage;

        if(loaded) {
        if(imageprepared[current] == false) {
            while(((checkImageFlags =
                            checkImage(image[current],this))&ALLBITS) == 0) {
                try {Thread.sleep(100);} catch (InterruptedException e){}
            }
            imageprepared[current] = true;
        }
        // Copy the prepared image into a new image of the max size
        if(imagecentered[current] == false) {
            image[current] = centerImage(image[current], maxwidth, maxheight,
                                        width[current], height[current]);
            imagecentered[current] = true;
        }

        // Current Image
        imageCanvas.getGraphics().drawImage(image[current], 0, 0, this);
        }
    }
            
    public Image centerImage(Image image, int maxwidth, int maxheight, 
                                    int width, int height) {
        Image tmpimage;
        Graphics tmpGC;

        if(maxwidth == width && maxheight == height) {
            return image;
        }
        tmpimage = createImage(maxwidth, maxheight);
        tmpGC = tmpimage.getGraphics();
        tmpGC.setColor(getBackground());
        tmpGC.fillRect(0, 0, maxwidth, maxheight);
        tmpGC.drawImage(image, (maxwidth - width)/2,
                          (maxheight - height)/2, this);
        return tmpimage;
    }

    public synchronized boolean imageUpdate(Image img,
                                            int infoflags, 
                                            int x, int y,
                                            int width,
                                            int height){
        int i;
        int cur = -1;

        for(i=0;i<num_images;i++) {
            if(img.equals(image[i])) {
                cur = i;
                break;
            }
        }
        if((infoflags & ABORT) != 0) {
            if(cur != -1)
                System.out.println("Abort image = " + imagename[cur]);
            img.flush();
            img = null;
            return false;
        }
        if((infoflags & ERROR) != 0) {
            if(cur != -1)
                System.out.println("Error getting image = " + imagename[cur]);
            img.flush();
            img = null;
            return false;
        }
        if((infoflags & ALLBITS) != 0) {
            if(cur != -1)
                imageprepared[cur] = true;
            return true;
        }
        return (infoflags & (ALLBITS|ABORT)) == 0;
    }

    public void update(Graphics g) {
        paint(g);
    }

    public void run() {
        Thread.currentThread().setPriority(Thread.MIN_PRIORITY);
        while (ssthread != null) {
            repaint();
            if(currentClip != null)
                currentClip.stop();
            if(sound[current] != null) {
                // Stop the currently playing audio clip
                // If it has already terminated, then calling stop won't hurt
                if(playaudio && sound[current] != null)
                    sound[current].play();
                currentClip = sound[current];
            }
            if(autoPlay == true) {
                try {Thread.sleep(delayBetweenImages);} 
                        catch (InterruptedException e){}
                current++;
                if(current >= num_images)
                    current = 0;
            }
        }
        ssthread = null;
    }

    public void start() {
        if(ssthread == null && autoPlay == true) {
            threadSuspended = false;
            ssthread = new Thread(this);
            ssthread.start();
        }
        if(currentClip != null)
            currentClip.stop();
        if(sound[current] != null) {
            // Stop the currently playing audio clip
            // If it has already terminated, then calling stop won't hurt
            if(playaudio && sound[current] != null)
                sound[current].play();
            currentClip = sound[current];
        }
    }    

    public void stop() {
        if(ssthread != null)
            ssthread.stop();
        ssthread = null;
        if(currentClip != null)
            currentClip.stop();
    }

    public boolean mouseDown(Event evt, int x, int y) {
        if(autoPlay == true)
            return true;
        else {
            if((evt.modifiers & Event.SHIFT_MASK) != 0) {
                current--;
                if(current < 0)
                    current = num_images - 1;
            } else {
                current++;
                if(current >= num_images)
                    current = 0;
            }
            if(currentClip != null)
                currentClip.stop();
            if(sound[current] != null) {
                // Stop the currently playing audio clip
                // If it has already terminated, then calling stop won't hurt
                if(playaudio && sound[current] != null)
                    sound[current].play();
                currentClip = sound[current];
            }
            repaint();
        }
        return true;
    }

    public boolean action(Event evt, Object o) {
        if(autoPlay == true)
            return true;
        else {
            if(evt.target.equals(nextButton)) {
                current++;
                if(current >= num_images)
                    current = 0;
            } else if(evt.target.equals(prevButton)) {
                current--;
                if(current < 0)
                    current = num_images - 1;
            } else if(evt.target instanceof Checkbox) {
                if(audioCB.getState()) {
                    playaudio = true;
                    if(sound[current] != null)
                        sound[current].play();
                    currentClip = sound[current];
                } else {
                    playaudio = false;
                    if(currentClip != null)
                        currentClip.stop();
                }
                // Return so that the audio clip doesn't get played
                return true;
            }
            if(currentClip != null)
                currentClip.stop();
            if(sound[current] != null) {
                // Stop the currently playing audio clip
                // If it has already terminated, then calling stop won't hurt
                if(playaudio && sound[current] != null)
                    sound[current].play();
                currentClip = sound[current];
            }
            repaint();
        }
        return true;
    }

    public void togglethread() {
        if (threadSuspended == true) {
            threadSuspended = false;
            ssthread.resume();
        } else {
            threadSuspended = true;
            ssthread.suspend();
        }
    }

    public boolean keyDown(Event evt, int key) {
        int x = evt.x;
        int y = evt.y;

        if(key == 'F') {
            delayBetweenImages *= .9;
            if(delayBetweenImages < 2)
                delayBetweenImages = 2;
            getAppletContext().showStatus("Setting image delay to: " +
                        delayBetweenImages);
        } else if(key == 'S') {
            if(delayBetweenImages < 10)
                delayBetweenImages++;
            else
                delayBetweenImages *= 1.1;
            getAppletContext().showStatus("Setting image delay to: " +
                        delayBetweenImages);
        } else if(key == 'a') {
            playaudio = !playaudio;
            if(playaudio && currentClip != null)
                currentClip.play();
            if(!playaudio && currentClip != null)
                currentClip.stop();
            if(controls) {
                if(playaudio)
                    audioCB.setState(true);
                else
                    audioCB.setState(false);
            }
        }
        return true;
    }
}

// We need this Canvas in order to get the expose event when
// our window is obscured by another, or deiconified
class ssCanvas extends Canvas {
    Applet applet;

    public ssCanvas(Applet a) {
        applet = a;
    }

    public void paint(Graphics g) {
        applet.paint(g);
    }
}
