import java.applet.Applet;
import java.applet.AudioClip;
import java.awt.*;
import java.awt.image.*;

public class SlideShowFade extends Applet implements Runnable {
    Image image[][];
    int imagedisplayed;
    boolean threadSuspended;
    Thread ssthread = null;
    int maxwidth, maxheight;
    boolean ImagePrepared;
    int checkImageFlags;
    int num_images = 0;
    String imagename[];
    boolean imageprepared[][];
    int width[];
    int height[];
    String soundname[];
    AudioClip sound[];
    AudioClip currentClip;
    int i, j;
    int current, next;
    Color       app_bg_color;         // Applet background color
    Panel       controlPanel;
    Panel       canvasPanel;
    ssCanvas    imageCanvas;
    boolean     controls;
    Button      nextButton, prevButton;
    boolean     autoPlay;
    int         fadesize;
    int         bg_red, bg_green, bg_blue;
    int         index;
    boolean     loaded = false;
    int         delayBetweenFades;
    int         delayBetweenImages;
    boolean     forward = true;
    Checkbox    audioCB;
    boolean     playaudio = true;

    public void init() {
        num_images   = GetParm.toInt(this,   "NumImages", 0);
        app_bg_color = GetParm.toColor(this, "AppBGColor", null);
        controls     = GetParm.toBoolean(this, "Controls", false);
        autoPlay     = GetParm.toBoolean(this, "AutoPlay", false);
        fadesize     = GetParm.toInt(this, "FadeSize", 10);
        delayBetweenFades = GetParm.toInt(this, "DelayBetweenFades", 20);
        delayBetweenImages = GetParm.toInt(this, "DelayBetweenImages", 5);
        delayBetweenImages *= 1000;

        // If autoPlay is true then force controls off
        if(autoPlay == true)
            controls = false;

        // Set the background and foreground colors, if specified
        if(app_bg_color != null)
            setBackground(app_bg_color);
        else
            app_bg_color = getBackground();

        bg_red = getBackground().getRed();
        bg_green = getBackground().getGreen();
        bg_blue = getBackground().getBlue();

        if(num_images > 0) {
            image         = new Image[num_images][fadesize];
            imagename     = new String[num_images];
            imageprepared = new boolean[num_images][fadesize];
            width         = new int[num_images];
            height        = new int[num_images];
            soundname     = new String[num_images];
            sound         = new AudioClip[num_images];
        } else {
            System.out.println("You must specify the number of images (NumImages)");
            return;
        }
        for(i=0;i<num_images;i++) {
            for(j=0;j<fadesize;j++)
                imageprepared[i][j] = false;
            imagename[i] = GetParm.toString(this, "Image" + (i+1), null);
            if(imagename == null) {
                System.out.println("Missing image " + (i+1));
                return;
            }
            image[i][0] = getImage(getDocumentBase(), imagename[i]);

            // Get the sounds
            soundname[i] = GetParm.toString(this, "Sound" + (i+1), null);
            if(soundname[i] != null) {
                sound[i] = getAudioClip(getDocumentBase(), soundname[i]);
                if(sound[i] == null) {
                    soundname[i] = null;
                }
            }
        }


        // Start loading the other images
        for(i=0;i<num_images;i++)
            imageprepared[i][0] = prepareImage(image[i][0], this);

        // Make sure we have the images loaded
        for(i=0;i<num_images;i++) {
            if(imageprepared[i][0] == false) {
                while(((checkImageFlags = checkImage(image[i][0],this))&ALLBITS) == 0) {
                    try {Thread.sleep(delayBetweenFades);} 
                            catch (InterruptedException e){}
                }
                imageprepared[i][0] = true;
            }
        }

        // Get width and height of all images
        for(i=0;i<num_images;i++) {
            while((width[i] = image[i][0].getWidth(this)) == -1) {
                try {Thread.sleep(delayBetweenFades);} 
                            catch (InterruptedException e){}
            }
            while((height[i] = image[i][0].getHeight(this)) == -1) {
                try {Thread.sleep(delayBetweenFades);} 
                            catch (InterruptedException e){}
            }
        }

        maxwidth = Math.max(width[0], width[1]);
        maxheight = Math.max(height[0], height[1]);
        for(i=0;i<num_images;i++) {
            maxwidth = Math.max(maxwidth, width[i]);
            maxheight = Math.max(maxheight, height[i]);
        }

        for(i=0;i<num_images;i++) {
            for(j=1;j<fadesize;j++) {
                ImageFilter colorfilter = new FadeFilter(bg_red, bg_green,
                                        bg_blue, fadesize, fadesize-j-1);
                image[i][j] = createImage(
                    new FilteredImageSource(image[i][0].getSource(),
                                    colorfilter));
                imageprepared[i][j] = prepareImage(image[i][j], this);
            }
        }
        // Make sure we have the images loaded
        for(i=0;i<num_images;i++) {
            for(j=1;j<fadesize;j++) {
                if(imageprepared[i][j] == false) {
                    while(((checkImageFlags = checkImage(image[i][j],this))&ALLBITS) == 0) {
                        try {Thread.sleep(delayBetweenFades);} 
                                catch (InterruptedException e){}
                    }
                    imageprepared[i][0] = true;
                }
            }
        }
        loaded = true;


        imagedisplayed = 0;

        current = 0;
        setLayout(new BorderLayout());
        canvasPanel = new Panel();      // Default to Flow layout
        add("Center", canvasPanel);

        imageCanvas = new ssCanvas(this);
        canvasPanel.add(imageCanvas);
        if(controls == true) {
            controlPanel = new Panel();
            nextButton = new Button("Next");
            prevButton = new Button("Previous");
            audioCB    = new Checkbox("Audio");
            audioCB.setState(true);
            controlPanel.add(nextButton);
            controlPanel.add(prevButton);
            controlPanel.add(audioCB);
            add("South", controlPanel);
        }
    }

    public void paint(Graphics g) {

        //System.out.println("paint ("+current+","+index+")");

        imageCanvas.reshape(0, 0, width[current], height[current]);
        imageCanvas.getGraphics().setColor(getBackground());
        if(width[current] < canvasPanel.size().width)
            imageCanvas.getGraphics().fillRect(width[current], 0,
                        canvasPanel.size().width-width[current],
                        imageCanvas.size().height);
        if(height[current] < canvasPanel.size().height)
            imageCanvas.getGraphics().fillRect(0, height[current],
                        imageCanvas.size().width,
                        canvasPanel.size().height-height[current]);
        if(loaded)
            imageCanvas.getGraphics().drawImage(image[current][index], 0, 0, this);
    }
            
    public synchronized boolean imageUpdate(Image img,
                                            int infoflags, 
                                            int x, int y,
                                            int width,
                                            int height){
        int i, j;
        int cur = -1;
        int curj = -1;

        for(i=0;i<num_images;i++) {
            for(j=0;j<fadesize;j++) {
                if(img.equals(image[i][j])) {
                    cur = i;
                    curj = j;
                    break;
                }
            }
        }
        if((infoflags & ABORT) != 0) {
            if(cur != -1)
                System.out.println("Abort image = " + imagename[cur]);
            img.flush();
            img = null;
            return false;
        }
        if((infoflags & ERROR) != 0) {
            if(cur != -1)
                System.out.println("Error getting image = " + imagename[cur]);
            img.flush();
            img = null;
            return false;
        }
        if((infoflags & ALLBITS) != 0) {
            if(cur != -1)
                imageprepared[cur][curj] = true;
            return true;
        }
        return (infoflags & (ALLBITS|ABORT)) == 0;
    }

    public void update(Graphics g) {
        paint(g);
    }

    public void run() {
        boolean first = true;
        Thread.currentThread().setPriority(Thread.MIN_PRIORITY);
        while (ssthread != null) {
            if(autoPlay) {
                if(first) {
                    first = false;
                    repaint();
                    if(playaudio && sound[current] != null)
                        sound[current].play();
                    currentClip = sound[current];
                    try {Thread.sleep(delayBetweenImages);}
                        catch (InterruptedException e){}
                }

                for(index=0;index<fadesize;index++) {
                    repaint();
                    try {Thread.sleep(delayBetweenFades);}
                            catch (InterruptedException e){}
                }
                current++;
                if(current >= num_images)
                    current = 0;
                // Stop audio now since the image is no longer being displayed
                if(currentClip != null)
                    currentClip.stop();
                if(playaudio && sound[current] != null)
                    sound[current].play();
                currentClip = sound[current];
                for(index=fadesize-1;index>=0;index--) {
                    repaint();
                    try {Thread.sleep(delayBetweenFades);}
                        catch (InterruptedException e){}
                }
                // Set index to 0 in case we get expose event
                index = 0;
                try {Thread.sleep(delayBetweenImages);} 
                        catch (InterruptedException e){}
            } else {
                if(first) {
                    first = false;
                    repaint();
                    if(playaudio && sound[current] != null)
                        sound[current].play();
                    currentClip = sound[current];
                    togglethread();
                }

                for(index=0;index<fadesize;index++) {
                    repaint();
                    try {Thread.sleep(delayBetweenFades);}
                            catch (InterruptedException e){}
                }
                if(forward) {
                    current++;
                    if(current >= num_images)
                        current = 0;
                } else {
                    current--;
                    if(current < 0)
                        current = num_images - 1;
                }
                if(currentClip != null)
                    currentClip.stop();
                if(playaudio && sound[current] != null)
                    // Stop the currently playing audio clip
                    // If it has already terminated, 
                    // then calling stop won't hurt
                    sound[current].play();
                currentClip = sound[current];
                for(index=fadesize-1;index>=0;index--) {
                    repaint();
                    try {Thread.sleep(delayBetweenFades);}
                        catch (InterruptedException e){}
                }
                // Set index to 0 in case we get expose event
                index = 0;
                togglethread();
            }
        }
        ssthread = null;
    }

    public void start() {
        if(ssthread == null) {
            threadSuspended = false;
            ssthread = new Thread(this);
            ssthread.start();
        }
        if(currentClip != null)
            currentClip.stop();
        if(sound[current] != null) {
            // Stop the currently playing audio clip
            // If it has already terminated, then calling stop won't hurt
            if(playaudio && sound[current] != null)
                sound[current].play();
            currentClip = sound[current];
        }
    }    

    public void stop() {
        ssthread = null;
        if(currentClip != null)
            currentClip.stop();
    }

    public boolean mouseDown(Event evt, int x, int y) {

        if(autoPlay == true)
            return true;
        if((evt.modifiers & Event.SHIFT_MASK) != 0) {
            forward = false;
        } else {
            forward = true;
        }
        if(autoPlay == false)
            togglethread();
        return true;
    }

    public boolean action(Event evt, Object o) {
        if(autoPlay == true)
            return true;
        if(evt.target.equals(nextButton)) {
            forward = true;
        } else if(evt.target.equals(prevButton)) {
            forward = false;
        } else if(evt.target instanceof Checkbox) {
            if(audioCB.getState()) {
                playaudio = true;
                if(sound[current] != null)
                    sound[current].play();
                currentClip = sound[current];
            } else {
                playaudio = false;
                if(currentClip != null)
                    currentClip.stop();
            }
            // Return so that we don't toggle the thread for checkbox
            return true;
        }
        if(autoPlay == false)
            togglethread();
        return true;
    }

    public void togglethread() {
        if (threadSuspended == true) {
            threadSuspended = false;
            ssthread.resume();
        } else {
            threadSuspended = true;
            ssthread.suspend();
        }
    }

    public boolean keyDown(Event evt, int key) {
        int x = evt.x;
        int y = evt.y;

        if(key == 'f') {
            delayBetweenFades *= .9;
            if(delayBetweenFades < 2)
                delayBetweenFades = 2;
            getAppletContext().showStatus("Setting delay to: " +
                        delayBetweenFades);
            return true;
        } else if(key == 's') {
            if(delayBetweenFades < 10)
                delayBetweenFades++;
            else
                delayBetweenFades *= 1.1;
            getAppletContext().showStatus("Setting delay to: " +
                        delayBetweenFades);
            return true;
        } else if(key == 'F') {
            delayBetweenImages *= .9;
            if(delayBetweenImages < 2)
                delayBetweenImages = 2;
            getAppletContext().showStatus("Setting delay to: " +
                        delayBetweenImages);
            return true;
        } else if(key == 'S') {
            if(delayBetweenImages < 10)
                delayBetweenImages++;
            else
                delayBetweenImages *= 1.1;
            getAppletContext().showStatus("Setting delay to: " +
                        delayBetweenImages);
            return true;
        } else if(key == 'a') {
            playaudio = !playaudio;
            if(playaudio && currentClip != null)
                currentClip.play();
            if(!playaudio && currentClip != null)
                currentClip.stop();
            if(controls) {
                if(playaudio)
                    audioCB.setState(true);
                else
                    audioCB.setState(false);
            }
        }
        return true;
    }
}

// We need this Canvas in order to get the expose event when
// our window is obscured by another, or deiconified
class ssCanvas extends Canvas {
    Applet applet;

    public ssCanvas(Applet a) {
        applet = a;
    }

    public void paint(Graphics g) {
        applet.paint(g);
    }
}

class FadeFilter extends RGBImageFilter {
    int bg_red, bg_green, bg_blue;
    int total;
    int index;
 
    public FadeFilter() {
        // The filter's operation does not depend on the
        // pixel's location, so IndexColorModels can be
        // filtered directly.
        canFilterIndexColorModel = true;
    }
 
    public FadeFilter(int r, int g, int b, int t, int i) {
        bg_red = r;
        bg_green = g;
        bg_blue = b;
        total = t;
        index = i;
        canFilterIndexColorModel = true;
    }
 
    int count = 0;
 
    public int filterRGB(int x, int y, int rgb) {
        int red, green, blue, alpha;
        count++;
 
        red   = (rgb & 0x00ff0000) >> 16;
        green = (rgb & 0x0000ff00) >> 8;
        blue  = rgb & 0x000000ff;
        alpha = (rgb & 0xff000000) >>> 24;
        if(red > bg_red)
            red = bg_red + (red - bg_red)*index/total;
        else
            red = bg_red - (bg_red - red)*index/total;
        if(green > bg_green)
            green = bg_green + (green - bg_green)*index/total;
        else
            green = bg_green - (bg_green - green)*index/total;
        if(blue > bg_blue)
            blue = bg_blue + (blue - bg_blue)*index/total;
        else
            blue = bg_blue - (bg_blue - blue)*index/total;
 
 
        return ((alpha << 24) | (red << 16) | (green << 8) | blue);
    }
}
