import java.applet.Applet;
import java.applet.AudioClip;
import java.awt.*;

public class SlideShowSwap extends SlideShow {
    int next;
    int r, q;
    int delayBetweenMoves;
    int moveSize;
    boolean imageStill = false;
    boolean forward = true;

    public void init() {
        super.init();
        delayBetweenMoves = GetParm.toInt(this, "DelayBetweenMoves", 20);
        moveSize = GetParm.toInt(this, "MoveSize", 5);
        current = 0;
        next = 1;
    }

    public void paint(Graphics g) {
        Graphics g1, g2;

        if(loaded) {

        // Make sure current image is loaded
        if(imageprepared[current] == false) {
            while(((checkImageFlags =
                            checkImage(image[current],this))&ALLBITS) == 0) {
                try {Thread.sleep(100);} catch (InterruptedException e){}
            }
            imageprepared[current] = true;
        }

        // Make sure next image is loaded
        if(imageprepared[next] == false) {
            while(((checkImageFlags =
                            checkImage(image[next],this))&ALLBITS) == 0) {
                try {Thread.sleep(100);} catch (InterruptedException e){}
            }
            imageprepared[next] = true;
        }

        // Copy the prepared image into a new image of the max size
        if(imagecentered[current] == false) {
            image[current] = centerImage(image[current], maxwidth, maxheight,
                                        width[current], height[current]);
            imagecentered[current] = true;
        }

        // Copy the prepared image into a new image of the max size
        if(imagecentered[next] == false) {
            image[next] = centerImage(image[next], maxwidth, maxheight,
                                        width[next], height[next]);
            imagecentered[next] = true;
        }

        if(forward) {
            // Current Image
            g1 = imageCanvas.getGraphics().create(0, 0, maxwidth, maxheight);
            g1.clipRect(q, 0, maxwidth-q, maxheight);
            g1.drawImage(image[current], r, 0, this);

            // Next Image
            g2 = imageCanvas.getGraphics().create(0, 0, maxwidth, maxheight);
            g2.clipRect(0, 0, q, maxheight);
            g2.drawImage(image[next], -r, 0, this);
        } else {
            // Current Image
            g1 = imageCanvas.getGraphics().create(0, 0, maxwidth, maxheight);
            g1.clipRect(0, 0, maxwidth-q, maxheight);
            g1.drawImage(image[current], -r, 0, this);

            // Next Image
            g2 = imageCanvas.getGraphics().create(0, 0, maxwidth, maxheight);
            g2.clipRect(maxwidth-q, 0, q, maxheight);
            g2.drawImage(image[next], r, 0, this);
        }

        }
    }

    public void run() {
        boolean first = true;
        int sleeptime = delayBetweenMoves;

        Thread.currentThread().setPriority(Thread.MIN_PRIORITY);
        while (ssthread != null) {
            if(first == true) {
                repaint();
                first = false;
                if(playaudio && sound[current] != null)
                    sound[current].play();
                currentClip = sound[current];
                if(autoPlay)
                    try {Thread.sleep(delayBetweenImages);} 
                        catch (InterruptedException e){}
                else {
                    togglethread();
                }
                continue;
            }

            // The switch is complete
            if(r <= 0 && q > 0) {
                r = q = 0;
                // use next because it will soon become the current
                currentClip.stop();
                if(playaudio && sound[next] != null)
                    sound[next].play();
                currentClip = sound[next];
                imageStill = true;
            } else if(q >= (maxwidth/2)) {
                r -= moveSize;
                q += moveSize;
            } else {
                q += moveSize;
                r += moveSize;
            }
            if(autoPlay) {
                if(imageStill == true) {
                    sleeptime = delayBetweenImages;
                } else {
                    sleeptime = delayBetweenMoves;
                }
            }
            if(imageStill == true) {
                if(forward) {
                    current++;
                    if(current >= num_images)
                        current = 0;
                    next = current + 1;
                    if(next >= num_images)
                        next = 0;
                } else {
                    current--;
                    if(current < 0)
                        current = num_images - 1;
                    next = current - 1;
                    if(next < 0)
                        next = num_images - 1;
                }

                imageStill = false;
                if(autoPlay == false)
                    togglethread();
            }
            repaint();
            try {Thread.sleep(sleeptime);} 
                    catch (InterruptedException e){}
        }
        ssthread = null;
    }

    public void start() {
        if(ssthread == null) {
            threadSuspended = false;
            ssthread = new Thread(this);
            ssthread.start();
        }
        if(currentClip != null)
            currentClip.stop();
        if(sound[current] != null) {
            // Stop the currently playing audio clip
            // If it has already terminated, then calling stop won't hurt
            if(playaudio && sound[current] != null)
                sound[current].play();
            currentClip = sound[current];
        }
    }

    public boolean action(Event evt, Object o) {
        if(autoPlay == true)
            return true;
        if(evt.target.equals(nextButton)) {
            forward = true;
            next = current + 1;
            if(next >= num_images)
                next = 0;
        } else if(evt.target.equals(prevButton)) {
            forward = false;
            next = current - 1;
            if(next < 0)
                next = num_images - 1;
        } else if(evt.target instanceof Checkbox) {
            if(audioCB.getState()) {
                playaudio = true;
                if(playaudio && sound[current] != null)
                    sound[current].play();
                currentClip = sound[current];
            } else {
                playaudio = false;
                if(currentClip != null)
                    currentClip.stop();
            }
            // Return so that we don't toggle the thread for checkbox
            return true;
        }
        if(autoPlay == false) {
            togglethread();
        }
        return true;
    }

    public boolean mouseDown(Event evt, int x, int y) {
        int tmp;

        if((evt.modifiers & Event.SHIFT_MASK) != 0) {
            forward = !forward;
            q = maxwidth-q;
            tmp = next;
            next = current;
            current = tmp;
        } else {
            if(forward) {
                next = current + 1;
                if(next >= num_images)
                    next = 0;
            } else {
                next = current - 1;
                if(next < 0)
                    next = num_images - 1;
            }
        }
        if(autoPlay == false)
            togglethread();
        return true;
    }

    public boolean keyDown(Event evt, int key) {
        int x = evt.x;
        int y = evt.y;

        if(key == 'f') {
            delayBetweenMoves *= .9;
            if(delayBetweenMoves < 2)
                delayBetweenMoves = 2;
            getAppletContext().showStatus("Setting delay to: " +
                        delayBetweenMoves);
            return true;
        } else if(key == 's') {
            if(delayBetweenMoves < 10)
                delayBetweenMoves++;
            else
                delayBetweenMoves *= 1.1;
            getAppletContext().showStatus("Setting delay to: " +
                        delayBetweenMoves);
            return true;
        } else if(key == 'm') {
            moveSize += 1;
            getAppletContext().showStatus("New distance: " + moveSize);
            return true;
        } else if(key == 'l') {
            moveSize -= 1;
            if(moveSize <= 0)
                moveSize = 1;
            getAppletContext().showStatus("New distance: " + moveSize);
            return true;
        }
        return super.keyDown(evt, key);
    }
}
