import java.applet.Applet;
import java.applet.AudioClip;

//
// Play an audio file
// This applet just plays an audio 
// No text or graphics are included
//
public class Audio extends Applet implements Runnable {
    AudioClip   clip;           // The Audio Clip
    String      audio_file;     // The name of the audio file
    Thread      audio_thread;   // The audio thread
    int         playdelay;      // Delay in seconds between playing
    boolean     loop;           // If the audio should play in loop

    public void init() {
        // Get the Audio parameter
        audio_file = GetParm.toString(this,  "Audio", null);
        playdelay  = GetParm.toInt(this,     "PlayDelay", 0);
        loop       = GetParm.toBoolean(this, "Loop", false);

        // If the user did not specify an audio file then 
        // print error message and exit
        if(audio_file == null) {
            System.out.println(
                "You must specify the audio " +
                "file as parameter \"Audio\"");
        }

        // If the user specified an non-existent or invalid
        // file, then print error message and exit
        if(audio_file != null) {
            clip = getAudioClip(getDocumentBase(), audio_file);
            if(clip == null) {
                System.out.println(
                    "Error getting " + audio_file + ". " +
                    "Make sure the file exists and is an audio file");
            }
        }
    }

    //
    // run() is called by the start method of Thread.
    // 
    public void run() {
        // If the user specified loop but no playdelay, or they
        // explicitly specified playdelay=0 then loop forever
        if(loop && playdelay == 0)
            clip.loop();
        if(loop) {
            // If the user specified loop and a playdelay then
            // play the audio clip and then sleep the specified
            // number of seconds
            while(audio_thread != null) {
                if(clip != null) {
                    clip.play();
                    try {
                        Thread.sleep(playdelay*1000);
                    } catch (InterruptedException e) {}
                } else
                    audio_thread = null;
            }
        } else {
            // The user did not specify loop so
            // play it once then leave
            if(clip != null)
                clip.play();
        }
    }

    //
    // This method is invoked when the applet is loaded or
    // in other word, when the page on which the applet lives
    // is visited. 
    // Start the thread!
    //
    public void start() {
        if(audio_thread == null) {
            audio_thread = new Thread(this);
            audio_thread.start();
        }
	if(clip != null)
	    clip.stop();
    }
        
    //
    // Stop the thread!
    // This should be called when the the user leaves the page
    // or when the browser is iconified
    //
    public void stop() {
        if(clip != null)
            clip.stop();
	if(audio_thread != null)
	    audio_thread.stop();
        audio_thread = null;
    }
}
