import java.awt.*;
import java.applet.Applet;
import java.applet.AudioClip;

//
// AudioButton is a simple audio applet.  Two buttons are
// created labeled Play and Stop.  When the Play button 
// is pressed the audio begins and runs to completion unless
// one of the buttons is pressed again before completion.
// If the Stop button is pressed then the audio stops.
// If the Play button is pressed then the audio begins 
// again at the beginning.
//
public class AudioButton extends Applet {
    AudioClip   clip;           // The Audio Clip
    Button      play, stop;     // The Play and Stop Buttons
    String      audio_file;     // The name of the audio file
    Color       bg_color;       // The applet background color
    Color       fg_color;       // The applet foreground color
    Label       label;          // The label object
    String      str;            // The string to display for label
    Panel       button_panel;   // The button panel
    Panel       label_panel;    // The label panel
    String      play_label;     // The Play button label
    String      stop_label;     // The Stop button label
    Font        font;           // The font for the label
    Font        buttonfont;     // The font for the buttons

    public void init() {
        bg_color       = GetParm.toColor(this,  "BGColor", null);
        fg_color       = GetParm.toColor(this,  "FGColor", null);
        str            = GetParm.toString(this, "Label", null);
        play_label     = GetParm.toString(this, "PlayLabel", "Play");
        stop_label     = GetParm.toString(this, "StopLabel", "Stop");
        font           = GetParm.toFont(this,   "Font", "Style",
                                    "PointSize");
        buttonfont     = GetParm.toFont(this,   "ButtonFont",
                                    "ButtonStyle", "ButtonPointSize");
        audio_file     = GetParm.toString(this, "Audio", null);


        setFont(font);

        // Set the background and foreground colors, if specified
        if(bg_color != null)
            setBackground(bg_color);
        else
            bg_color = getBackground();
        if(fg_color != null)
            setForeground(fg_color);
        else
            fg_color = getForeground();

        // If the user did not specify an audio file then 
        // print error message
        if(audio_file == null) {
            System.out.println(
                "You must specify the audio " +
                "file as parameter \"Audio\"");
        }

        // If the user specified an non-existent or invalid
        // file, then print error message
        if(audio_file != null) {
            clip = getAudioClip(getDocumentBase(), audio_file);
            if(clip == null) {
                System.out.println(
                    "Error getting " + audio_file + ". " +
                    "Make sure the file exists and is an audio file");
                // Overwrite the str String to display as error label
                str = "Error loading " + audio_file;
            }
        }

        setLayout(new GridLayout(2, 1));
        // Create the Play and Stop buttons
        button_panel = new Panel();
        add(button_panel);
        label_panel = new Panel();
        add(label_panel);
        play = new Button(play_label);
        stop = new Button(stop_label);
        play.setFont(buttonfont);
        stop.setFont(buttonfont);
        button_panel.add(play);
        button_panel.add(stop);
        if(str != null) {
            label = new Label(str);
            label_panel.add(label);
        }
        if(audio_file == null || clip == null) {
            play.disable();
            stop.disable();
        }
    }

    // The action function gets called when either of
    // the two buttons are pressed.  Look at the Object
    // (which contains the label) to determine which
    // button was pressed.
    //
    // If the audio is already plyaing and the 
    // play button is pressed, the audio will
    // begin again from the beginning

    public boolean action(Event evt, Object o) {
        if(evt.target instanceof Button) {
            if(o.equals(play_label))
                clip.play();
            else if(o.equals(stop_label))
                clip.stop();
        }
        return true;
    }

    public void stop() {
        clip.stop();
    }
}
