import java.awt.*;
import java.applet.*;
import java.net.*;
import java.io.*;

//
// AudioImageText is the derived from BasicImage.  It adds
// the functionality of playing an audio clip with
// either the image, the text, or both
//
public class AudioImageText extends BasicImage {
    protected AudioClip tx_clip;        // The Text Audio Clip
    protected String    tx_audio_file;  // The name of the Text audio file
    protected boolean   tx_playing;     // Is the Text clip currently playing
    protected AudioClip img_clip;       // The Image Audio Clip
    protected String    img_audio_file; // The name of the Image audio file
    protected boolean   img_playing;    // Is the Image clip currently playing

    public void init() {
        super.init();
        tx_audio_file = GetParm.toString(this, "TxAudio", null);
        img_audio_file = GetParm.toString(this, "ImgAudio", null);

        // If the user specified an non-existent or invalid
        // file, then print error message and exit
        if(tx_audio_file != null) {
            tx_clip = getAudioClip(getDocumentBase(), tx_audio_file);
            if(tx_clip == null) {
                System.out.println(
                    "Error getting " + tx_audio_file + ". " +
                    "Make sure the file exists and is an audio file");
            }
        }
        // If the user specified an non-existent or invalid
        // file, then print error message and exit
        if(img_audio_file != null) {
            img_clip = getAudioClip(getDocumentBase(), img_audio_file);
            if(img_clip == null) {
                System.out.println(
                    "Error getting " + img_audio_file + ". " +
                    "Make sure the file exists and is an audio file");
            }
        }
    }

    //
    // Check the x and y coordinates to see if we're actually
    // inside the boundaries of the text or image or both
    //
    public boolean mouseMove(Event evt, int x, int y) {
        // Process the image first
        if(x > img_x_coord && x < img_x_coord + img_width &&
           y > img_y_coord && y < img_y_coord + img_height) {
            if(!img_playing) {
                img_playing = true;
                if(img_clip != null)
                    img_clip.play();
            } // else it's already playing, leave it alone
        } else {
            // Since we don't get any kind of notification when the audio
            // clip has terminated (due to playing to completion) we set
            // playing to false everytime we move out of the boundaries
            // of the string
            img_playing = false;
            if(img_clip != null) {
                img_clip.stop();
            }
        }

        // Process the text last
        if(x > tx_x_coord && x < tx_x_coord + tx_width &&
           y < tx_y_coord+tx_descent && y > tx_y_coord - tx_ascent) {
            if(!tx_playing) {
                tx_playing = true;
                if(tx_clip != null)
                    tx_clip.play();
            } // else it's already playing, leave it alone
        } else {
            // Since we don't get any kind of notification when the audio
            // clip has terminated (due to playing to completion) we set
            // playing to false everytime we move out of the boundaries
            // of the string
            tx_playing = false;
            if(tx_clip != null) {
                tx_clip.stop();
            }
        }
        return true;
    }

    //
    // Just to be sure, when we exit the applet, stop the clips
    // It should already be stopped
    //
    public boolean mouseExit(Event evt, int x, int y) {
        // Just to be sure
        tx_playing = false;
        if(tx_clip != null)
            tx_clip.stop();
        img_playing = false;
        if(img_clip != null)
            img_clip.stop();
        return true;
    }
}
