import java.awt.*;
import java.applet.*;
import java.net.*;
import java.io.*;

//
// AudioText displays a string and also plays
// an audio file when the pointer enters the text.
// If the pointer moves out of the text before the
// audio has finished playing then the audio is terminated.
//
public class AudioText extends BasicText {
    protected AudioClip   tx_clip;           // The Audio Clip
    protected String      tx_audio_file;     // The name of the audio file
    protected boolean     tx_playing;        // Is the clip currently playing

    public void init() {
        super.init();
        tx_audio_file = GetParm.toString(this, "TxAudio", null);

        // If the user specified an non-existent or invalid
        // file, then print error message and exit
        if(tx_audio_file != null) {
            tx_clip = getAudioClip(getDocumentBase(), tx_audio_file);
            if(tx_clip == null) {
                System.out.println(
                    "Error getting " + tx_audio_file + ". " +
                    "Make sure the file exists and is an audio file");
            }
        }
    }

    //
    // Check the x and y coordinates to see if we're actually
    // inside the boundaries of the text
    //
    public boolean mouseMove(Event evt, int x, int y) {
        if(x > tx_x_coord && x < tx_x_coord + tx_width &&
           y < tx_y_coord+tx_descent && y > tx_y_coord - tx_ascent) {
            if(!tx_playing) {
                tx_playing = true;
                if(tx_clip != null)
                    tx_clip.play();
            } // else it's already playing, leave it alone
        } else {
            // Since we don't get any kind of notification when the audio
            // clip has terminated (due to playing to completion) we set
            // playing to false everytime we move out of the boundaries
            // of the string
            tx_playing = false;
            if(tx_clip != null) {
                tx_clip.stop();
            }
        }
        return true;
    }

    //
    // Just to be sure, when we exit the applet, stop the clip
    // It should already be stopped
    //
    public boolean mouseExit(Event evt, int x, int y) {
        tx_playing = false;
        if(tx_clip != null)
            tx_clip.stop();
        return true;
    }
}
