import java.awt.*;
import java.applet.Applet;
import java.net.*;
import java.io.*;

//
// BasicImage is the derived from BasicText.  It adds
// the functionality of displaying an image in addition
// to text.
//
public class BasicImage extends BasicText {
    protected Image       img;                  // The image
    protected String      img_file;             // Image file name
    protected ImgGetr     img_gettr;            // The Image Getter class
    protected boolean     img_wait_load;        // Load flag
    protected int         img_width;            // Image width
    protected int         img_height;           // Image height
    protected int         img_xoffset;          // Image X offset
    protected int         img_yoffset;          // Image Y offset
    protected int         img_y_coord;          // Image X coordinate
    protected int         img_x_coord;          // Image Y coordinate
    protected boolean     img_horizcenter;      // Image horizontally centered
    protected boolean     img_vertcenter;       // Image vertically centered
    protected int         img_borderwidth;      // Image border width
    protected Color       img_bordercolor;      // Image border color
    protected int         img_bordermargin;     // Image border margin
    protected String      img_fr_type;          // Image frame type
    protected int         img_fr_thick;         // Image frame thickness
    protected int         img_fr_margin;        // Image frame margin

    Font                  loading_font;

    public void init() {
        super.init();
        img_file        = GetParm.toString(this,  "Image", null);
        img_wait_load   = GetParm.toBoolean(this, "ImgLoadWait", false);
        img_xoffset     = GetParm.toInt(this,     "ImgXOffset", 0);
        img_yoffset     = GetParm.toInt(this,     "ImgYOffset", -1);
        img_horizcenter = GetParm.toBoolean(this, "ImgHorizCenter", false);
        img_vertcenter  = GetParm.toBoolean(this, "ImgVertCenter",  false);
        img_borderwidth = GetParm.toInt(this,     "ImgBorderWidth", 0);
        img_bordercolor = GetParm.toColor(this,   "ImgBorderColor",
                                                   Color.black);
        img_bordermargin= GetParm.toInt(this,     "ImgBorderMargin", 0);
        img_fr_thick    = GetParm.toInt(this,     "ImgFrameThickness", 0);
        img_fr_type     = GetParm.toString(this,  "ImgFrameType", "ShadowIn");
        img_fr_margin   = GetParm.toInt(this,     "ImgFrameMargin", 0);


        if(img_file != null) {
            img_gettr     = new ImgGetr(this);
            if(img_wait_load)
                repaint();
            img = img_gettr.process(getDocumentBase(), img_file, false);
            img_gettr.prepareImage(img, this);
            img_width = img_gettr.getWidth(img, this);
            img_height = img_gettr.getHeight(img, this);
        }
    }

    //
    // The paint method draw the image, and draw the border and/or
    // frame if there is one.
    // If the user has specified ImgLoadWait then display a message
    // on the screen while waiting.
    //
    public void paint(Graphics g) {
        if(img != null) {

            img_x_coord = calcULXcoord(img_horizcenter, size().width,
                                            img_width, img_xoffset);
            img_y_coord = calcULYcoord(img_vertcenter, size().height,
                                            img_height, img_yoffset);

            if(img_wait_load && !img_gettr.isImagePrepared(img, this)) {
                if(loading_font == null)
                    loading_font = new Font("Dialog", Font.BOLD, 14);
                g.setFont(loading_font);
                g.drawString("Loading " + img_file + "...", 
                                img_x_coord,
                                img_y_coord + (img_height/2));
            } else if(img_wait_load && img_gettr.isImagePrepared(img, this)) {
                // This is called one time when the img is finally
                // loaded.  Clear the background
                g.setColor(app_bg_color);
                g.fillRect(0, 0, size().width, size().height);

                g.drawImage(img, img_x_coord, img_y_coord, this);
                drawFrame(g, img_x_coord, img_y_coord, img_width,
                                img_height, img_fr_type,
                                img_fr_thick, img_fr_margin);
                drawBorder(g, img_x_coord, img_y_coord, img_width,
                                img_height, img_borderwidth,
                                img_bordermargin, img_bordercolor);

                img_wait_load = false;        // Because it's already loaded
            } else {
                g.drawImage(img, img_x_coord, img_y_coord, this);
                drawFrame(g, img_x_coord, img_y_coord, img_width,
                                img_height, img_fr_type,
                                img_fr_thick, img_fr_margin);
                drawBorder(g, img_x_coord, img_y_coord, img_width,
                                img_height, img_borderwidth,
                                img_bordermargin, img_bordercolor);
            }
        }
        // Call the superclass's paint method. This should draw the text if any
        // We call super.paint after drawing the image so that the text can
        // be place "on top of" the image, if desired
        super.paint(g);
    }

    public synchronized boolean imageUpdate(Image img,
					    int infoflags, 
					    int x, int y,
					    int width,
					    int height){
	if((infoflags & ERROR) != 0) {
	    System.out.println("Error getting image = " + img_file);
	    img.flush();
	    img = null;
	    return false;
	}
	if((infoflags & ABORT) != 0) {
	    System.out.println("Abort image = " + img_file);
	    img.flush();
	    img = null;
	    return false;
	}
	if((infoflags & ALLBITS) != 0) {
	    return true;
	}
	return super.imageUpdate(img, infoflags, x, y, width, height);
    }
}
