import java.awt.*;
import java.applet.Applet;
import java.net.*;
import java.io.*;

//
// BasicText is the base class applet upon which most of the
// applets in the Fundamentals chapter are based.
// Basic handles lots of generic things like drawing
// borders, frames, and underline.  Many of the basic
// parameters are gotten here.
//
public class BasicText extends Applet {
    // Parameters that apply to the entire applet
    protected int         app_borderwidth;      // Applet border width
    protected Color       app_bordercolor;      // Applet border color
    protected String      app_fr_type;          // Applet frame type
    protected int         app_fr_thick;         // Applet frame thickness
    protected Color       app_bg_color;         // Applet background color

    // Parameters specific to Text
    protected String      tx;                   // Text
    protected int         tx_xoffset;           // Text X offset
    protected int         tx_yoffset;           // Text Y offset
    protected Color       tx_color;             // Text color
    protected boolean     tx_horizcenter;       // Text horizontally centered
    protected boolean     tx_vertcenter;        // Text vertically centered
    protected Font        tx_font;              // Text font
    protected FontMetrics tx_fontmetrics;       // Text font metrics
    protected boolean     tx_underline;         // Text underline
    protected int         tx_width;             // Text width
    protected int         tx_height;            // Text height
    protected int         tx_ascent;            // Text ascent
    protected int         tx_descent;           // Text descent
    protected int         tx_borderwidth;       // Text border width
    protected Color       tx_bordercolor;       // Text border color
    protected int         tx_bordermargin;      // Text border margin
    protected String      tx_fr_type;           // Text frame type
    protected int         tx_fr_thick;          // Text frame thickness
    protected int         tx_fr_margin;         // Text frame margin
    protected int         tx_x_coord;           // Text X Coordinate
    protected int         tx_y_coord;           // Text Y Coordinate

    public void init() {
        // Applet parameters
        app_borderwidth = GetParm.toInt(this,     "AppBorderWidth", 0);
        app_bordercolor = GetParm.toColor(this,   "AppBorderColor", Color.black);
        app_fr_type     = GetParm.toString(this,  "AppFrameType", "ShadowIn");
        app_fr_thick    = GetParm.toInt(this,     "AppFrameThickness", 0);
        app_bg_color    = GetParm.toColor(this,   "AppBGColor", null);

        // Text parameters
        tx              = GetParm.toString(this, "Text", "");
        tx_xoffset      = GetParm.toInt(this,     "TxXOffset", 0);
        tx_yoffset      = GetParm.toInt(this,     "TxYOffset", -1);
        tx_color        = GetParm.toColor(this,   "TxColor", null);
        tx_horizcenter  = GetParm.toBoolean(this, "TxHorizCenter", false);
        tx_vertcenter   = GetParm.toBoolean(this, "TxVertCenter",  false);
        tx_font         = GetParm.toFont(this,    "TxFont",
                                                  "TxStyle",
                                                  "TxPointSize");
        tx_underline    = GetParm.toBoolean(this, "TxUnderLine", false);
        tx_borderwidth  = GetParm.toInt(this,     "TxBorderWidth", 0);
        tx_bordercolor  = GetParm.toColor(this,   "TxBorderColor", Color.black);
        tx_bordermargin = GetParm.toInt(this,     "TxBorderMargin", 0);
        tx_fr_thick     = GetParm.toInt(this,     "TxFrameThickness", 0);
        tx_fr_type      = GetParm.toString(this,  "TxFrameType", "ShadowIn");
        tx_fr_margin    = GetParm.toInt(this,     "TxFrameMargin", 0);

        setFont(tx_font);
        tx_fontmetrics = getFontMetrics(tx_font);

        tx_width  = tx_fontmetrics.stringWidth(tx);
        tx_height = tx_fontmetrics.getHeight();
        tx_ascent = tx_fontmetrics.getMaxAscent();
        tx_descent = tx_fontmetrics.getDescent();

        // Set the background and foreground colors, if specified
        if(app_bg_color != null)
            setBackground(app_bg_color);
        else
            app_bg_color = getBackground();
        if(tx_color != null)
            setForeground(tx_color);
        else
            tx_color = getForeground();
    }

    //
    // Overwrite the default update method
    // This may not be necessary for this case
    // but it certainly doesn't hurt
    //
    public void update(Graphics g) {
        paint(g);
    }

    //
    // Our paint method handles several tasks:
    //    Calculates the x and y coordinates base on params
    //    Draws the applet frame
    //    Draws the applet border
    //    Draws the text frame
    //    Draws the text border
    //    Draws the underline
    //    Draws the text
    //
    public void paint(Graphics g) {
        // We never know when the size might change so if horizcenter
        // is true then recalculate the x_coord every time
        tx_x_coord = calcULXcoord(tx_horizcenter, size().width, tx_width,
                                        tx_xoffset);
        tx_y_coord = calcLLYcoord(tx_vertcenter, size().height, tx_height,
                                        tx_descent, tx_yoffset);

        // Draw frame for applet
        drawFrame(g, 0, 0, size().width, size().height,
                            app_fr_type, app_fr_thick, 0);
        // Draw border for applet
        drawBorder(g, 0, 0, size().width, size().height,
                            app_borderwidth, 0, app_bordercolor);
        // Draw frame for text
        drawFrame(g, tx_x_coord, tx_y_coord-tx_ascent,
                            tx_width, tx_height,
                            tx_fr_type, tx_fr_thick, tx_fr_margin);
        // Draw border for text
        drawBorder(g, tx_x_coord, tx_y_coord-tx_ascent,
                            tx_width, tx_height,
                            tx_borderwidth, tx_bordermargin, tx_bordercolor);
        drawUnderline(g, tx_x_coord, tx_y_coord+Math.max(1, (tx_descent/4)),
                            tx_width, tx_color);
        g.setColor(tx_color);
        g.drawString(tx, tx_x_coord, tx_y_coord);
    }

    //
    // Calculate the upper left corner position when centering
    // an item that's drawn in relation to its upper left corner
    // such as images
    //
    protected int calcULCenter(int panelsize, int objectsize) {
        return ((panelsize / 2) - (objectsize / 2));
    }

    //
    // Calculate the lower left corner position when centering
    // an item that's drawn in relation to its lower left corner
    // such as text
    //
    protected int calcLLCenter(int panelsize, int objectsize) {
        return ((panelsize / 2) + (objectsize / 2));
    }

    //
    // A general purpose text drawing method
    // This method is used by the Img class
    //
    protected void drawText(Graphics g, String txt, int x, int y,
                                            Color c, Font f) {
        if(txt != null) {
            g.setColor(c);
            g.setFont(f);
            g.drawString(txt, x, y);
        }
        
    }

    //
    // The drawBorder method draws a border at the 
    // specified x and y position
    //
    protected void drawBorder(Graphics g,
                           int x,
                           int y,
                           int w,
                           int h,
                           int borderwidth,
                           int margin,
                           Color c) {
        if(borderwidth > 0) {
            g.setColor(c);
            for(int i=0;i<borderwidth;i++) {
                g.drawRect(x-margin+i,
                           y-margin+i,
                           w+(margin*2)-(i*2)-1,
                           h+(margin*2)-(i*2)-1);
            }
        }
    }

    void drawTopLine(Graphics g,
                    int x, int y, int w, int h, int i, int margin) {
        g.drawLine(x-margin+i, y-margin+i, x+w+margin-i-1, y-margin+i);
    }

    void drawBottomLine(Graphics g,
                    int x, int y, int w, int h, int i, int margin) {
        g.drawLine(x-margin+i, y+h+margin-i-1, x+w+margin-i-1, y+h+margin-i-1);
    }

    void drawLeftLine(Graphics g,
                    int x, int y, int w, int h, int i, int margin) {
        g.drawLine(x-margin+i, y-margin+i, x-margin+i, y+h+margin-i-1);
    }

    void drawRightLine(Graphics g,
                    int x, int y, int w, int h, int i, int margin) {
        g.drawLine(x+w+margin-i-1, y-margin+i, x+w+margin-i-1, y+h+margin-i-1);
    }
        

    //
    // The drawFrame draws a frame at the specified x and y position
    // A frame can be one of the following four types: ShadowIn,
    // ShadowOut, ShadowEtchedIn, ShadowEtchedOut
    //
    protected void drawFrame(Graphics g,    // Graphics to draw to
                           int x,           // X upper left position
                           int y,           // Y upper left position
                           int w,           // width
                           int h,           // height
                           String type,     // Type of frame
                           int thickness,   // thickness of frame
                           int margin) {    // Margin around object
        int i;
        if(thickness == 0)
            return;
        Color darker = AdjColor.darkenit(app_bg_color, .50);
        Color slightlydarker = AdjColor.darkenit(app_bg_color, .10);
        Color brighter = AdjColor.brightenit(app_bg_color, .50);

        if(thickness > 0) {
            if(type.equalsIgnoreCase("shadowout")) {
                for(i=0;i<thickness;i++) {
                    g.setColor(brighter);
                    // TOP
                    drawTopLine(g, x, y, w, h, i, margin);
                    // LEFT
                    drawLeftLine(g, x, y, w, h, i, margin);

                    g.setColor(darker);
                    // BOTTOM
                    drawBottomLine(g, x, y, w, h, i, margin);
                    // RIGHT
                    drawRightLine(g, x, y, w, h, i, margin);
                }
            } else if(type.equalsIgnoreCase("shadowetchedin")) {
                for(i=0;i<thickness;i++) {
                    if(i == 0)
                        g.setColor(darker);
                    else if (i == thickness-1)
                        g.setColor(brighter);
                    else
                        g.setColor(slightlydarker);
                    // TOP
                    drawTopLine(g, x, y, w, h, i, margin);
                    // LEFT
                    drawLeftLine(g, x, y, w, h, i, margin);

                    if(i == 0)
                        g.setColor(brighter);
                    else if (i == thickness-1)
                        g.setColor(darker);
                    else
                        g.setColor(slightlydarker);
                    // BOTTOM
                    drawBottomLine(g, x, y, w, h, i, margin);
                    // RIGHT
                    drawRightLine(g, x, y, w, h, i, margin);
                }

            } else if(type.equalsIgnoreCase("shadowetchedout")) {
                for(i=0;i<thickness;i++) {
                    if(i == 0)
                        g.setColor(brighter);
                    else if (i == thickness-1)
                        g.setColor(darker);
                    else
                        g.setColor(app_bg_color);
                    // TOP
                    drawTopLine(g, x, y, w, h, i, margin);
                    // LEFT
                    drawLeftLine(g, x, y, w, h, i, margin);

                    if(i == 0)
                        g.setColor(darker);
                    else if (i == thickness-1)
                        g.setColor(brighter);
                    else
                        g.setColor(app_bg_color);
                    // BOTTOM
                    drawBottomLine(g, x, y, w, h, i, margin);
                    // RIGHT
                    drawRightLine(g, x, y, w, h, i, margin);
                }

            } else {    // Default to "shadowin"
                for(i=0;i<thickness;i++) {
                    g.setColor(darker);
                    // TOP
                    drawTopLine(g, x, y, w, h, i, margin);
                    // LEFT
                    drawLeftLine(g, x, y, w, h, i, margin);

                    g.setColor(brighter);
                    // BOTTOM
                    drawBottomLine(g, x, y, w, h, i, margin);
                    // RIGHT
                    drawRightLine(g, x, y, w, h, i, margin);
                }
            }
        }
    }

    //
    // The drawUnderline method draws an underline
    // The underline is actually a filled rectangle
    // whose size is based on the width of a pipe
    // character from the same font.  It works!
    //
    protected void drawUnderline(Graphics g,
                           int x,
                           int y,
                           int w,
                           Color c) {
        if(tx_underline == true) {
            int pipewidth = tx_fontmetrics.charWidth('|');;
            g.setColor(c);
            g.fillRect(x, y, w, Math.max(1, pipewidth/4));
        }
    }

    //
    // Calculate the upper left X coordinate based
    // on parameters such as horizcenter and xoffset
    //
    protected int calcULXcoord(boolean hcent, int appletwidth,
                                int objectwidth, int xoffset) {
        int x;

        // Horizontal (x coordinate) Position
        if(hcent)
            x = calcULCenter(appletwidth, objectwidth);
        else if (xoffset == -1)
            x = 0;
        else
            x = xoffset;
        
        return x;
    }

    //
    // Calculate the upper left Y coordinate based
    // on parameters such as vertcenter and yoffset
    //
    protected int calcULYcoord(boolean hcent, int appletheight,
                            int objectheight, int yoffset) {
        int y;

        // Vertical (y coordinate) Position
        if(hcent)
            y = calcULCenter(appletheight, objectheight);
        else if (yoffset == -1)
            y = 0;
        else
            y = yoffset;

        return y;
    }

    //
    // Calculate the lower left Y coordinate based
    // on parameters such as vertcenter and yoffset
    //
    protected int calcLLYcoord(boolean hcent, int appletheight,
                    int objectheight, int objectdescent, int yoffset) {
        int y;

        // Vertical (y coordinate) Position
        if(hcent)
            y = calcLLCenter(appletheight, objectheight) - objectdescent;
        else if (yoffset == -1)
            y = objectheight - objectdescent;
        else
            y = yoffset;

        return y;
    }
}
