import java.awt.*;
import java.applet.*;
import java.net.*;
import java.io.*;

//
// ImageMap does everything thing that URLAudImgTxtBG does
// but in addition is provides an "image map" capability
//
public class ImageMap extends URLAudImgTxtBG {
    protected int       num_maps;       // Number of Maps
    protected int       map_x1[];       // X1 coord for map
    protected int       map_y1[];       // Y1 coord for map
    protected int       map_x2[];       // X2 coord for map
    protected int       map_y2[];       // Y2 coord for map
    protected URL       map_url[];      // URLs for maps
    protected boolean   testmode;       // Test mode
    

    public void init() {
        int i;

        super.init();
        testmode = GetParm.toBoolean(this, "TestMode",   false);
        num_maps = GetParm.toInt(this,     "AppNumMaps", 0);

        if(num_maps > 0) {
            map_x1  = new int[num_maps];
            map_y1  = new int[num_maps];
            map_x2  = new int[num_maps];
            map_y2  = new int[num_maps];
            map_url = new URL[num_maps];
        }

        for(i=0;i<num_maps;i++) {
            map_x1[i]  = GetParm.toInt(this, "Map"+(i+1)+"_X1",  0);
            map_y1[i]  = GetParm.toInt(this, "Map"+(i+1)+"_Y1",  0);
            map_x2[i]  = GetParm.toInt(this, "Map"+(i+1)+"_X2",  0);
            map_y2[i]  = GetParm.toInt(this, "Map"+(i+1)+"_Y2",  0);
            map_url[i] = GetParm.toURL(this, "Map"+(i+1)+"_URL");
        }
    }    

    //
    // If the user clicks the mouse then load the URL
    // ImageMap url take precedence over urls of text
    // or image
    //
    // If we find an image entry then use it and return
    //
    public boolean mouseDown(Event evt, int x, int y) {
        // Handle the text first (Text takes precedence over images)
        int i;
        for(i=0;i<num_maps;i++) {
            if(x > map_x1[i] && x < map_x2[i] &&
               y > map_y1[i] && y < map_y2[i]) {
                if(map_url[i] != null) {
                    if(testmode) {
                        getAppletContext().showStatus("Loading URL: " +
                                    map_url[i].toString());
                        System.out.println("Loading URL: " +
                                    map_url[i].toString());
                    } else
                        getAppletContext().showDocument(map_url[i]);
                    return true;
                }
            }
        }    
        super.mouseDown(evt, x, y);
        return true;
    }

    //
    // The ImageMap urls take precedence over text and image
    // urls.
    //
    public boolean mouseMove(Event evt, int x, int y) {
        int i;

        if(testmode) {
            getAppletContext().showStatus("X: " + x + " Y: " + y);
            return true;
        }
        tx_status_shown = false;
        for(i=0;i<num_maps;i++) {
            if(x > map_x1[i] && x < map_x2[i] &&
               y > map_y1[i] && y < map_y2[i]) {
                if(map_url[i] != null) {
                    getAppletContext().showStatus(map_url[i].toString());
                    tx_status_shown = true;
                }
            }
        }

        //
        // If a image map url was found (hence status_shown == true)
        // then return and don't bother to call the parent mouseMove
        // which would simply display the url for the text or image.
        //
        if(tx_status_shown == true) {
            return true;
        }

        // The parent will handle clearing the showStatus line
        super.mouseMove(evt, x, y);
        return true;
    }
}
