import java.awt.*;
import java.applet.*;
import java.net.*;
import java.io.*;

//
// URLAudImgTxt does everything that AudImTx does
// plus it loads a URL when a mouse button is
// pressed with the pointer over the text or
// image. 
//
public class URLAudImgTxt extends AudioImageText {
    protected URL               tx_url;         // The Text URL
    protected URL               img_url;        // The Image URL
    protected boolean           tx_status_shown;
    protected boolean           img_status_shown;

    public void init() {
        super.init();

        tx_url     = GetParm.toURL(this, "TxURL");
        img_url    = GetParm.toURL(this, "ImgURL");

        tx_status_shown  = false;
        img_status_shown = false;
    }

    //
    // If the user clicks the mouse then load the URL
    // We check the text position first, so if there is
    // overlapping text and image then the text URL will
    // be loaded, not the imgae URL
    //
    public boolean mouseDown(Event evt, int x, int y) {
        // Handle the text first (Text takes precedence over images)
        if(x > tx_x_coord && x < tx_x_coord + tx_width &&
           y < tx_y_coord+tx_descent && y > tx_y_coord - tx_ascent) {
            if(tx_url != null) {
                getAppletContext().showDocument(tx_url);
                System.out.println("loading " + tx_url.toString());
                return true;
            }
        }    
        // Handle the image last
        if(x > img_x_coord && x < img_x_coord + img_width &&
           y > img_y_coord && y < img_y_coord + img_height) {
            if(img_url != null) {
                getAppletContext().showDocument(img_url);
                System.out.println("loading " + img_url.toString());
                return true;
            }
        }    
        return true;
    }

    //
    // Inherit the mouseMove behavior from AudioImageText
    // In addition, update the status line by calling
    // showStatus.  This indicates to the user when
    // they will go if they click the mouse
    //
    public boolean mouseMove(Event evt, int x, int y) {
        int i;

        super.mouseMove(evt, x, y);
        // Handle the text first
        if(x > tx_x_coord && x < tx_x_coord + tx_width &&
           y < tx_y_coord+tx_descent && y > tx_y_coord - tx_ascent) {
            if(tx_url != null) {
                getAppletContext().showStatus(tx_url.toString());
                tx_status_shown = true;
            }
        }    

        // Handle the image last
        // If a Text URL is being displayed then don't
        // display the image URL
        if(x > img_x_coord && x < img_x_coord + img_width &&
           y > img_y_coord && y < img_y_coord + img_height) {
            if(img_url != null && tx_status_shown == false) {
                getAppletContext().showStatus(img_url.toString());
                img_status_shown = true;
            }
        }    
        // If we are not over either text or image that has 
        // a corresponding URL then clear the status line
        if(tx_status_shown == false && img_status_shown == false)
            getAppletContext().showStatus(null);
        return true;
    }


    // When the mouse exits the applet, make sure the
    // status line is clear
    public boolean mouseExit(Event evt, int x, int y) {
        super.mouseExit(evt, x, y);
        getAppletContext().showStatus(null);
        return true;
    }
}
