import java.awt.*;
import java.awt.image.*;
import java.applet.*;
import java.net.*;
import java.io.*;

//
// BasicTxt is the base class applet upon which most of the
// applets in the Fundamentals chapter are based.
// Basic handles lots of generic things like drawing
// borders, frames, and underline.  Many of the basic
// parameters are gotten here.
//
public class OffScr extends Applet {
    // Parameters that apply to the entire applet
    protected int         app_borderwidth;      // Applet border width
    protected Color       app_bordercolor;      // Applet border color
    protected String      app_fr_type;          // Applet frame type
    protected int         app_fr_thick;         // Applet frame thickness
    protected Color       app_bg_color;         // Applet background color

    // Parameters specific to Text
    protected String      tx;                   // Text
    protected int         tx_xoffset;           // Text X offset
    protected int         tx_yoffset;           // Text Y offset
    protected Color       tx_color;             // Text color
    protected Color       tx_bg_color;          // Text background color
    protected boolean     tx_horizcenter;       // Text horizontally centered
    protected boolean     tx_vertcenter;        // Text vertically centered
    protected Font        tx_font;              // Text font
    protected FontMetrics tx_fontmetrics;       // Text font metrics
    protected boolean     tx_underline;         // Text underline
    protected int         tx_width;             // Text width
    protected int         tx_height;            // Text height
    protected int         tx_ascent;            // Text ascent
    protected int         tx_descent;           // Text descent
    protected int         tx_borderwidth;       // Text border width
    protected Color       tx_bordercolor;       // Text border color
    protected int         tx_bordermargin;      // Text border margin
    protected String      tx_fr_type;           // Text frame type
    protected int         tx_fr_thick;          // Text frame thickness
    protected int         tx_fr_margin;         // Text frame margin
    protected int         tx_x_coord;           // Text X Coordinate
    protected int         tx_y_coord;           // Text Y Coordinate

    // Parameters specific to the Image
    protected String      img_file;             // Image file name
    protected ImgGetr     img_gettr;            // The Image Getter class
    protected boolean     img_wait_load;        // Load flag
    protected int         img_width;            // Image width
    protected int         img_height;           // Image height
    protected int         img_xoffset;          // Image X offset
    protected int         img_yoffset;          // Image Y offset
    protected int         img_y_coord;          // Image X coordinate
    protected int         img_x_coord;          // Image Y coordinate
    protected boolean     img_horizcenter;      // Image horizontally centered
    protected boolean     img_vertcenter;       // Image vertically centered
    protected int         img_borderwidth;      // Image border width
    protected Color       img_bordercolor;      // Image border color
    protected int         img_bordermargin;     // Image border margin
    protected String      img_fr_type;          // Image frame type
    protected int         img_fr_thick;         // Image frame thickness
    protected int         img_fr_margin;        // Image frame margin
    Font                  loading_font;

    // Parameters specific to Audio
    protected AudioClip    tx_clip;             // The Text Audio Clip
    protected String       tx_audio_file;       // The Text audio file
    protected boolean      tx_playing;          // Text clip currently playing
    protected AudioClip    img_clip;            // The Image Audio Clip
    protected String       img_audio_file;      // The Image audio file
    protected boolean      img_playing;         // Image clip currently playing

    // Parameters specific to URLs
    protected URL          tx_url;              // The Text URL
    protected URL          img_url;             // The Image URL

    protected boolean      tile;             // Tiled?
    protected Image        bgimage;          // The background image
    protected String       bgimage_file;     // Name of the background image
    protected ImgGetr      ig;               // The Image Getter class
    protected int          bgimage_width;    // Background image width
    protected int          bgimage_height;   // Background image height
    protected int          bgimageXoffset;   // BG image X ofset
    protected int          bgimageYoffset;   // BG image Y ofset
    protected Image        tiled_bgimage;    // Tile BG?
    protected int          app_width;        // App width
    protected int          app_height;       // App height
    protected Graphics     bg_g;             // Background Graphics

    // Parameters specific to Image Maps
    protected int          num_maps;         // Number of Maps
    protected int          map_x1[];         // X1 coord for map
    protected int          map_y1[];         // Y1 coord for map
    protected int          map_x2[];         // X2 coord for map
    protected int          map_y2[];         // Y2 coord for map
    protected URL          map_url[];        // URLs for maps
    protected boolean      testmode;         // Test mode

    // Parameters specific to OffScreen images
    protected   Image         tx_osi;         // Text Off-Screen Image
    protected   Graphics      tx_g;           // Graphics of tx_osi
    protected   int           tx_pixels[];    // The text image
    protected   String        tx_filter_str;  // Text filters in string format
    protected   Image         img_osi;        // Image Off-Screen Image
    protected   Graphics      img_g;          // Graphics of img_osi
    protected   int           img_pixels[];   // The text image
    protected   String        img_filter_str; // Img filters in string format
    boolean                   retval;         // Retval from grabPixels;


    public void init() {
        int i;

        // ======== Applet ===========
        app_borderwidth = GetParm.toInt(this,     "AppBorderWidth", 0);
        app_bordercolor = GetParm.toColor(this,   "AppBorderColor", Color.black);
        app_fr_type     = GetParm.toString(this,  "AppFrameType", "ShadowIn");
        app_fr_thick    = GetParm.toInt(this,     "AppFrameThickness", 0);
        app_bg_color    = GetParm.toColor(this,   "AppBGColor", null);

        // ======== Text ===========
        tx              = GetParm.toString(this, "Text", null);
        tx_xoffset      = GetParm.toInt(this,     "TxXOffset", 0);
        tx_yoffset      = GetParm.toInt(this,     "TxYOffset", -1);
        tx_color        = GetParm.toColor(this,   "TxColor", null);
        tx_bg_color     = GetParm.toColor(this,   "TxBGColor", null);
        tx_horizcenter  = GetParm.toBoolean(this, "TxHorizCenter", false);
        tx_vertcenter   = GetParm.toBoolean(this, "TxVertCenter",  false);
        tx_font         = GetParm.toFont(this,    "TxFont",
                                                  "TxStyle",
                                                  "TxPointSize");
        tx_underline    = GetParm.toBoolean(this, "TxUnderLine", false);
        tx_borderwidth  = GetParm.toInt(this,     "TxBorderWidth", 0);
        tx_bordercolor  = GetParm.toColor(this,   "TxBorderColor", Color.black);
        tx_bordermargin = GetParm.toInt(this,     "TxBorderMargin", 0);
        tx_fr_thick     = GetParm.toInt(this,     "TxFrameThickness", 0);
        tx_fr_type      = GetParm.toString(this,  "TxFrameType", "ShadowIn");
        tx_fr_margin    = GetParm.toInt(this,     "TxFrameMargin", 0);

        setFont(tx_font);
        tx_fontmetrics = getFontMetrics(tx_font);

        // Set the background and foreground colors, if specified
        // We must set these colors here before using them
        // in the next block of code
        if(app_bg_color != null)
            setBackground(app_bg_color);
        else
            app_bg_color = getBackground();
        if(tx_color != null)
            setForeground(tx_color);
        else
            tx_color = getForeground();

        if(tx != null) {
            // This is somewhat wasteful to create this
            // because it may not be needed if the user
            // has specified TxFilter.  However, if
            // not then we must generate the pixels
            ftext ps = new ftext();
            ps.setText(tx);
            ps.setApplet(this);
            ps.setFont(tx_font);
            ps.setBackground(app_bg_color);
            ps.setForeground(tx_color);
            tx_pixels = ps.filter(null, 0, 0);
            tx_width = ps.getWidth();
            tx_height = ps.getHeight();

            tx_width  = tx_fontmetrics.stringWidth(tx);
            tx_height = tx_fontmetrics.getHeight();
            tx_ascent = tx_fontmetrics.getMaxAscent();
            tx_descent = tx_fontmetrics.getDescent();
        }

        // ======== Image ===========
        img_file        = GetParm.toString(this,  "Image", null);
        img_wait_load   = GetParm.toBoolean(this, "ImgLoadWait", false);
        img_xoffset     = GetParm.toInt(this,     "ImgXOffset", 0);
        img_yoffset     = GetParm.toInt(this,     "ImgYOffset", -1);
        img_horizcenter = GetParm.toBoolean(this, "ImgHorizCenter", false);
        img_vertcenter  = GetParm.toBoolean(this, "ImgVertCenter",  false);
        img_borderwidth = GetParm.toInt(this,     "ImgBorderWidth", 0);
        img_bordercolor = GetParm.toColor(this,   "ImgBorderColor",
                                                   Color.black);
        img_bordermargin= GetParm.toInt(this,     "ImgBorderMargin", 0);
        img_fr_thick    = GetParm.toInt(this,     "ImgFrameThickness", 0);
        img_fr_type     = GetParm.toString(this,  "ImgFrameType", "ShadowIn");
        img_fr_margin   = GetParm.toInt(this,     "ImgFrameMargin", 0);


        if(img_file != null) {
            img_gettr     = new ImgGetr(this);
            if(img_wait_load)
                repaint();
            img_osi = img_gettr.process(getDocumentBase(), img_file, false);
            img_gettr.prepareImage(img_osi, this);
            img_width = img_gettr.getWidth(img_osi, this);
            img_height = img_gettr.getHeight(img_osi, this);
            img_gettr.loadImageAndWait(img_osi, this);
        
            img_pixels = img_gettr.getPixels(img_osi, this);
        }

        // ======== Audio ===========
        tx_audio_file = GetParm.toString(this, "TxAudio", null);
        img_audio_file = GetParm.toString(this, "ImgAudio", null);

        // If the user specified an non-existent or invalid
        // file, then print error message and exit
        if(tx_audio_file != null) {
            tx_clip = getAudioClip(getDocumentBase(), tx_audio_file);
            if(tx_clip == null) {
                System.out.println(
                    "Error getting " + tx_audio_file + ". " +
                    "Make sure the file exists and is an audio file");
            }
        }
        // If the user specified an non-existent or invalid
        // file, then print error message and exit
        if(img_audio_file != null) {
            img_clip = getAudioClip(getDocumentBase(), img_audio_file);
            if(img_clip == null) {
                System.out.println(
                    "Error getting " + img_audio_file + ". " +
                    "Make sure the file exists and is an audio file");
            }
        }

        // ======== URL ===========
        tx_url     = GetParm.toURL(this, "TxURL");
        img_url    = GetParm.toURL(this, "ImgURL");

        // ======== BG Image ===========
        tile           = GetParm.toBoolean(this, "AppTile",     false);
        bgimage_file   = GetParm.toString(this,  "AppBGImage",  null);
        bgimageXoffset = GetParm.toInt(this,     "AppBGImageXOffset",  0);
        bgimageYoffset = GetParm.toInt(this,     "AppBGImageYOffset",  0);

        if(bgimage_file != null) {
            ig = new ImgGetr(this);
            bgimage = ig.process(getDocumentBase(), bgimage_file, false);
            ig.prepareImage(bgimage, this);
            bgimage_width = ig.getWidth(bgimage, this);
            bgimage_height = ig.getHeight(bgimage, this);
        }

        // ========== Image Maps ===========
        testmode = GetParm.toBoolean(this, "TestMode",   false);
        num_maps = GetParm.toInt(this,     "AppNumMaps", 0);

        if(num_maps > 0) {
            map_x1  = new int[num_maps];
            map_y1  = new int[num_maps];
            map_x2  = new int[num_maps];
            map_y2  = new int[num_maps];
            map_url = new URL[num_maps];
        }

        for(i=0;i<num_maps;i++) {
            map_x1[i]  = GetParm.toInt(this, "Map"+(i+1)+"_X1",  0);
            map_y1[i]  = GetParm.toInt(this, "Map"+(i+1)+"_Y1",  0);
            map_x2[i]  = GetParm.toInt(this, "Map"+(i+1)+"_X2",  0);
            map_y2[i]  = GetParm.toInt(this, "Map"+(i+1)+"_Y2",  0);
            map_url[i] = GetParm.toURL(this, "URL"+(i+1));
        }

        // ======== Filters ===========
        tx_filter_str  = GetParm.toString(this, "TxFilter",  null);
        img_filter_str = GetParm.toString(this, "ImgFilter", null);
    }

    //
    // Overwrite the default update method
    // This may not be necessary for this case
    // but it certainly doesn't hurt
    //
    public void update(Graphics g) {
        paint(g);
    }

    //
    // Our paint method handles several tasks:
    //    Calculates the x and y coordinates base on params
    //    Draws the applet frame
    //    Draws the applet border
    //    Draws the text frame
    //    Draws the text border
    //    Draws the underline
    //    Draws the text
    //
    public void paint(Graphics g) {
        tilebackground(g);
        if(img_osi != null) {

            img_x_coord = calcULXcoord(img_horizcenter, size().width,
                                            img_width, img_xoffset);
            img_y_coord = calcULYcoord(img_vertcenter, size().height,
                                            img_height, img_yoffset);

            g.drawImage(img_osi, img_x_coord, img_y_coord, this);
            drawFrame(g, img_x_coord, img_y_coord, img_width,
                            img_height, img_fr_type,
                            img_fr_thick, img_fr_margin);
            drawBorder(g, img_x_coord, img_y_coord, img_width,
                            img_height, img_borderwidth,
                            img_bordermargin, img_bordercolor);
        }
        // We never know when the size might change so if horizcenter
        // is true then recalculate the x_coord every time
        if(tx_osi != null) {
            tx_x_coord = calcULXcoord(tx_horizcenter, size().width,
                            tx_width, tx_xoffset);
            tx_y_coord = calcULYcoord(tx_vertcenter, size().height,
                                tx_height, tx_yoffset);

            // Draw frame for applet
            drawFrame(g, 0, 0, size().width, size().height,
                                app_fr_type, app_fr_thick, 0);
            // Draw border for applet
            drawBorder(g, 0, 0, size().width, size().height,
                                app_borderwidth, 0, app_bordercolor);

            // Draw the text
            g.drawImage(tx_osi, tx_x_coord, tx_y_coord, this);

            // Draw frame for text
            drawFrame(g, tx_x_coord, tx_y_coord,
                                tx_width, tx_height,
                                tx_fr_type, tx_fr_thick, tx_fr_margin);


            // Draw border for text
            drawBorder(g, tx_x_coord, tx_y_coord,
                                tx_width, tx_height,
                                tx_borderwidth, tx_bordermargin,
                                tx_bordercolor);
        }
    }

    //
    // If the user clicks the mouse then load the URL
    // We check the text position first, so if there is
    // overlapping text and image then the text URL will
    // be loaded, not the imgae URL
    //
    public boolean mouseDown(Event evt, int x, int y) {
        int i;

        // Handle the image maps first
        for(i=0;i<num_maps;i++) {
            if(x > map_x1[i] && x < map_x2[i] &&
               y > map_y1[i] && y < map_y2[i]) {
                if(map_url[i] != null) {
                    if(testmode) {
                        getAppletContext().showStatus("Loading URL: " +
                                    map_url[i].toString());
                        System.out.println("Loading URL: " +
                                    map_url[i].toString());
                    } else
                        getAppletContext().showDocument(map_url[i]);
                    return true;
                }
            }
        }    

        // Handle the text next (Text takes precedence over images)
        if(x > tx_x_coord && x < tx_x_coord + tx_width &&
           y > tx_y_coord && y < tx_y_coord + tx_height) {
            if(tx_url != null) {
                getAppletContext().showDocument(tx_url);
                System.out.println("loading " + tx_url.toString());
                return true;
            }
        }    

        // Handle the image last
        if(x > img_x_coord && x < img_x_coord + img_width &&
           y > img_y_coord && y < img_y_coord + img_height) {
            if(img_url != null) {
                getAppletContext().showDocument(img_url);
                System.out.println("loading " + img_url.toString());
                return true;
            }
        }    
        return true;
    }

    //
    // Check the x and y coordinates to see if we're actually
    // inside the boundaries of the text or image or both
    //
    public boolean mouseMove(Event evt, int x, int y) {
        boolean tx_status_shown = false;
        boolean img_status_shown = false;
        int i;

        if(testmode) {
            getAppletContext().showStatus("X: " + x + " Y: " + y);
            return true;
        }

        //
        // If a image map url was found (hence tx_status_shown == true)
        // then return and don't bother to call the other mouseMove
        // stuff would simply display the url for the text or image.
        //
        if(tx_status_shown == true) {
            return true;
        }

        // ========== AUDIO =============
        // Process the image first
        if(x > img_x_coord && x < img_x_coord + img_width &&
           y > img_y_coord && y < img_y_coord + img_height) {
            if(!img_playing) {
                img_playing = true;
                if(img_clip != null)
                    img_clip.play();
            } // else it's already playing, leave it alone
        } else {
            // Since we don't get any kind of notification when the audio
            // clip has terminated (due to playing to completion) we set
            // playing to false everytime we move out of the boundaries
            // of the string
            img_playing = false;
            if(img_clip != null) {
                img_clip.stop();
            }
        }

        // Process the text last
        if(x > tx_x_coord && x < tx_x_coord + tx_width &&
           y > tx_y_coord && y < tx_y_coord + tx_height) {
            if(!tx_playing) {
                tx_playing = true;
                if(tx_clip != null)
                    tx_clip.play();
            } // else it's already playing, leave it alone
        } else {
            // Since we don't get any kind of notification when the audio
            // clip has terminated (due to playing to completion) we set
            // playing to false everytime we move out of the boundaries
            // of the string
            tx_playing = false;
            if(tx_clip != null) {
                tx_clip.stop();
            }
        }

        // ========== IMAGE MAPS =============
        for(i=0;i<num_maps;i++) {
            if(x > map_x1[i] && x < map_x2[i] &&
               y > map_y1[i] && y < map_y2[i]) {
                if(map_url[i] != null) {
                    getAppletContext().showStatus(map_url[i].toString());
                    tx_status_shown = true;
                }
            }
        }

        // ========== URLs =============
        // Handle the text first
        if(x > tx_x_coord && x < tx_x_coord + tx_width &&
           y > tx_y_coord && y < tx_y_coord + tx_height) {
            if(tx_url != null) {
                getAppletContext().showStatus(tx_url.toString());
                tx_status_shown = true;
            }
        }    

        // Handle the image last
        // If a Text URL is being displayed then don't
        // display the image URL
        if(x > img_x_coord && x < img_x_coord + img_width &&
           y > img_y_coord && y < img_y_coord + img_height) {
            if(img_url != null && tx_status_shown == false) {
                getAppletContext().showStatus(img_url.toString());
                img_status_shown = true;
            }
        }    
        // If we are not over either text or image that has 
        // a corresponding URL then clear the status line
        if(tx_status_shown == false && img_status_shown == false)
            getAppletContext().showStatus(null);
        return true;
    }

    //
    // Just to be sure, when we exit the applet, stop the clips
    // It should already be stopped
    //
    public boolean mouseExit(Event evt, int x, int y) {
        // Just to be sure
        tx_playing = false;
        if(tx_clip != null)
            tx_clip.stop();
        img_playing = false;
        if(img_clip != null)
            img_clip.stop();
        getAppletContext().showStatus(null);
        return true;
    }

    //
    // Calculate the upper left corner position when centering
    // an item that's drawn in relation to its upper left corner
    // such as images
    //
    protected int calcULCenter(int panelsize, int objectsize) {
        return ((panelsize / 2) - (objectsize / 2));
    }

    //
    // Calculate the lower left corner position when centering
    // an item that's drawn in relation to its lower left corner
    // such as text
    //
    protected int calcLLCenter(int panelsize, int objectsize) {
        return ((panelsize / 2) + (objectsize / 2));
    }

    //
    // A general purpose text drawing method
    // This method is used by the Img class
    //
    protected void drawText(Graphics g, String txt, int x, int y,
                                            Color c, Font f) {
        if(txt != null) {
            g.setColor(c);
            g.setFont(f);
            g.drawString(txt, x, y);
        }
        
    }

    //
    // The drawBorder method draws a border at the 
    // specified x and y position
    //
    protected void drawBorder(Graphics g,
                           int x,
                           int y,
                           int w,
                           int h,
                           int borderwidth,
                           int margin,
                           Color c) {
        if(borderwidth > 0) {
            g.setColor(c);
            for(int i=0;i<borderwidth;i++) {
                g.drawRect(x-margin+i,
                           y-margin+i,
                           w+(margin*2)-(i*2)-1,
                           h+(margin*2)-(i*2)-1);
            }
        }
    }

    void drawTopLine(Graphics g,
                    int x, int y, int w, int h, int i, int margin) {
        g.drawLine(x-margin+i, y-margin+i, x+w+margin-i-1, y-margin+i);
    }

    void drawBottomLine(Graphics g,
                    int x, int y, int w, int h, int i, int margin) {
        g.drawLine(x-margin+i, y+h+margin-i-1, x+w+margin-i-1, y+h+margin-i-1);
    }

    void drawLeftLine(Graphics g,
                    int x, int y, int w, int h, int i, int margin) {
        g.drawLine(x-margin+i, y-margin+i, x-margin+i, y+h+margin-i-1);
    }

    void drawRightLine(Graphics g,
                    int x, int y, int w, int h, int i, int margin) {
        g.drawLine(x+w+margin-i-1, y-margin+i, x+w+margin-i-1, y+h+margin-i-1);
    }
        

    //
    // The drawFrame draws a frame at the specified x and y position
    // A frame can be one of the following four types: ShadowIn,
    // ShadowOut, ShadowEtchedIn, ShadowEtchedOut
    //
    protected void drawFrame(Graphics g,    // Graphics to draw to
                           int x,           // X upper left position
                           int y,           // Y upper left position
                           int w,           // width
                           int h,           // height
                           String type,     // Type of frame
                           int thickness,   // thickness of frame
                           int margin) {    // Margin around object
        int i;
        if(thickness == 0)
            return;
        Color darker = AdjColor.darkenit(app_bg_color, .50);
        Color slightlydarker = AdjColor.darkenit(app_bg_color, .10);
        Color brighter = AdjColor.brightenit(app_bg_color, .50);

        if(thickness > 0) {
            if(type.equalsIgnoreCase("shadowout")) {
                for(i=0;i<thickness;i++) {
                    g.setColor(brighter);
                    // TOP
                    drawTopLine(g, x, y, w, h, i, margin);
                    // LEFT
                    drawLeftLine(g, x, y, w, h, i, margin);

                    g.setColor(darker);
                    // BOTTOM
                    drawBottomLine(g, x, y, w, h, i, margin);
                    // RIGHT
                    drawRightLine(g, x, y, w, h, i, margin);
                }
            } else if(type.equalsIgnoreCase("shadowetchedin")) {
                for(i=0;i<thickness;i++) {
                    if(i == 0)
                        g.setColor(darker);
                    else if (i == thickness-1)
                        g.setColor(brighter);
                    else
                        g.setColor(slightlydarker);
                    // TOP
                    drawTopLine(g, x, y, w, h, i, margin);
                    // LEFT
                    drawLeftLine(g, x, y, w, h, i, margin);

                    if(i == 0)
                        g.setColor(brighter);
                    else if (i == thickness-1)
                        g.setColor(darker);
                    else
                        g.setColor(slightlydarker);
                    // BOTTOM
                    drawBottomLine(g, x, y, w, h, i, margin);
                    // RIGHT
                    drawRightLine(g, x, y, w, h, i, margin);
                }

            } else if(type.equalsIgnoreCase("shadowetchedout")) {
                for(i=0;i<thickness;i++) {
                    if(i == 0)
                        g.setColor(brighter);
                    else if (i == thickness-1)
                        g.setColor(darker);
                    else
                        g.setColor(app_bg_color);
                    // TOP
                    drawTopLine(g, x, y, w, h, i, margin);
                    // LEFT
                    drawLeftLine(g, x, y, w, h, i, margin);

                    if(i == 0)
                        g.setColor(darker);
                    else if (i == thickness-1)
                        g.setColor(brighter);
                    else
                        g.setColor(app_bg_color);
                    // BOTTOM
                    drawBottomLine(g, x, y, w, h, i, margin);
                    // RIGHT
                    drawRightLine(g, x, y, w, h, i, margin);
                }

            } else {    // Default to "shadowin"
                for(i=0;i<thickness;i++) {
                    g.setColor(darker);
                    // TOP
                    drawTopLine(g, x, y, w, h, i, margin);
                    // LEFT
                    drawLeftLine(g, x, y, w, h, i, margin);

                    g.setColor(brighter);
                    // BOTTOM
                    drawBottomLine(g, x, y, w, h, i, margin);
                    // RIGHT
                    drawRightLine(g, x, y, w, h, i, margin);
                }
            }
        }
    }

    //
    // The drawUnderline method draws an underline
    // The underline is actually a filled rectangle
    // whose size is based on the width of a pipe
    // character from the same font.  It works!
    //
    protected void drawUnderline(Graphics g,
                           int x,
                           int y,
                           int w,
                           Color c) {
        if(tx_underline == true) {
            int pipewidth = tx_fontmetrics.charWidth('|');;
            g.setColor(c);
            g.fillRect(x, y, w, Math.max(1, pipewidth/4));
        }
    }

    //
    // Calculate the upper left X coordinate based
    // on parameters such as horizcenter and xoffset
    //
    protected int calcULXcoord(boolean hcent, int appletwidth,
                                int objectwidth, int xoffset) {
        int x;

        // Horizontal (x coordinate) Position
        if(hcent)
            x = calcULCenter(appletwidth, objectwidth);
        else if (xoffset == -1)
            x = 0;
        else
            x = xoffset;
        
        return x;
    }

    //
    // Calculate the upper left Y coordinate based
    // on parameters such as vertcenter and yoffset
    //
    protected int calcULYcoord(boolean hcent, int appletheight,
                            int objectheight, int yoffset) {
        int y;

        // Vertical (y coordinate) Position
        if(hcent)
            y = calcULCenter(appletheight, objectheight);
        else if (yoffset == -1)
            y = 0;
        else
            y = yoffset;

        return y;
    }

    //
    // Calculate the lower left Y coordinate based
    // on parameters such as vertcenter and yoffset
    //
    protected int calcLLYcoord(boolean hcent, int appletheight,
                    int objectheight, int objectdescent, int yoffset) {
        int y;

        // Vertical (y coordinate) Position
        if(hcent)
            y = calcLLCenter(appletheight, objectheight) - objectdescent;
        else if (yoffset == -1)
            y = objectheight - objectdescent;
        else
            y = yoffset;

        return y;
    }

    // 
    // Tile the background with the specified image
    //
    protected void tilebackground(Graphics g) {
        int i, j;

        if(bgimage == null)
            return;
        if(ig.isImagePrepared(bgimage, this) == false)
            ig.loadImageAndWait(bgimage, this);
        //
        // If the applet has changed size or this is the first time
        // tilebackground has been called then create the bg image
        //
        if(app_width != size().width || app_height != size().height) {
            app_width  = size().width;
            app_height = size().height;

            tiled_bgimage = createImage(size().width, size().height); 
            bg_g = tiled_bgimage.getGraphics();
            bg_g.setColor(app_bg_color);
            bg_g.fillRect(0, 0, size().width, size().height);
            if(tile & (bgimageXoffset != 0 || bgimageYoffset != 0)) {
                for(i=-bgimage_height;i<app_height;i+=bgimage_height) {
                    for(j=-bgimage_width;j<app_width;j+=bgimage_width) {
                        bg_g.drawImage(bgimage, j+bgimageYoffset,
                                        i+bgimageXoffset, this);
                    }
                }

            } else if(tile) {
                for(i=-bgimage_height;i<app_height;i+=bgimage_height) {
                    for(j=-bgimage_width;j<app_width;j+=bgimage_width) {
                        bg_g.drawImage(bgimage, j+2, i+2, this);
                    }
                }
            } else {
                bg_g.drawImage(bgimage, bgimageXoffset,
                                        bgimageYoffset, this);
            }
        }

        if(tiled_bgimage != null) {
            g.drawImage(tiled_bgimage, 0, 0, this);
        }
    }

    public void doStandardFilters() {
        Filter f;

        //========= INVOKE IMAGE FILTERS ==============
        f = new Filter();
        if(img_osi != null && img_filter_str != null) {
            img_pixels = f.invoke(this, null, img_filter_str, img_pixels,
                            img_width, img_height, false, app_bg_color,
                            false);
            img_width  = f.getWidth();
            img_height = f.getHeight();
        }

        //========= CREATE IMAGE FROM PIXELS ==============
        if(img_osi != null) {
            img_osi = createImage(new MemoryImageSource(img_width,
                            img_height, ColorModel.getRGBdefault(),
                            img_pixels, 0, img_width));
        }

        //========= INVOKE FILTERS ==============
        if(tx != null && tx_filter_str != null) {
            if(tx_bg_color == null) {
                tx_pixels = f.invoke(this, tx, tx_filter_str, tx_pixels,
                            tx_width, tx_height, true, app_bg_color,
                            tx_underline);
            } else {
                tx_pixels = f.invoke(this, tx, tx_filter_str, tx_pixels,
                            tx_width, tx_height, false, tx_bg_color,
                            tx_underline);
            }
            tx_width  = f.getWidth();
            tx_height = f.getHeight();
        }

        //========= CREATE IMAGE FROM PIXELS ==============
        if(tx != null) {
            tx_osi = createImage(new MemoryImageSource(tx_width,
                            tx_height, ColorModel.getRGBdefault(),
                            tx_pixels, 0, tx_width));
        }
    }

    public synchronized boolean imageUpdate(Image img,
                                            int infoflags, 
                                            int x, int y,
                                            int width,
                                            int height){
        if((infoflags & ERROR) != 0) {
            if(img.equals(this.img_osi))
                System.out.println("Error getting image = " + img_file);
            else if(img.equals(bgimage))
                System.out.println("Error getting image = " + bgimage_file);
            img.flush();
            img = null;
            return false;
        }
        if((infoflags & ABORT) != 0) {
            if(img.equals(this.img_osi))
                System.out.println("Abort image = " + img_file);
            else if(img.equals(bgimage))
                System.out.println("Abort image = " + bgimage_file);
            img.flush();
            img = null;
            return false;
        }
        if((infoflags & ALLBITS) != 0) {
            return true;
        }
        return super.imageUpdate(img, infoflags, x, y, width, height);
    }
}
